package biggests.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.tree.*;

import biggests.tree.*;
import biggests.utils.*;
import de.schlichtherle.io.File;
import smadeira.biclustering.*;
import smadeira.ontologizer.GeneSetResult;
import smadeira.ontologizer.GOFrontEnd;
import java.applet.Applet;
import java.net.URL;
import java.applet.AppletContext;
import java.net.MalformedURLException;
import java.util.Arrays;
import java.lang.reflect.Method;

/**
 * <p>Title: BiGGEsTS Frame</p>
 *
 * <p>Description: The main frame of the application.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
public class BiggestsFrame extends JFrame implements ActionListener {

/*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//             ATTRIBUTES
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************/
    /**
     * The paths to the main directories of BiGGEsTS:
     * Function_Results -> to store resulting files from function analysis
     * GO_Files -> directory where the application keeps gene ontology files
     * Sessions -> the preferred location to save session archive files
     * Datasets -> the preferred location to keep the dataset files
     */
    private static String[] pathsToFolders = {"Function_Results",
                                             "GO_Files",
                                             "Sessions",
                                             "Datasets"};

    /**
     * Path to DOT binary (external program to convert dot
     * files to svg/png files).
     */
    private static String pathToDOT = null;

    private static String pathToNumDend = "numdend.txt";

    public static int SCROLL_BAR_INCREMENT = 100;

    public static boolean locked = false;

    /**
     * URL to download gene ontology files.
     * (files like gene_ontology.obo and related)
     */
    private static String geneOntologyFilesURL =
            "ftp://ftp.geneontology.org/pub/go/ontology/obo_format_1_0/";

    /**
     * URL to download gene association files.
     * (files like gene_association.[organism_specific_file_extension])
     */
    private static String geneAssociationFilesURL =
            "ftp://ftp.geneontology.org/pub/go/gene-associations/";

    /**
     * Main panel from main frame, which holds all other graphic components.
     */
    private JPanel panelContent;

    /**
     * Right panel from main frame, which holds the tabbed pane
     * and the context label.
     */
    private JPanel panelRight;

    /**
     * Dataset matrices and biclusters tree.
     */
    private DatasetTree tree;

    /**
     * Information text pane - displays information about the current selected
     * dataset, matrix or bicluster on the tree.
     */
    private JTextPane textPaneInfo;

    /**
     * Context label, which contains information about what the user may do.
     */
    private JTextField labelContext;

    /**
     * Tabbed pane which contains the main functionalities of the application:<br/>
     * Loading;<br/>
     * Preprocessing;<br/>
     * Biclustering;<br/>
     * Analyzing;<br/>
     * Post-processing;<br/>
     * Comparing.
     */
    private JTabbedPane tabbedPane;

    /**
     * Analyzing tabbed pane that contains the analyzing options:<br/>
     * Matrix analysis;<br/>
     * Expression analysis;<br/>
     * Function analysis.
     */
    private JTabbedPane tabbedPaneAnalyzing;

    /**
     * Matrix Analyzing tabbed pane which contains the matrix analyzing options:<br/>
     * Matrix analysis by values;<br/>
     * Matrix analysis by colors;<br/>
     * Matrix analysis by symbols.
     */
    private JTabbedPane tabbedPaneAnalyzingMatrix;

    /**
     * Shows function analysis options.
     */
    private JTabbedPane tabbedPaneAnalyzingFunction;


    /**
     * Tree scroll pane: scrolls tree view.
     */
    private JScrollPane scrollTree;

    /**
     * Info area scroll pane: scrolls the information text area.
     */
    private JScrollPane scrollInfo;

    /**
     * Load scroll pane: scrolls the load panel.
     */
    private JScrollPane scrollLoading;

    /**
     * Preprocessing scroll pane: scrolls the preprocessing panel.
     */
    private JScrollPane scrollPreProcessing;

    /**
     * Biclustering scroll pane: scrolls the biclustering panel.
     */
    private JScrollPane scrollBiclustering;

    /**
     * Post-processing scroll pane: scrolls the post-processing panel.
     */
    private JScrollPane scrollPostProcessing;

    /**
     * Comparing scroll pane: scrolls the comparing panel.
     */
//    private JScrollPane scrollComparing;

    /**
     * Scrolls expression analysis - bicluster time points chart panel.
     */
    private JScrollPane scrollAnalyzingExpressionBiclusterTimePoints;

    /**
     * Scrolls expression analysis - all time points chart panel.
     */
    private JScrollPane scrollAnalyzingExpressionAllTimePoints;

    /**
     * Scrolls expression analysis - bicluster pattern chart panel.
     */
    private JScrollPane scrollAnalyzingExpressionBiclusterPattern;

    /**
     * Scrolls function analysis - functions table panel.
     */
    private JScrollPane scrollAnalyzingFunctionTable;

    /**
     * Scrolls function analysis - significance graph panel.
     */
    private JScrollPane scrollAnalyzingFunctionSignificanceGraph;

    /**
     * Scrolls matrix analysis by values panel.
     */
    private JScrollPane scrollAnalyzingMatrixValues;

    /**
     * Scrolls matrix analysis by colors panel.
     */
    private JScrollPane scrollAnalyzingMatrixColors;

    /**
     * Scrolls matrix analysis by symbols panel.
     */
    private JScrollPane scrollAnalyzingMatrixSymbols;

    /**
     * Scrolls matrix analysis by dendrogram panel.
     */
    private JScrollPane scrollAnalyzingMatrixDendrogram;

    /**
     * Loading panel: panel which contains the load options.
     */
    private PanelLoading panelLoading;

    /**
     * Preprocessing panel: shows the options for preprocessing tasks.
     */
    private PanelPreProcessing panelPreProcessing;

    /**
     * Biclustering panel: shows all the options to apply biclustering
     * algorithms.
     */
    private PanelBiclustering panelBiclustering;

    /**
     * Post-processing panel: shows options to post-process biclusters.
     */
    private PanelPostProcessing panelPostProcessing;

    /**
     * Comparing panel: shows options to compare biclusters.
     */
//    private JPanel panelComparing;

    /**
     * Shows expression analysis options.
     */
    private JTabbedPane tabbedPaneAnalyzingExpression;

    /**
     * Shows expression analysis of bicluster's genes in bicluster time points.
     */
    private PanelExpression panelAnalyzingExpressionBiclusterTimePoints;

    /**
     * Shows expression analysis of bicluster's genes in all time points.
     */
    private PanelExpression panelAnalyzingExpressionAllTimePoints;

    /**
     * Shows expression analysis of bicluster's pattern.
     */
    private PanelExpression panelAnalyzingExpressionBiclusterPattern;

    /**
     * Shows matrix analysis by values.
     */
    private PanelAnalyzingMatrixTables panelAnalyzingMatrixValues;

    /**
     * Shows matrix analysis by colors.
     */
    private PanelAnalyzingMatrixTables panelAnalyzingMatrixColors;

    /**
     * Shows matrix analysis by symbols.
     */
    private PanelAnalyzingMatrixTables panelAnalyzingMatrixSymbols;

    /**
     * Shows matrix analysis by dendrogram.
     */
    private PanelAnalyzingMatrixDendrogram panelAnalyzingMatrixDendrogram;

    /**
     * Shows function analysis - functions table.
     */
    private PanelAnalyzingFunctionTable panelAnalyzingFunctionTable;

    /**
     * Shows function analysis - significant functions graph.
     */
    private PanelAnalyzingFunctionSignificanceGraph
            panelAnalyzingFunctionSignificanceGraph;

    /**
     * Left-right split pane: splits left and right sides of the main panel.
     */
    private JSplitPane splitLeftRight;

    /**
     * Top-bottom split pane: splits top and bottom in the left side panel
     * of the main panel.
     */
    private JSplitPane splitTopBottom;

    /**
     * Application's menu bar.
     */
    private JMenuBar menu;

    /**
     * Session menu.
     */
    private JMenu session;

    /**
     * Load dataset menu item.
     */
    private JMenuItem sessionLoadDataset;

    /**
     * Load session menu item.
     */
    private JMenuItem sessionLoadSession;

    /**
     * Save session menu item.
     */
    private JMenuItem sessionSaveSession;

    /**
     * Exit application menu item.
     */
    private JMenuItem sessionExit;

    /**
     * Help menu.
     */
    private JMenu help;

    /**
     * Help about menu item.
     */
    private JMenuItem helpAbout;

    /**
     * Settings menu.
     */
    private JMenu settings;

    /**
     * Settings colors menu.
     */
    private JMenu settingsColors;

    /**
     * Red and green color scheme.
     */
    private JMenuItem colorsRedGreen;

    /**
     * Blue and yellow color scheme.
     */
    private JMenuItem colorsYellowBlue;

    /*****************************************************************************
     //*****************************************************************************
     //*****************************************************************************
     //                  CONSTRUCTOR
     //*****************************************************************************
     //*****************************************************************************
     //*****************************************************************************/
    /**
     * Application's main frame constructor.
     *
     */
    public BiggestsFrame() {
        try {
            File dend = new File(pathToNumDend);
            if (dend.exists() && dend.isFile()) {
                BufferedReader br = new BufferedReader(new FileReader(dend));
                PanelAnalyzingMatrixDendrogram.setNumberOfDendrograms(
                        Integer.parseInt(br.readLine()));
                br.close();
            }
            if (pathToDOT == null) {
                BiggestsInstall.initDOT();
            }

            BiggestsIO.initializeFolders(pathsToFolders);
            BiggestsIO.deleteFolderContent(this.getPathToFunctionResults());
            PanelAnalyzingMatrixDendrogram.deleteTempFilesAndDirectory();
        } catch (IOException ex) {
            JOptionPane.showMessageDialog(null, "Initialization error\n" +
                                          ex.getMessage(), "Error",
                                          JOptionPane.ERROR_MESSAGE);
        }

        try {
            setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
            jbInit();
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /**
     * Component initialization.
     *
     * @throws java.lang.Exception
     */
    private void jbInit() throws Exception {
        this.setTitle("BiGGEsTS - Biclustering Gene Expression Time Series");
        this.addComponentListener(new BiggestsFrame_this_componentAdapter(this));
        this.addWindowStateListener(new
                                    BiggestsFrame_this_windowStateAdapter(this));
        this.addWindowListener(new BiggestsFrame_this_windowAdapter(this));
        this.setMinimumSize(new Dimension(200,200));
        if (BiggestsInstall.isLinux(System.getProperty("os.name").toLowerCase())) {
            this.setSize(new Dimension(600, 600));
        }
        session = new JMenu("Session");
        sessionLoadDataset = new JMenuItem("Load dataset");
        sessionLoadSession = new JMenuItem("Load session");
        sessionSaveSession = new JMenuItem("Save session");
        sessionExit = new JMenuItem("Exit");

        sessionLoadDataset.addActionListener(this);
        sessionLoadSession.addActionListener(this);
        sessionSaveSession.addActionListener(this);
        sessionExit.addActionListener(this);

        session.add(sessionLoadDataset);
        session.add(sessionLoadSession);
        session.add(sessionSaveSession);
        session.add(sessionExit);

        settings = new JMenu("Settings");
        settingsColors = new JMenu("Color scheme");
        colorsRedGreen = new JRadioButtonMenuItem("Red and green");
        colorsYellowBlue = new JRadioButtonMenuItem("Yellow and blue");
        ButtonGroup buttonGroup = new ButtonGroup();
        buttonGroup.add(colorsRedGreen);
        buttonGroup.add(colorsYellowBlue);
        settingsColors.add(colorsRedGreen);
        settingsColors.add(colorsYellowBlue);
        settings.add(settingsColors);
        colorsRedGreen.addActionListener(this);
        colorsYellowBlue.addActionListener(this);
        colorsRedGreen.setSelected(true);

        help = new JMenu("Help");
        helpAbout = new JMenuItem("About...");
        helpAbout.addActionListener(this);
        help.add(helpAbout);

        menu = new JMenuBar();
        menu.add(session);
        menu.add(settings);
        menu.add(help);
        this.setJMenuBar(menu);

        labelContext = new JTextField();
        labelContext.setHorizontalAlignment(SwingConstants.CENTER);
        labelContext.setText("Load a dataset.");
        labelContext.setEditable(false);

        final BiggestsFrame frame = this;
        tree = new DatasetTree() {
            public void setSelectionPath(TreePath selectionPath) {
                if (selectionPath == null) {
                    return;
                }
                Object node = ((DefaultMutableTreeNode)
                                                 selectionPath.
                                                 getLastPathComponent()).
                                     getUserObject();

                if (node instanceof NodeInfo) {
                    int type = ((NodeInfo)node).getNodeObjectType();
                    if (type == NodeInfo.BICLUSTER_SINGLE ||
                        type == NodeInfo.BICLUSTER_GROUP ||
                        type == NodeInfo.POSTPROCESSED_GROUP ||
                        type == NodeInfo.POSTPROCESSED_SINGLE) {
                        lock();
                        frame.startTaskInLabelContext("Rendering...");
                    }
                }

                Thread applyTreePathSelectionThread =
                        new TreePathSelectionThread(
                        selectionModel, selectionPath);
                applyTreePathSelectionThread.start();
            }
        };

        scrollTree = new JScrollPane();
        scrollTree.getViewport().add(tree, null);
        scrollTree.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);

        textPaneInfo = new JTextPane();
        textPaneInfo.setMargin(new Insets(10, 10, 10, 10));
        textPaneInfo.setEditable(false);
        textPaneInfo.setText("Information\n\nThis box displays information " +
                             "about matrices, algorithm applications, etc.");
        scrollInfo = new JScrollPane();
        scrollInfo.getViewport().add(textPaneInfo, null);
        scrollInfo.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);

        panelLoading = new PanelLoading(this);
        panelPreProcessing = new PanelPreProcessing(this);
        panelBiclustering = new PanelBiclustering(this);
        panelPostProcessing = new PanelPostProcessing(this);
//        panelComparing = new JPanel();

        panelAnalyzingExpressionBiclusterTimePoints =
            new CustomPanelExpression(this,
                                      PanelExpression.BICLUSTER_TIME_POINTS,
                                      "Analyze bicluster time points charts.");
        panelAnalyzingExpressionAllTimePoints =
            new CustomPanelExpression(this, PanelExpression.ALL_TIME_POINTS,
            "Analyze all time points charts.");
        panelAnalyzingExpressionBiclusterPattern =
            new CustomPanelExpression(this, PanelExpression.PATTERN,
                                "Analyze pattern charts.");

        scrollAnalyzingExpressionBiclusterTimePoints =
                new JScrollPane(this.
                                panelAnalyzingExpressionBiclusterTimePoints);
        scrollAnalyzingExpressionBiclusterTimePoints.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollAnalyzingExpressionAllTimePoints =
                new JScrollPane(this.panelAnalyzingExpressionAllTimePoints);
        scrollAnalyzingExpressionAllTimePoints.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollAnalyzingExpressionBiclusterPattern =
                new JScrollPane(this.panelAnalyzingExpressionBiclusterPattern);
        scrollAnalyzingExpressionBiclusterPattern.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);


        tabbedPaneAnalyzingExpression = new TabbedPaneAnalyzingExpression(this);
        tabbedPaneAnalyzingExpression.add("Bicluster time points",
                                          scrollAnalyzingExpressionBiclusterTimePoints);
        tabbedPaneAnalyzingExpression.add("All time points",
                                          scrollAnalyzingExpressionAllTimePoints);
        tabbedPaneAnalyzingExpression.add("Bicluster pattern",
                                          scrollAnalyzingExpressionBiclusterPattern);
        tabbedPaneAnalyzingExpression.setTabPlacement(JTabbedPane.BOTTOM);

        panelAnalyzingFunctionTable = new PanelAnalyzingFunctionTable(this);
        panelAnalyzingFunctionSignificanceGraph = new
                PanelAnalyzingFunctionSignificanceGraph(this);
        scrollAnalyzingFunctionTable = new JScrollPane(
                panelAnalyzingFunctionTable);
        scrollAnalyzingFunctionTable.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollAnalyzingFunctionSignificanceGraph = new JScrollPane(
                panelAnalyzingFunctionSignificanceGraph);
        scrollAnalyzingFunctionSignificanceGraph.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);

        /**
         * add the function analyzing functionality tabs:
         *
         * Function analysis - table tab -> index 0
         * Matrix analysis - significance graph tab -> index 1
         */
        tabbedPaneAnalyzingFunction = new JTabbedPane();

        tabbedPaneAnalyzingFunction.add("Table", scrollAnalyzingFunctionTable);
        tabbedPaneAnalyzingFunction.add("Significance Graph",
                scrollAnalyzingFunctionSignificanceGraph);

        tabbedPaneAnalyzingFunction.setTabPlacement(JTabbedPane.BOTTOM);

        panelAnalyzingMatrixValues = new CustomPanelAnalyzingMatrixTables(
                this, PanelAnalyzingMatrixTables.VALUES,
                "Analyze expression matrices.");
        panelAnalyzingMatrixColors = new CustomPanelAnalyzingMatrixTables(
                this, PanelAnalyzingMatrixTables.COLORS,
                "Analyze expression heatmaps.");
        panelAnalyzingMatrixSymbols = new CustomPanelAnalyzingMatrixTables(
                this, PanelAnalyzingMatrixTables.SYMBOLS,
                "Analyze discretized heatmaps.");

        panelAnalyzingMatrixDendrogram = new PanelAnalyzingMatrixDendrogram(this);

        scrollAnalyzingMatrixValues = new JScrollPane(
                panelAnalyzingMatrixValues);
        scrollAnalyzingMatrixValues.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollAnalyzingMatrixColors = new JScrollPane(
                panelAnalyzingMatrixColors);
        scrollAnalyzingMatrixColors.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollAnalyzingMatrixSymbols = new JScrollPane(
                panelAnalyzingMatrixSymbols);
        scrollAnalyzingMatrixSymbols.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollAnalyzingMatrixDendrogram = new JScrollPane(
                panelAnalyzingMatrixDendrogram);
        scrollAnalyzingMatrixDendrogram.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);

        /**
         * add the matrix analyzing functionality tabs:
         *
         * Matrix analysis by values tab -> index 0
         * Matrix analysis by colors tab -> index 1
         * Matrix analysis by symbols tab -> index 2
         */
        tabbedPaneAnalyzingMatrix = new TabbedPaneAnalyzingMatrix(this);

        tabbedPaneAnalyzingMatrix.add("Values", scrollAnalyzingMatrixValues);
        tabbedPaneAnalyzingMatrix.add("Colors", scrollAnalyzingMatrixColors);
        tabbedPaneAnalyzingMatrix.add("Symbols", scrollAnalyzingMatrixSymbols);
        tabbedPaneAnalyzingMatrix.add("Dendrogram",
                                      scrollAnalyzingMatrixDendrogram);
        tabbedPaneAnalyzingMatrix.setTabPlacement(JTabbedPane.BOTTOM);

        /**
         * add the analyzing functionality tabs:
         *
         * Matrix analysis tab -> index 0
         * Expression analysis tab -> index 1
         * Function analysis tab -> index 2
         */
        tabbedPaneAnalyzing = new TabbedPaneAnalyzing(this);

        tabbedPaneAnalyzing.add("Matrix", tabbedPaneAnalyzingMatrix);
        tabbedPaneAnalyzing.add("Expression", tabbedPaneAnalyzingExpression);
        tabbedPaneAnalyzing.add("Function", tabbedPaneAnalyzingFunction);

        scrollLoading = new JScrollPane(panelLoading);
        scrollLoading.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollPreProcessing = new JScrollPane(panelPreProcessing);
        scrollPreProcessing.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollBiclustering = new JScrollPane(panelBiclustering);
        scrollBiclustering.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
        scrollPostProcessing = new JScrollPane(panelPostProcessing);
        scrollPostProcessing.getVerticalScrollBar().
                setUnitIncrement(SCROLL_BAR_INCREMENT);
//        scrollComparing = new JScrollPane(panelComparing);

        /**
         * add the main functionality tabs:
         *
         * Loading tab -> index 0
         * Preprocessing tab -> index 1
         * Biclustering tab -> index 2
         * Analyzing tab -> index 3
         * Post-processing tab -> index 4
         * Comparing tab -> index 5
         */
        tabbedPane = new TabbedPane(this);

        tabbedPane.add("Loading", scrollLoading);
        tabbedPane.add("Preprocessing", scrollPreProcessing);
        tabbedPane.add("Biclustering", scrollBiclustering);
        tabbedPane.add("Analyzing", tabbedPaneAnalyzing);
        tabbedPane.add("Post-processing", scrollPostProcessing);
//        tabbedPane.add("Comparing", scrollComparing);

        /**
         * set initial enabled/disabled tabs:
         *
         * Loading -> index 0
         * Preprocessing -> index 1
         * Biclustering -> index 2
         * Analyzing -> index 3
         * Post-processing -> index 4
         * Comparing -> index 5
         */
        this.setTabbedPaneEnabledTabs(true, false, false, false, false, false);

        panelRight = new JPanel();
        panelRight.setLayout(new BorderLayout());
        panelRight.add(tabbedPane, BorderLayout.CENTER);
        panelRight.add(labelContext, BorderLayout.SOUTH);

        splitTopBottom = new JSplitPane();
        splitTopBottom.setOrientation(JSplitPane.VERTICAL_SPLIT);
        splitTopBottom.setLeftComponent(scrollTree);
        splitTopBottom.setRightComponent(scrollInfo);
        splitTopBottom.add(scrollTree, JSplitPane.TOP);
        splitTopBottom.add(scrollInfo, JSplitPane.BOTTOM);

        splitLeftRight = new JSplitPane();
        splitLeftRight.setLeftComponent(splitTopBottom);
        splitLeftRight.setRightComponent(panelRight);
        splitLeftRight.add(panelRight, JSplitPane.RIGHT);
        splitLeftRight.add(splitTopBottom, JSplitPane.LEFT);

        panelContent = (JPanel) getContentPane();
        panelContent.setLayout(new BorderLayout());
        panelContent.add(splitLeftRight, BorderLayout.CENTER);
        this.setExtendedState(this.getExtendedState() | JFrame.MAXIMIZED_BOTH);
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        this.setSplitDividersLocations(screenSize);
    }

    public static void setPathToDOT(String path) {
        pathToDOT = path;
    }

    /**
     * Sets split panes dividers locations.
     *
     * @param size <code>Dimension</code> the dimension of the main
     *        window after some changes have been produced.
     */
    private void setSplitDividersLocations(Dimension size) {
        int leftPrefWidth = (int) (size.width * 0.22);
        int leftTopPrefHeight = (int) (size.height * 0.6);

        splitTopBottom.setDividerLocation(leftTopPrefHeight);
        splitLeftRight.setDividerLocation(leftPrefWidth);
    }

    /******************************************************************************
//******************************************************************************
//******************************************************************************
//      GETTERS AND SETTERS
//******************************************************************************
//******************************************************************************
//******************************************************************************/

         /**
          * Returns the about menu.
          *
          * @return <code>JMenu</code> the application's about menu
          */
         public JMenu getHelp() {
             return help;
         }

    /**
     * Returns the context label, which displays tips about functionality
     * options to user.
     *
     * @return <code>JTextField</code> the context label
     */
    public JTextField getLabelContext() {
        return labelContext;
    }

    /**
     * Returns the application's menu bar.
     *
     * @return <code>JMenuBar</code> the application's menu bar
     */
    public JMenuBar getMenu() {
        return menu;
    }

    /**
     * Returns the expression analysis by all time points chart panel.
     *
     * @return <code>PanelExpression</code> the expression analysis by all time points panel
     */
    public PanelExpression getPanelAnalyzingExpressionAllTimePoints() {
        return panelAnalyzingExpressionAllTimePoints;
    }

    /**
     * Returns the expression analysis by bicluster pattern chart panel.
     *
     * @return <code>PanelExpression</code> the expression analysis by bicluster pattern
     */
    public PanelExpression getPanelExpressionBiclusterPattern() {
        return panelAnalyzingExpressionBiclusterPattern;
    }

    /**
     * Returns the expression analysis by bicluster time points chart panel.
     *
     * @return <code>PanelExpression</code> the expression analysis by bicluster time points chart panel
     */
    public PanelExpression getPanelExpressionBiclusterTimePoints() {
        return panelAnalyzingExpressionBiclusterTimePoints;
    }

    /**
     * Returns the function analysis - significance graph panel.
     *
     * @return <code>PanelAnalyzingFunctionSignificanceGraph</code>
     */
    public PanelAnalyzingFunctionSignificanceGraph
            getPanelAnalyzingFunctionSignificanceGraph() {
        return panelAnalyzingFunctionSignificanceGraph;
    }

    /**
     * Returns the function analysis - table panel.
     *
     * @return <code>PanelAnalyzingFunctionTable</code>
     */
    public PanelAnalyzingFunctionTable getPanelAnalyzingFunctionTable() {
        return panelAnalyzingFunctionTable;
    }

    /**
     * Returns the matrix analysis by colors panel.
     *
     * @return <code>PanelAnalyzingMatrixTables</code> the matrix analysis by colors panel
     */
    public PanelAnalyzingMatrixTables getPanelAnalyzingMatrixColors() {
        return panelAnalyzingMatrixColors;
    }

    /**
     * Returns the matrix analysis by symbols panel.
     *
     * @return <code>PanelAnalyzingMatrixTables</code> the matrix analysis by symbols panel
     */
    public PanelAnalyzingMatrixTables getPanelAnalyzingMatrixSymbols() {
        return panelAnalyzingMatrixSymbols;
    }

    /**
     * Returns the matrix analysis by values panel.
     *
     * @return <code>PanelAnalyzingMatrixTables</code> the matrix analysis by values panel
     */
    public PanelAnalyzingMatrixTables getPanelAnalyzingMatrixValues() {
        return panelAnalyzingMatrixValues;
    }

    /**
     * Returns the matrix analysis by dendrogram panel.
     *
     * @return PanelAnalyzingMatrixDendrogram
     */
    public PanelAnalyzingMatrixDendrogram getPanelAnalyzingMatrixDendrogram() {
        return panelAnalyzingMatrixDendrogram;
    }

    /**
     * Returns biclustering panel.
     *
     * @return <code>PanelBiclustering</code>
     */
    public PanelBiclustering getPanelBiclustering() {
        return panelBiclustering;
    }

    /**
     * Returns this frames' content panel.
     *
     * @return this frame's content <code>JPanel</code>
     */
    public JPanel getPanelContent() {
        return panelContent;
    }

    /**
     * Returns the loading panel.
     *
     * @return <code>PanelLoading</code>
     */
    public PanelLoading getPanelLoading() {
        return panelLoading;
    }

    /**
     * Returns preprocessing panel.
     *
     * @return <code>PanelPreProcessing</code>
     */
    public PanelPreProcessing getPanelPreProcessing() {
        return panelPreProcessing;
    }

    /**
     * Returns the panel on the right, which holds the tabs and
     * visualization panels.
     *
     * @return the <code>JPanel</code> on the right
     */
    public JPanel getPanelRight() {
        return panelRight;
    }

    /**
     * Returns the information text pane, which displays informations about the
     * selected datasets, matrices or biclusters in the tree.
     *
     * @return JTextPane the information text pane
     */
    public JTextPane getTextPaneInfo() {
        return textPaneInfo;
    }

    /**
     * Returns the dataset tree object.
     *
     * @return DatasetTree the dataset tree
     */
    public DatasetTree getTree() {
        return tree;
    }

    /**
     * Returns the path to function results' folder/directory.
     *
     * @return the <code>String</code> path
     */
    public String getPathToFunctionResults() {
        return pathsToFolders[0];
    }

    /**
     * Returns the path to gene ontology files folder/directory.
     *
     * @return <code>String</code> path
     */
    public String getPathToGO_Files() {
        return pathsToFolders[1];
    }

    /**
     * Returns the path to stored session files folder/directory.
     *
     * @return <code>String</code> path
     */
    public String getPathToSessions() {
        return pathsToFolders[2];
    }

    /**
     * Returns the path to the directory with dataset files.
     *
     * @return <code>String</code> path
     */
    public String getPathToDatasets() {
        return pathsToFolders[3];
    }

    /**
     * Returns the paths to function results, gene ontology files and sessions.
     *
     * @return the <code>String[]</code> array with paths to function
     *             results, gene ontology files and sessions
     */
    public String[] getPathsToFolders() {
        return pathsToFolders;
    }

    /**
     * Returns the path to dot binary file.
     *
     * @return String the path to dot binary file
     */
    public String getPathToDOT() {
        return this.pathToDOT;
    }

    /**
     * Returns gene ontology files URL.
     *
     * @return String the gene ontology files URL
     */
    public String getGeneOntologyFilesURL() {
        return geneOntologyFilesURL;
    }

    /**
     * Returns gene association files URL.
     *
     * @return String the gene association files URL
     */
    public String getGeneAssociationFilesURL() {
        return geneAssociationFilesURL;
    }

    /**
     * Returns the scroll pane which holds the expression
     * analysis - all time points chart panel.
     *
     * @return the <code>JScrollPane</code> which holds the expression
     * analysis - all time points chart panel
     */
    public JScrollPane getScrollAnalyzingExpressionAllTimePoints() {
        return scrollAnalyzingExpressionAllTimePoints;
    }

    /**
     * Returns the scroll pane which holds the expression
     * analysis - bicluster pattern chart panel.
     *
     * @return the <code>JScrollPane</code> which holds the expression
     * analysis - bicluster pattern chart panel
     */
    public JScrollPane getScrollAnalyzingExpressionBiclusterPattern() {
        return scrollAnalyzingExpressionBiclusterPattern;
    }

    /**
     * Returns the scroll pane which holds the expression
     * analysis - bicluster time points chart panel.
     *
     * @return the <code>JScrollPane</code> which holds the expression
     * analysis - bicluster time points chart panel
     */
    public JScrollPane getScrollAnalyzingExpressionBiclusterTimePoints() {
        return scrollAnalyzingExpressionBiclusterTimePoints;
    }

    /**
     * Returns the scroll pane which holds the function
     * analysis - significance graph panel.
     *
     * @return the <code>JScrollPane</code> which holds the function
     * analysis - significance graph panel
     */
    public JScrollPane getScrollAnalyzingFunctionSignificanceGraph() {
        return scrollAnalyzingFunctionSignificanceGraph;
    }

    /**
     * Returns the scroll pane which holds the function
     * analysis - table of functions panel.
     *
     * @return the <code>JScrollPane</code> which holds the function
     * analysis - table of functions panel
     */
    public JScrollPane getScrollAnalyzingFunctionTable() {
        return scrollAnalyzingFunctionTable;
    }

    /**
     * Returns the scroll pane which holds the matrix analysis by colors panel.
     *
     * @return the <code>JScrollPane</code> which holds the matrix analysis by colors panel
     */
    public JScrollPane getScrollAnalyzingMatrixColors() {
        return scrollAnalyzingMatrixColors;
    }

    /**
     * Returns the scroll pane which holds the matrix analysis by symbols panel.
     *
     * @return the <code>JScrollPane</code> which holds the matrix analysis by symbols panel
     */
    public JScrollPane getScrollAnalyzingMatrixSymbols() {
        return scrollAnalyzingMatrixSymbols;
    }

    /**
     * Returns the scroll pane which holds the matrix analysis by values panel.
     *
     * @return the <code>JScrollPane</code> which holds the matrix analysis by values panel
     */
    public JScrollPane getScrollAnalyzingMatrixValues() {
        return scrollAnalyzingMatrixValues;
    }

    /**
     * Returns the scroll pane which scrolls biclustering panel.
     *
     * @return the <code>JScrollPane</code> which scrolls biclustering panel
     */
    public JScrollPane getScrollBiclustering() {
        return scrollBiclustering;
    }

    /**
     * Returns the scroll pane which scrolls information text pane.
     *
     * @return the <code>JScrollPane</code> which scrolls information text pane
     */
    public JScrollPane getScrollInfo() {
        return scrollInfo;
    }

    /**
     * Returns the scroll pane which scrolls loading panel.
     *
     * @return the <code>JScrollPane</code> which scrolls loading panel
     */
    public JScrollPane getScrollLoading() {
        return scrollLoading;
    }

    /**
     * Returns the scroll pane which scrolls preprocessing panel.
     *
     * @return the <code>JScrollPane</code> which scrolls preprocessing panel
     */
    public JScrollPane getScrollPreProcessing() {
        return scrollPreProcessing;
    }

    /**
     * Returns the scroll pane which scrolls the dataset tree.
     *
     * @return the <code>JScrollPane</code> which scrolls the dataset tree
     */
    public JScrollPane getScrollTree() {
        return scrollTree;
    }

    /**
     * Returns the left-right split pane.
     *
     * @return the left-right <code>JSplitPane</code>
     */
    public JSplitPane getSplitLeftRight() {
        return splitLeftRight;
    }

    /**
     * Returns the top-bottom split pane.
     *
     * @return the top-bottom <code>JSplitPane</code>
     */
    public JSplitPane getSplitTopBottom() {
        return splitTopBottom;
    }

    /**
     * Returns the main tabbed pane.
     *
     * @return the main <code>JTabbedPane</code>
     */
    public JTabbedPane getTabbedPane() {
        return tabbedPane;
    }

    /**
     * Returns the analysis tabbed pane.
     *
     * @return the analysis <code>JTabbedPane</code>
     */
    public JTabbedPane getTabbedPaneAnalyzing() {
        return tabbedPaneAnalyzing;
    }

    /**
     * Returns the expression analysis tabbed pane.
     *
     * @return the expression analysis <code>JTabbedPane</code>
     */
    public JTabbedPane getTabbedPaneAnalyzingExpression() {
        return tabbedPaneAnalyzingExpression;
    }

    /**
     * Returns the function analysis tabbed pane.
     *
     * @return the function analysis <code>JTabbedPane</code>
     */
    public JTabbedPane getTabbedPaneAnalyzingFunction() {
        return tabbedPaneAnalyzingFunction;
    }

    /**
     * Returns the matrix analysis tabbed pane.
     *
     * @return the matrix analysis <code>JTabbedPane</code>
     */
    public JTabbedPane getTabbedPaneAnalyzingMatrix() {
        return tabbedPaneAnalyzingMatrix;
    }

    /**
     * Sets a new dataset tree.
     *
     * @param t the new <code>DatasetTree</code>
     */
    public void setTree(DatasetTree t) {
        tree = t;
        scrollTree.getViewport().setView(tree);
        tree.reload();
        this.setVisible(true);
    }

    public void initTree() {
        tree = new DatasetTree() {
            public void setSelectionPath(TreePath selectionPath) {
                if (selectionPath == null) {
                    return;
                }
                Object node = ((DefaultMutableTreeNode)
                                                 selectionPath.
                                                 getLastPathComponent()).
                                     getUserObject();

                if (node instanceof NodeInfo) {
                    int type = ((NodeInfo)node).getNodeObjectType();
                    if (type == NodeInfo.BICLUSTER_SINGLE ||
                        type == NodeInfo.BICLUSTER_GROUP ||
                        type == NodeInfo.POSTPROCESSED_GROUP ||
                        type == NodeInfo.POSTPROCESSED_SINGLE) {
                        lock();
                        startTaskInLabelContext("Rendering...");
                    }
                }

                Thread applyTreePathSelectionThread =
                        new TreePathSelectionThread(
                        selectionModel, selectionPath);
                applyTreePathSelectionThread.start();
            }                        
        };
        scrollTree.getViewport().setView(tree);
        tree.reload();
        this.setVisible(true);        
    }

    /**
     * Enables/disables tabs on the main tabbed pane.<br/>
     *
     * @param loading         <code>boolean</code> indicating if tab
     *                        loading is to be enabled (true)
     *                        /disabled (false)
     * @param preprocessing   <code>boolean</code> indicating if tab
     *                        preprocessing is to be enabled (true)
     *                        /disabled (false)
     * @param biclustering    <code>boolean</code> indicating if tab
     *                        biclustering is to be enabled (true)
     *                        /disabled (false)
     * @param analyzing       <code>boolean</code> indicating if tab
     *                        analyzing is to be enabled (true)
     *                        /disabled (false)
     * @param postprocessing  <code>boolean</code> indicating if tab
     *                        post-processing is to be enabled (true)
     *                        /disabled (false)
     * @param comparing       <code>boolean</code> indicating if tab
     *                        comparing is to be enabled (true)
     *                        /disabled (false)
     */
    public void setTabbedPaneEnabledTabs(boolean loading,
                                         boolean preprocessing,
                                         boolean biclustering,
                                         boolean analyzing,
                                         boolean postprocessing,
                                         boolean comparing) {
        tabbedPane.setEnabledAt(0, loading); // loading tab
        tabbedPane.setEnabledAt(1, preprocessing); // preprocessing tab
        tabbedPane.setEnabledAt(2, biclustering); // biclustering tab
        tabbedPane.setEnabledAt(3, analyzing); // analyzing tab
        tabbedPane.setEnabledAt(4, postprocessing); // post-processing tab
//        tabbedPane.setEnabledAt(5, comparing);        // comparing tab
    }

    /**
     * Enables/disables tabs on the analyzing tabbed pane.<br/>
     *
     * @param matrix          <code>boolean</code> indicating if tab
     *                        matrix is to be enabled (true)
     *                        /disabled (false)
     * @param expression      <code>boolean</code> indicating if tab
     *                        expression is to be enabled (true)
     *                        /disabled (false)
     * @param function        <code>boolean</code> indicating if tab
     *                        function is to be enabled (true)
     *                        /disabled (false)
     */
    public void setTabbedPaneAnalyzingEnabledTabs(boolean matrix,
                                                  boolean expression,
                                                  boolean function) {
        tabbedPaneAnalyzing.setEnabledAt(0, matrix); // matrix tab
        tabbedPaneAnalyzing.setEnabledAt(1, expression); // expression tab
        tabbedPaneAnalyzing.setEnabledAt(2, function); // function tab
    }

    /**
     * Enables/disables tabs on the matrix analysis tabbed pane.
     *
     * @param values          <code>boolean</code> indicating if tab
     *                        values is to be enabled (true)
     *                        /disabled (false)
     * @param colors          <code>boolean</code> indicating if tab
     *                        colors is to be enabled (true)
     *                        /disabled (false)
     * @param symbols         <code>boolean</code> indicating if tab
     *                        symbols is to be enabled (true)
     *                        /disabled (false)
     * @param dendrogram      <code>boolean</code> indicating if tab
     *                        dendrogram is to be enabled (true)
     *                        /disabled (false)
     */
    public void setTabbedPaneAnalyzingMatrixEnabledTabs(boolean values,
            boolean colors,
            boolean symbols,
            boolean dendrogram) {
        tabbedPaneAnalyzingMatrix.setEnabledAt(0, values);
        tabbedPaneAnalyzingMatrix.setEnabledAt(1, colors);
        tabbedPaneAnalyzingMatrix.setEnabledAt(2, symbols);
        tabbedPaneAnalyzingMatrix.setEnabledAt(3, dendrogram);
    }

    /**
     * Clears all application's panels, so that they contain no information.
     */
    public void clearAllPanels() {
        panelLoading.clear();
        panelPostProcessing.clear();
//        panelComparing = new JPanel();
        panelAnalyzingExpressionBiclusterTimePoints.clear();
        panelAnalyzingExpressionAllTimePoints.clear();
        panelAnalyzingExpressionBiclusterPattern.clear();
        panelAnalyzingFunctionTable.clear();
        panelAnalyzingFunctionSignificanceGraph.clear();
        panelAnalyzingMatrixValues.clear();
        panelAnalyzingMatrixColors.clear();
        panelAnalyzingMatrixSymbols.clear();
        panelAnalyzingMatrixDendrogram.clear();
        panelBiclustering.clear();
    }

    private void selectedRootOrDatasetNode() {
        /**
         * we have paths such as:
         * 1. root
         * 2. root -> dataset X
         *
         * available options:
         *      LOADING
         */
        this.setTabbedPaneEnabledTabs(true, // loading tab
                                      false, // preprocessing tab
                                      false, // biclustering tab
                                      false, // analyzing tab
                                      false, // post-processing tab
                                      false); // comparing tab

        tabbedPane.getModel().setSelectedIndex(0); // loading tab
        labelContext.setText("Choose a dataset to load.");
        textPaneInfo.setText("\n\n\n\nBiGGEsTS Information Box");
    }

    private void selectedOriginalExpressionMatrixNode(TreePath selectionPath) {
        /**
         * we have paths such as:
         * 1. root -> dataset X -> original
         *       available options:
         *          LOADING
         *          PREPROCESSING
         *             FILTER (mandatory if the original matrix has
         *                     missing values;
         *                     inaccessible option if the original
         *                     matrix has no missing values)
         *             NORMALIZE
         *             DISCRETIZE
         *             SMOOTH
         *          BICLUSTERING
         *             CC_TSB ALGORITHM (available only if the matrix
         *                               has no missing values)
         *          ANALYZING
         *             MATRIX (VALUES; COLORS)
         *             FUNCTION (FUNCTION TABLE)
         *
         * 2. root -> dataset X -> preprocessed
         * 3. root -> dataset X -> discretized
         * 4. root -> dataset X -> biclusters
         * 5. root -> dataset X -> post-processed
         *       available options:
         *          LOADING
         */
        Object nodeObjectInfo = ((DefaultMutableTreeNode)
                                 selectionPath.getLastPathComponent()).
                                getUserObject();
        if (nodeObjectInfo instanceof String) {
            /**
             * case of paths 2, 3, 4 and 5
             */
            this.setTabbedPaneEnabledTabs(true, // loading tab
                                          false, // preprocessing tab
                                          false, // biclustering tab
                                          false, // analyzing tab
                                          false, // post-processing tab
                                          false); // comparing tab
            tabbedPane.getModel().setSelectedIndex(0);
            labelContext.setText("Choose a dataset to load.");
            textPaneInfo.setText("\n\n\n\nBiGGEsTS Information Box");
        } else {
            NodeInfo nodeInfo = (NodeInfo) nodeObjectInfo;
            ExpressionMatrix expressionMatrix = (ExpressionMatrix)
                                                nodeInfo.getNodeObject();
            boolean hasMissings = expressionMatrix.hasMissingValues();
            /**
             * case of path 1 - node object is an object that contains an
             * original expression matrix.
             */
            this.setTabbedPaneEnabledTabs(true, // loading tab
                                          true, // preprocessing tab
                                          !hasMissings, // biclustering tab
                                          true, // analyzing tab
                                          false, // post-processing tab
                                          false); // comparing tab
            tabbedPane.getModel().setSelectedIndex(3);
            labelContext.setText("Choose a functionality on the tabs.");
            try {
                textPaneInfo.setStyledDocument(nodeInfo.getInfo());
            } catch (BadLocationException ex) {
                ex.printStackTrace();
            }

            this.panelPreProcessing.paint(expressionMatrix,
                                          selectionPath);

            try {
                expressionMatrix.getGoTermsComputed();
                this.panelAnalyzingFunctionTable.paint(
                        expressionMatrix,
                        selectionPath);
            } catch (Exception ex) {
                ex.printStackTrace();
                this.tabbedPaneAnalyzing.setEnabledAt(2, false);
            }

            if (hasMissings) {
                this.tabbedPaneAnalyzing.setEnabledAt(0, true);
                this.tabbedPaneAnalyzing.setEnabledAt(1, false);

                tabbedPaneAnalyzing.getModel().setSelectedIndex(0);

                this.setTabbedPaneAnalyzingMatrixEnabledTabs(true, //values
                        true, //colors
                        false, //symbols
                        true); //dendrogram

                tabbedPaneAnalyzingMatrix.getModel().setSelectedIndex(1);
                panelAnalyzingMatrixValues.paint(nodeInfo,
                                                 selectionPath);
                panelAnalyzingMatrixColors.paint(nodeInfo,
                                                 selectionPath);
                panelAnalyzingMatrixDendrogram.paint((IMatrix)
                                                     nodeInfo.getNodeObject());
            } else {
                /**
                 * the expression matrix of the node object doesn't have missing values
                 */
                this.panelBiclustering.paint(expressionMatrix,
                                             selectionPath);
                this.tabbedPaneAnalyzing.setEnabledAt(0, true);
                this.tabbedPaneAnalyzing.setEnabledAt(1, false);

                tabbedPaneAnalyzing.getModel().setSelectedIndex(0);

                this.setTabbedPaneAnalyzingMatrixEnabledTabs(true, //values tab
                        true, //colors tab
                        false, //symbols tab
                        true); //dendrogram tab
                tabbedPaneAnalyzingMatrix.getModel().setSelectedIndex(1);

                /**
                 * initialize analyse matrix by values panel
                 */
                panelAnalyzingMatrixValues.paint(nodeInfo,
                                                 selectionPath);

                /**
                 * initialize analyse matrix by colors panel
                 */
                panelAnalyzingMatrixColors.paint(nodeInfo,
                                                 selectionPath);
            }
        }
    }

    private void selectedPreprocessedDiscretizedOrGroupsOfBiclusters(
            NodeInfo nodeInfo, TreePath selectionPath) {
        /**
         * we have paths such as:
         * 1. root -> dataset X -> preprocessed -> preprocessed Y
         * available options:
         *      LOADING
         *      PREPROCESSING
         *      BICLUSTERING
         *         CC TSB BICLUSTERING
         *      ANALYZING
         *         MATRIX (VALUES; COLORS)
         *
         * 2. root -> dataset X -> discretized -> discretized Y
         * available options:
         *      LOADING
         *      BICLUSTERING
         *         CCC BICLUSTERING
         *         E CCC BICLUSTERING
         *         q CLUSTERING
         *      ANALYZING
         *         MATRIX (VALUES; COLORS; SYMBOLS)
         *
         * 3. root -> dataset X -> biclusters -> bicluster group Y
         * available options:
         *     LOADING
         *     ANALYZING
         *         MATRIX (VALUES; COLORS; SYMBOLS - if it has discrete values)
         *         EXPRESSION (BICLUSTER TIME POINTS; PATTERN - if it has discrete values)
         *         FUNCTION (TABLE)
         *     POST-PROCESSING
         *     COMPARING
         *
         * 4. root -> dataset X -> post-processed -> post-processed group Y
         *     LOADING
         *     ANALYZING
         *         MATRIX (VALUES; COLORS; SYMBOLS - if it has discrete values)
         *         EXPRESSION (BICLUSTER TIME POINTS; PATTERN - if it has discrete values)
         *         FUNCTION (TABLE)
         *     POST-PROCESSING
         *     COMPARING
         */
        int nodeObjectType = nodeInfo.getNodeObjectType();

        try {
            textPaneInfo.setStyledDocument(nodeInfo.getInfo());
        } catch (BadLocationException ex) {
            ex.printStackTrace();
        }

        switch (nodeObjectType) {
        case NodeInfo.PREPROCESSED:
            labelContext.setText(
                    "Visualize preprocess matrix tables (by values or colors).");
            PreProcessedExpressionMatrix preprocessed = (
                    PreProcessedExpressionMatrix) nodeInfo.
                    getNodeObject();
            boolean hasMissings = preprocessed.hasMissingValues();
            this.setTabbedPaneEnabledTabs(true, // loading tab
                    true, // preprocessing tab
                    !hasMissings, // biclustering tab
                    true, // analyzing tab
                    false, // post-processing tab
                    false); // comparing

            tabbedPane.setSelectedIndex(3); // analyzing tab

            this.panelPreProcessing.paint(preprocessed, selectionPath);

            if (!hasMissings) {
                this.panelBiclustering.paint(preprocessed,
                        selectionPath);
            }

            try {
                preprocessed.getGoTermsComputed();
                this.panelAnalyzingFunctionTable.paint(
            preprocessed,
            selectionPath);
            } catch (Exception ex) {
                ex.printStackTrace();
            }

            this.tabbedPaneAnalyzing.setEnabledAt(0, true);
            this.tabbedPaneAnalyzing.setEnabledAt(1, false);

            tabbedPaneAnalyzing.setSelectedIndex(0); // matrix tab

            this.setTabbedPaneAnalyzingMatrixEnabledTabs(true, // values tab
                    true, // colors tab
                    false, // symbols tab
                    true); // dendrogram tab
            tabbedPaneAnalyzingMatrix.setSelectedIndex(1); // colors tab
            panelAnalyzingMatrixValues.paint(nodeInfo, selectionPath);
            panelAnalyzingMatrixColors.paint(nodeInfo, selectionPath);
            panelAnalyzingMatrixDendrogram.paint((IMatrix) nodeInfo.
                    getNodeObject());

            break;
        case NodeInfo.DISCRETIZED:
            DiscretizedExpressionMatrix discretized = (
                    DiscretizedExpressionMatrix) nodeInfo.
                    getNodeObject();
            labelContext.setText(
                    "Visualize discretized matrix tables (by values, colors or symbols).");
            this.setTabbedPaneEnabledTabs(true, // loading tab
                    false, // preprocessing tab
                    true, // biclustering tab
                    true, // analyzing tab
                    false, // post-processing
                    false); // comparing

            tabbedPane.getModel().setSelectedIndex(3);

            panelBiclustering.paint(discretized, selectionPath);

            tabbedPaneAnalyzing.getModel().setSelectedIndex(0);

            this.setTabbedPaneAnalyzingMatrixEnabledTabs(true, //values
                    true, //colors
                    true, //symbols
                    true); // dendrogram

            tabbedPaneAnalyzingMatrix.getModel().setSelectedIndex(2);

            /**
             * initialize analyzing matrix panels
             */
            this.panelAnalyzingMatrixValues.paint(nodeInfo,
                    selectionPath);
            this.panelAnalyzingMatrixColors.paint(nodeInfo,
                    selectionPath);
            this.panelAnalyzingMatrixSymbols.paint(nodeInfo,
                    selectionPath);
            panelAnalyzingMatrixDendrogram.paint((IMatrix) nodeInfo.
                    getNodeObject());

            try {
                discretized.getGoTermsComputed();
                    this.panelAnalyzingFunctionTable.paint(
                            discretized,
                            selectionPath);
            } catch (Exception ex) {
                ex.printStackTrace();
            }

            this.tabbedPaneAnalyzing.setEnabledAt(0, true);
            this.tabbedPaneAnalyzing.setEnabledAt(1, false);
            break;
        case NodeInfo.BICLUSTER_GROUP:
            labelContext.setText(
                    "Visualize biclusters matrices, expression charts and function analysis.");
            this.setTabbedPaneEnabledTabs(true, // loading tab
                    false, // preprocessing tab
                    false, // biclustering tab
                    true, // analyzing tab
                    true, // post-processing tab
                    true); // comparing tab

            tabbedPane.getModel().setSelectedIndex(3);

            this.setTabbedPaneAnalyzingEnabledTabs(true, // matrix tab
                    true, // expression tab
                    true); // function tab

            tabbedPaneAnalyzing.getModel().setSelectedIndex(0);

            /**
             * initialize function analysis panels (TABLE)
             */
            this.tabbedPaneAnalyzingFunction.setEnabledAt(0, true);
            this.tabbedPaneAnalyzingFunction.setEnabledAt(1, false);
            this.tabbedPaneAnalyzingFunction.getModel().setSelectedIndex(0);
            this.panelAnalyzingFunctionTable.paint((Biclustering)
                    nodeInfo.getNodeObject(), selectionPath, null);

            /**
             * initialize matrix analysis panels (VALUES; COLORS; SYMBOLS)
             * and expression analysis panels (BICLUSTER TIME POINTS; ALL TIME POINTS; PATTERN)
             */
            tabbedPaneAnalyzingMatrix.setEnabledAt(0, true); // values tab
            tabbedPaneAnalyzingMatrix.setEnabledAt(1, true); // colors tab
            tabbedPaneAnalyzingMatrix.setEnabledAt(3, false); // dendrogram tab
            tabbedPaneAnalyzingMatrix.getModel().setSelectedIndex(1);
            panelAnalyzingMatrixValues.paint(nodeInfo, selectionPath);
            panelAnalyzingMatrixColors.paint(nodeInfo, selectionPath);
            tabbedPaneAnalyzingExpression.setEnabledAt(0, true); // bicluster time points tab
            tabbedPaneAnalyzingExpression.setEnabledAt(1, false); // all time points tab
            panelAnalyzingExpressionBiclusterTimePoints.
                    paintBiclusterTimePointsCharts(nodeInfo);

            if (nodeInfo.getDataType() == NodeInfo.VALUES) {
                tabbedPaneAnalyzingMatrix.setEnabledAt(2, false); // symbols tab
                tabbedPaneAnalyzingExpression.setEnabledAt(2, false); // pattern tab
            } else {
                tabbedPaneAnalyzingMatrix.setEnabledAt(2, true); // symbols tab
                tabbedPaneAnalyzingExpression.setEnabledAt(2, true); // pattern tab
                panelAnalyzingMatrixSymbols.paint(nodeInfo,
                        selectionPath);
                panelAnalyzingExpressionBiclusterPattern.
                        paintBiclusterPatternCharts(nodeInfo);
            }

            /**
             * initialize post-processing panel
             */
            this.panelPostProcessing.paint((Biclustering) nodeInfo.
                    getNodeObject(), selectionPath);
            unlock();
            this.finishTaskInLabelContext("Analyze group of biclusters.");
            break;
        case NodeInfo.POSTPROCESSED_GROUP:
            labelContext.setText(
                    "Visualize biclusters matrices, expression charts and function analysis.");
            this.setTabbedPaneEnabledTabs(true, // loading tab
                    false, // preprocessing tab
                    false, // biclustering tab
                    true, // analyzing tab
                    true, // post-processing tab
                    true); // comparing tab

            tabbedPane.getModel().setSelectedIndex(3);

            this.setTabbedPaneAnalyzingEnabledTabs(true, // matrix tab
                    true, // expression tab
                    true); // function tab

            tabbedPaneAnalyzing.getModel().setSelectedIndex(0);

            /**
             * initialize function analysis panels (TABLE)
             */
            this.tabbedPaneAnalyzingFunction.setEnabledAt(0, true);
            this.tabbedPaneAnalyzingFunction.setEnabledAt(1, false);
            this.tabbedPaneAnalyzingFunction.getModel().setSelectedIndex(0);
            this.panelAnalyzingFunctionTable.paint((Biclustering)
                    nodeInfo.getNodeObject(), selectionPath, null);

            /**
             * initialize matrix analysis panels (VALUES; COLORS; SYMBOLS)
             * and expression analysis panels (BICLUSTER TIME POINTS; ALL TIME POINTS; PATTERN)
             */
            tabbedPaneAnalyzingMatrix.setEnabledAt(0, true); // values tab
            tabbedPaneAnalyzingMatrix.setEnabledAt(1, true); // colors tab
            tabbedPaneAnalyzingMatrix.setEnabledAt(3, false); // dendrogram tab
            tabbedPaneAnalyzingMatrix.getModel().setSelectedIndex(1);
            panelAnalyzingMatrixValues.paint(nodeInfo, selectionPath);
            panelAnalyzingMatrixColors.paint(nodeInfo, selectionPath);

            tabbedPaneAnalyzingExpression.setEnabledAt(0, true); // bicluster time points tab
            tabbedPaneAnalyzingExpression.setEnabledAt(1, false); // all time points tab
            panelAnalyzingExpressionBiclusterTimePoints.
                    paintBiclusterTimePointsCharts(nodeInfo);

            if (nodeInfo.getDataType() == NodeInfo.VALUES) {
                tabbedPaneAnalyzingMatrix.setEnabledAt(2, false); // symbols tab
                tabbedPaneAnalyzingExpression.setEnabledAt(2, false); // pattern tab
            } else {
                tabbedPaneAnalyzingMatrix.setEnabledAt(2, true); // symbols tab
                tabbedPaneAnalyzingExpression.setEnabledAt(2, true); // pattern tab

                panelAnalyzingMatrixSymbols.paint(nodeInfo,
                        selectionPath);
                panelAnalyzingExpressionBiclusterPattern.
                        paintBiclusterPatternCharts(nodeInfo);
            }

            this.panelPostProcessing.paint((Biclustering) nodeInfo.
                    getNodeObject(), selectionPath);
            unlock();
            this.finishTaskInLabelContext("Analyze group of biclusters.");
            break;
        }
    }

    private void selectedBicluster(NodeInfo nodeInfo, TreePath selectionPath) {
        /**
         * we have paths such as:
         * 1. root -> dataset X -> biclusters -> bicluster Y -> single bicluster Z
         * 2. root -> dataset X -> post-processed -> post-processed Y -> single post-processed Z
         *    available options:
         *       LOADING
         *       ANALYZING
         *          MATRIX (VALUES; COLORS; SYMBOLS - if it has discrete values)
         *          EXPRESSION (BICLUSTER TIME POINTS; ALL TIME POINTS; PATTERN - if it has discrete values)
         *          FUNCTION (TABLE; SIGNIFICANCE GRAPH - if it has significant biological functions)
         */
        try {
            textPaneInfo.setStyledDocument(nodeInfo.getInfo());
        } catch (BadLocationException ex) {
            ex.printStackTrace();
        }

        this.setTabbedPaneEnabledTabs(true, // loading tab
                                      false, // preprocessing tab
                                      false, // biclustering tab
                                      true, // analyzing tab
                                      false, // post-processing tab
                                      false); // comparing tab
        tabbedPane.getModel().setSelectedIndex(3);

        this.setTabbedPaneAnalyzingEnabledTabs(true, // matrix tab
                true, // expression tab
                true); // function tab
        tabbedPaneAnalyzing.getModel().setSelectedIndex(0);

        /**
         * initialize expression analysis panels (BICLUSTER TIME POINTS; ALL TIME POINTS; PATTERN)
         */
        tabbedPaneAnalyzingExpression.setEnabledAt(0, true); // bicluster time points tab
        tabbedPaneAnalyzingExpression.setEnabledAt(1, true); // all time points tab

        if (nodeInfo.getDataType() == NodeInfo.VALUES) {
            tabbedPaneAnalyzingExpression.setEnabledAt(2, false); // bicluster pattern tab
        } else {
            tabbedPaneAnalyzingExpression.setEnabledAt(2, true); // analyzing expression tab
            panelAnalyzingExpressionBiclusterPattern.
                    paintBiclusterPatternCharts(nodeInfo); // bicluster pattern tab
        }
        tabbedPaneAnalyzingExpression.getModel().setSelectedIndex(0);
        panelAnalyzingExpressionBiclusterTimePoints.
                paintBiclusterTimePointsCharts(nodeInfo); // bicluster time points tab
        panelAnalyzingExpressionAllTimePoints.paintAllTimePointsChart(
                nodeInfo,
                (NodeInfo) ((DefaultMutableTreeNode) selectionPath.
                            getPath()[3]).getUserObject()); // all time points tab

        Bicluster bicluster = (Bicluster)nodeInfo.getNodeObject();

        /**
         * initialize function panels (TABLE; SIGNIFICANCE GRAPH)
         */
        GOFrontEnd go = panelAnalyzingFunctionTable.getGOFrontEndForSelectionPath(selectionPath);
        GeneSetResult goProperties = this.panelAnalyzingFunctionTable.paint(
            bicluster, null, (float)0.01, selectionPath);

        if (go != null && goProperties != null) {
            if (bicluster.
                computeSignificant_GO_BonferroniCorrected_pValues_UserThreshold(
                goProperties, (float)0.01) > 0) {
                    Object[] path = selectionPath.getPath();
                    String pathToBiclusteringFunctionResults =
                            getPathToFunctionResults() +
                            File.separator + path[1].toString() +
                            File.separator +
                            ((DefaultMutableTreeNode) path[path.
                             length - 2]).getUserObject();

                    File directoryBiclusteringFunctionResults = new File(
                            pathToBiclusteringFunctionResults);
                    if (!directoryBiclusteringFunctionResults.exists()) {
                        directoryBiclusteringFunctionResults.mkdirs();
                    }

                    try {
                        bicluster.printGraphWithGeneOntologyTerms_ToFile(
                                go,
                                directoryBiclusteringFunctionResults.getAbsolutePath(),
                                "graph-BICLUSTER_" + bicluster.getID() + ".dot", (float)0.01);
                        panelAnalyzingFunctionSignificanceGraph.paint(
                                bicluster, selectionPath);
                        tabbedPaneAnalyzingFunction.setEnabledAt(1, true);
                    }
                    catch (Exception ex) {
                        ex.printStackTrace();
                        tabbedPaneAnalyzingFunction.setEnabledAt(1, false);
                    }
                }
        }


        /**
         * initialize matrix analysis panels (VALUES; COLORS; SYMBOLS)
         */
        tabbedPaneAnalyzingMatrix.setEnabledAt(0, true); // values tab
        panelAnalyzingMatrixValues.paint(nodeInfo, selectionPath); // values tab
        tabbedPaneAnalyzingMatrix.setEnabledAt(1, true); // colors tab
        tabbedPaneAnalyzingMatrix.setSelectedIndex(1); // colors tab
        tabbedPaneAnalyzingMatrix.setEnabledAt(3, false); // dendrogram tab
        panelAnalyzingMatrixColors.paint(nodeInfo, selectionPath); // colors tab

        if (nodeInfo.getDataType() == NodeInfo.VALUES) {
            tabbedPaneAnalyzingMatrix.setEnabledAt(2, false); // symbols tab
        } else {
            tabbedPaneAnalyzingMatrix.setEnabledAt(2, true); // symbols tab
            panelAnalyzingMatrixSymbols.paint(nodeInfo, selectionPath); // symbols tab
        }
        unlock();
        this.finishTaskInLabelContext("Analyze bicluster.");
    }

    /*****************************************************************************
    /*****************************************************************************
    /*****************************************************************************
    /      EVENT HANDLERS
    /*****************************************************************************
    /*****************************************************************************
    /*****************************************************************************/

    /**
     * Close operation. Exits the application. Asks the user if he/she wants
     * to save current session.
     *
     * @param saveWithCancelOption <code>boolean</code> <code>true</code>
     * if save dialog should have a cancel option, <code>false</code> if only
     * yes and no options should be displayed
     * @throws Exception
     */
    public void closeOperation(boolean saveWithCancelOption) {
        boolean succeeded;
        try {
            File dend = new File(pathToNumDend);
            PrintWriter pw = new PrintWriter(dend);
            pw.print(PanelAnalyzingMatrixDendrogram.getNumberOfDendrograms());
            pw.close();

            succeeded = this.saveCurrentSession(saveWithCancelOption);
            panelAnalyzingMatrixDendrogram.endTreeViewApp();
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(this,
                                          "Unable to save session. " +
                                          ex.getMessage(),
                                          "Unable to save session",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        }

        if (!succeeded) {
            return;
        }

        // delete the contents of Results folder
        BiggestsIO.deleteFolderContent(this.getPathToFunctionResults());

        //this.dispose();
        System.exit(0);
    }

    /**
     * Saves current session to a zip archive file.
     *
     * @param cancelOption boolean
     * @return boolean <code>true</code> if operation succeeded,
     * <code>false</code> if operation did not succeed or if user
     * canceled
     * @throws Exception
     */
    public boolean saveCurrentSession(boolean cancelOption) throws Exception {
        int option = -1;
        if (cancelOption) {
            option = JOptionPane.showOptionDialog(this,
                                                  "Save current session?",
                                                  "Save session?",
                                                  JOptionPane.YES_NO_CANCEL_OPTION,
                                                  JOptionPane.QUESTION_MESSAGE, null, null,
                                                  JOptionPane.YES_OPTION);

            if (option == JOptionPane.CANCEL_OPTION) {
                return false;
            }
        }
        else {
            option = JOptionPane.showOptionDialog(this,
                                                  "Save current session?",
                                                  "Save session?",
                                                  JOptionPane.YES_NO_OPTION,
                                                  JOptionPane.QUESTION_MESSAGE, null, null,
                                                  JOptionPane.YES_OPTION);
        }
        if (option == JOptionPane.NO_OPTION) {
            return true;
        }

        JFileChooser pathFileChooser = new JFileChooser(this.getPathToSessions());

        String[] invalidPaths = new String[2];
        invalidPaths[0] = this.getPathToFunctionResults();
        invalidPaths[1] = this.getPathToGO_Files();

        /**
         * initialize the file chooser with the intended file filter
         * and invalid paths
         */
        pathFileChooser.addChoosableFileFilter(new FileFilterDirectory(
                invalidPaths, ".zip"));

        /**
         * turn off accept all file filter
         */
        pathFileChooser.setAcceptAllFileFilterUsed(false);

        /**
         * show save dialog to the user
         */
        option = pathFileChooser.showSaveDialog(this);

        if (option != JFileChooser.APPROVE_OPTION) {
            return false;
        }

        /**
         * user hit approve option; get the specified path for the file
         */
        String newPath = pathFileChooser.getSelectedFile().getPath();

        File newFile = new File(newPath);
        if (newFile.exists()) {
            if (newFile.isFile()) {
                JOptionPane.showMessageDialog(this, "Unable to save session." +
                                              "\nFilename is not correct.",
                                              "Filename missing",
                                              JOptionPane.ERROR_MESSAGE);
                return false;
            }
            if (newFile.isDirectory()) {
                int opt = JOptionPane.showOptionDialog(this,
                        "The specified file already exists." +
                        "\nOverwrite?",
                        "Overwrite?",
                        JOptionPane.YES_NO_OPTION,
                        JOptionPane.QUESTION_MESSAGE,
                        null, null, JOptionPane.NO_OPTION);
                if (opt == JOptionPane.NO_OPTION) {
                    /**
                     * the user does not want to overwrite the file
                     */
                    return false;
                }
            }
        }

        /**
         * if the specified file path does not end with the valid
         * extension for the selected file filter, then add that
         * extension to the filename
         */
        String extension = ((FileFilterDirectory) pathFileChooser.getFileFilter()).
                           getValidFileExtension();
        if (!newPath.toLowerCase().endsWith(extension)) {
            newPath = newPath + extension;
        }

        newFile = new File(newPath);
        if (newFile.exists() && newFile.isFile()) {
            int opt = JOptionPane.showOptionDialog(this,
                    "The specified file already exists." +
                    "\nOverwrite?",
                    "Overwrite?",
                    JOptionPane.YES_NO_OPTION,
                    JOptionPane.QUESTION_MESSAGE,
                    null, null, JOptionPane.NO_OPTION);
            if (opt == JOptionPane.NO_OPTION) {
                return false;
            }
        }

        BiggestsIO.saveSession2(tree,
                               this.getPathToFunctionResults(),
                               newPath);
        return true;
    }

    /**
     * Handles the event fired when the main window of the
     * application is closing.
     *
     * @param e <code>WindowEvent</code>
     */
    public void this_windowClosing(WindowEvent e) {
        this.closeOperation(true);
    }

    /**
     * Handles the event fired when the user clicks the menu item Session ->
     * Load Dataset. Enables Loading tab.
     *
     * @param actionEvent <code>ActionEvent</code>
     */
    public void sessionLoadDataset_actionPerformed(ActionEvent actionEvent) {
        this.tabbedPane.setSelectedIndex(0);
    }

    /**
     * Handles the event fired when the user clicks the menu item Session ->
     * Load Session. Allows the user to load a session. Displays a load
     * session dialog.
     *
     * @param actionEvent <code>ActionEvent</code>
     */
    public void sessionLoadSession_actionPerformed(ActionEvent actionEvent) {
        try {
            boolean succeeded = this.saveCurrentSession(true);
            if (!succeeded) {
                return;
            }
            panelAnalyzingMatrixDendrogram.endTreeViewApp();
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(this,
                                          "Unable to save session. " +
                                          ex.getMessage(),
                                          "Unable to save session",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        }

        JFileChooser datasetFileChooser = new JFileChooser(this.
                getPathToSessions());
        try {
            datasetFileChooser.setFileFilter(new FileFilterExtension(".zip"));
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        datasetFileChooser.setAcceptAllFileFilterUsed(false);
        int option = datasetFileChooser.showOpenDialog(this);

        if (option == JFileChooser.APPROVE_OPTION) {
            try {
                BiggestsIO.restoreSession2(this,
                                          datasetFileChooser.
                                          getSelectedFile().getPath());
            } catch (Exception ex) {
                ex.printStackTrace();
                JOptionPane.showMessageDialog(this, "Unable to load " +
                                              "session. " + ex.getMessage(),
                                              "Unable to load session",
                                              JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }

    /**
     * Handles the event fired when the user clicks the menu item Session ->
     * Save Session. Allows the user to save current session. Displays a save
     * session dialog.
     *
     * @param actionEvent <code>ActionEvent</code>
     */
    public void sessionSaveSession_actionPerformed(ActionEvent actionEvent) {
        try {
            this.saveCurrentSession(true);
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(this,
                                          "Unable to save session. " +
                                          ex.getMessage(),
                                          "Unable to save session",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        }
    }

    /**
     * Handles the event fired when the user clicks Session -> Exit item menu.
     * Exits the application. First asks the user if he/she wants to save
     * current session.
     *
     * @param actionEvent <code>ActionEvent</code>
     */
    public void sessionExit_actionPerformed(ActionEvent actionEvent) {
        this.closeOperation(true);
    }

    /**
     * Handles the event fired when the user clicks Help -> About item menu.
     * Shows about dialog box.
     *
     * @param actionEvent <code>ActionEvent</code>
     */
    public void helpAbout_actionPerformed(ActionEvent actionEvent) {
        JPanel panel = new JPanel();
        panel.setBackground(Color.white);
        panel.setLayout(new BoxLayout(panel, BoxLayout.PAGE_AXIS));

        JTextPane text = new JTextPane();
        text.setText("BiGGEsTS 1.0.3\n\nCredits\n" +
                                      "Joana P. Gonalves, Sara C. Madeira\n" +
                                      "\n\nWebsite:");
        text.setEditable(false);
        JButton button = new JButton("http://kdbio.inesc-id.pt/software/biggests");
        button.setBackground(Color.white);
        button.addActionListener(new OpenURL());

        panel.add(text);
        panel.add(button);
        JOptionPane.showMessageDialog(this, panel, "About BiGGEsTS",
                                      JOptionPane.INFORMATION_MESSAGE);
    }

    public void colorsRedGreen_actionPerformed(ActionEvent actionEvent) {
        if (colorsRedGreen.isSelected()) {
            HeatmapColorScheme.setHeatmapColorScheme(
                    HeatmapColorScheme.RED_GREEN_SCHEME);
            if (panelAnalyzingMatrixSymbols.getNodeInfo().getDataType() == NodeInfo.SYMBOLS) {
                this.panelAnalyzingMatrixSymbols.paintAgain();
                this.panelAnalyzingMatrixSymbols.repaint();
            }
            this.panelAnalyzingMatrixColors.paintAgain();
            this.panelAnalyzingMatrixColors.repaint();
            this.setVisible(true);
        }
    }

    public void colorsYellowBlue_actionPerformed(ActionEvent actionEvent) {
        if (colorsYellowBlue.isSelected()) {
            HeatmapColorScheme.setHeatmapColorScheme(
                    HeatmapColorScheme.YELLOW_BLUE_SCHEME);
            if (panelAnalyzingMatrixSymbols.getNodeInfo().getDataType() == NodeInfo.SYMBOLS) {
                this.panelAnalyzingMatrixSymbols.paintAgain();
                this.panelAnalyzingMatrixSymbols.repaint();
            }
            this.panelAnalyzingMatrixColors.paintAgain();
            this.panelAnalyzingMatrixColors.repaint();
            this.setVisible(true);
        }
    }

    public void actionPerformed(ActionEvent actionEvent) {
        Object source = actionEvent.getSource();
        if (source == sessionLoadDataset) {
            this.sessionLoadDataset_actionPerformed(actionEvent);
            return;
        }
        if (source == sessionLoadSession) {
            this.sessionLoadSession_actionPerformed(actionEvent);
            return;
        }
        if (source == sessionSaveSession) {
            this.sessionSaveSession_actionPerformed(actionEvent);
            return;
        }
        if (source == sessionExit) {
            this.sessionExit_actionPerformed(actionEvent);
            return;
        }
        if (source == helpAbout) {
            this.helpAbout_actionPerformed(actionEvent);
            return;
        }
        if (source == colorsRedGreen) {
            this.colorsRedGreen_actionPerformed(actionEvent);
            return;
        }
        if (source == colorsYellowBlue) {
            this.colorsYellowBlue_actionPerformed(actionEvent);
            return;
        }
    }

    /**
     * Handles the event fired when the state of the main window changes.
     *
     * @param e WindowEvent
     */
    public void this_windowStateChanged(WindowEvent e) {
        Dimension size = e.getWindow().getSize();
        if (e.getNewState() == Frame.NORMAL) {
            int width = 800;
            int height = 600;
            Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
            if (width > screenSize.getWidth()) {
                width = (int) screenSize.getWidth();
            }
            if (height > screenSize.getHeight()) {
                height = (int) screenSize.getHeight();
            }

            size = new Dimension(width, height);
            this.setSize(size);
            this.setLocation((screenSize.width - this.getSize().width) / 2,
                             (screenSize.height - this.getSize().height) / 2);
        }
        if (e.getNewState() == Frame.MAXIMIZED_BOTH) {
            size = Toolkit.getDefaultToolkit().getScreenSize();
        }
        this.setSplitDividersLocations(size);
    }

    /**
     * Handles the event fired when the user resizes the main window.
     *
     * @param e ComponentEvent
     */
    public void this_componentResized(ComponentEvent e) {
        this.setSplitDividersLocations(this.getSize());
    }

// ###########################################################################
// ###########################################################################
//      SPECIAL EVENT HANDLERS FOR DEALING WITH THE USER'S SELECTION ON
//      THE TABS OR THE NODES OF THE TREE THAT INVOLVE LONG RENDERING TASKS:
//      - special subclasses of JTabbedPane and DatasetTree that override
//        the setSeletionIndex and setSelectionPath methods, respectively, for
//        displaying a warning message to the user before the rendering
//        task begins and creating and starting the thread that will perform
//        the actual selection of the tab or the tree node and the rendering
//        of the corresponding panels
//      - threads for updating the tab or tree node selection model
//        and processing the rendering tasks independently
//      - special panels that override the JPanel's paintChildren method
//        for updating the displayed information message when the rendering
//        task finishes
// ###########################################################################
// ###########################################################################

    private static class CustomPanelAnalyzingMatrixTables extends
            PanelAnalyzingMatrixTables {
        String customMessage;

        public CustomPanelAnalyzingMatrixTables(BiggestsFrame frame,
                                                int tableType,
                                                String customMessage) {
            super(frame, tableType);
            this.customMessage= customMessage;
        }

        public void paintChildren(Graphics g) {
            super.paintChildren(g);
        }
    }

    private static class CustomPanelExpression extends PanelExpression {
        String customMessage;

        public CustomPanelExpression(BiggestsFrame frame, int expressionType,
                                     String customMessage) {
            super(frame, expressionType);
            this.customMessage = customMessage;
        }

        public void paintChildren(Graphics g) {
            super.paintChildren(g);
        }
    }

    private static class TabbedPane extends JTabbedPane {
        private BiggestsFrame frame;

        protected TabbedPane(BiggestsFrame frame) {
            super();
            this.frame = frame;
        }

        public void setSelectedIndex(int selectedIndex) {
            if (!locked) {
                String message = frame.getLabelContext().getText();
                if (model.getSelectedIndex() != -1 &&
                    selectedIndex != model.getSelectedIndex()) {
                    switch (selectedIndex) {
                    case 0:
                        message = "Load a dataset.";
                        break;
                    case 1:
                        message = "Preprocess matrix.";
                        break;
                    case 2:
                        message = "Extract biclusters.";
                        break;
                    case 3:
                        frame.startTaskInLabelContext("Rendering...");
                        message = "Analysis tools.";
                        break;
                    case 4:
                        message = "Post-process biclusters.";
                        break;
                    }
                }
                Thread applyIndexSelectionThread =
                        new
                        TabbedPaneIndexSelectionThread(frame,
                                                       model, selectedIndex, message);
                applyIndexSelectionThread.start();
            }
            else {
                model.setSelectedIndex(selectedIndex);
            }
        }
    }

    private static class TabbedPaneAnalyzing extends JTabbedPane {
        private BiggestsFrame frame;

        protected TabbedPaneAnalyzing(BiggestsFrame frame) {
            super();
            this.frame = frame;
        }

        public void setSelectedIndex(int selectedIndex) {
            if (!locked) {
                String message = frame.getLabelContext().getText();
                if (model.getSelectedIndex() != -1 &&
                    selectedIndex != model.getSelectedIndex()) {
                    switch (selectedIndex) {
                    case 0:
                        frame.startTaskInLabelContext("Rendering matrices.");
                        message = "Analyze matrices.";
                        break;
                    case 1:
                        frame.startTaskInLabelContext("Rendering charts.");
                        message = "Analyze charts.";
                        break;
                    case 2:
                        message = "Analyze GO terms and functional enrichment.";
                    }
                }
                Thread applyIndexSelectionThread =
                        new
                        TabbedPaneIndexSelectionThread(frame, model, selectedIndex, message);
                applyIndexSelectionThread.start();
            }
            else {
                model.setSelectedIndex(selectedIndex);
            }
        }
    }

    private static class TabbedPaneAnalyzingMatrix extends JTabbedPane {
        private BiggestsFrame frame;

        protected TabbedPaneAnalyzingMatrix(BiggestsFrame frame) {
            super();
            this.frame = frame;
        }

        public void setSelectedIndex(int selectedIndex) {
            if (!locked) {
                String message = frame.getLabelContext().getText();
                if (model.getSelectedIndex() != -1 &&
                    selectedIndex != model.getSelectedIndex()) {

                    switch (selectedIndex) {
                    case 0:
                        frame.startTaskInLabelContext(
                                "Rendering expression matrices.");
                        message = "Analyze expression matrices.";
                        break;
                    case 1:
                        frame.startTaskInLabelContext(
                                "Rendering expression heatmaps.");
                        message = "Analyze expression heatmaps.";
                        break;
                    case 2:
                        frame.startTaskInLabelContext(
                                "Rendering symbol heatmaps.");
                        message = "Analyze symbol heatmaps.";
                        break;
                    }
                }
                Thread applyIndexSelectionThread =
                        new TabbedPaneIndexSelectionThread(frame, model,
                        selectedIndex, message);
                applyIndexSelectionThread.start();
            }
            else {
                model.setSelectedIndex(selectedIndex);
                unlock();
                frame.finishTaskInLabelContext("Analyze matrix.");
            }
        }
    }

    private static class TabbedPaneAnalyzingExpression extends JTabbedPane {
        private BiggestsFrame frame;

        protected TabbedPaneAnalyzingExpression(BiggestsFrame frame) {
            this.frame = frame;
        }

        public void setSelectedIndex(int selectedIndex) {
            if(!locked) {
                String message = frame.getLabelContext().getText();
                if (model.getSelectedIndex() != -1 &&
                    selectedIndex != model.getSelectedIndex()) {
                    switch (selectedIndex) {
                    case 0:
                        frame.startTaskInLabelContext(
                                "Rendering bicluster expression charts.");
                        message = "Analyze expression charts.";
                        break;
                    case 1:
                        frame.startTaskInLabelContext(
                                "Rendering all time points expression charts.");
                        message = "Analyze all time points expression charts.";
                        break;
                    case 2:
                        frame.startTaskInLabelContext(
                                "Rendering pattern charts.");
                        message = "Analyze pattern charts.";
                        break;
                    }
                }
                Thread applyIndexSelectionThread =
                        new TabbedPaneIndexSelectionThread(frame,
                        model, selectedIndex, message);
                applyIndexSelectionThread.start();
            }
            else {
                model.setSelectedIndex(selectedIndex);
            }
        }
    }

    private static class TabbedPaneIndexSelectionThread extends Thread {
        private SingleSelectionModel model;
        private int selectedIndex;
        private BiggestsFrame frame;
        private String finishMessage;

        public TabbedPaneIndexSelectionThread(BiggestsFrame frame, SingleSelectionModel model,
                                              int selectedIndex, String message) {
            this.model = model;
            this.selectedIndex = selectedIndex;
            this.frame = frame;
            this.finishMessage = message;
        }

        public void run () {
            model.setSelectedIndex(selectedIndex);
            frame.finishTaskInLabelContext(finishMessage);
        }
    }

    private class TreePathSelectionThread extends Thread {
        private TreeSelectionModel model;
        private TreePath selectionPath;

        public TreePathSelectionThread(TreeSelectionModel model,
                                       TreePath selectionPath) {
            this.model = model;
            this.selectionPath = selectionPath;
        }

        public void run() {
            Object[] path = {tree.getRoot()};
            try {
                path = selectionPath.getPath();
            } catch (NullPointerException ex) {
                model.setSelectionPath(new TreePath(tree.getRoot()));
                return;
            }

            model.setSelectionPath(new TreePath(path));
            tree.repaint();

            clearAllPanels();

            if (path.length < 3) {
                selectedRootOrDatasetNode();
            }
            else
            if (path.length == 3) {
                selectedOriginalExpressionMatrixNode(selectionPath);
            } else {
                NodeInfo nodeInfo = (NodeInfo) ((DefaultMutableTreeNode)
                                                selectionPath.
                                                getLastPathComponent()).
                                    getUserObject();
                if (path.length == 4) {
                    selectedPreprocessedDiscretizedOrGroupsOfBiclusters(
                                    nodeInfo,
                                    selectionPath);
                }
                else if (path.length == 5) {
                    selectedBicluster(nodeInfo, selectionPath);
                }
             }
             setVisible(true);
        }
    }

    protected void startTaskInLabelContext(String message) {
        labelContext.setText(message);
        HeatmapColorScheme.setBackAndForegroundHighlightColors(
            labelContext);
        labelContext.revalidate();
        labelContext.paintImmediately(labelContext.getBounds());
        this.setVisible(true);
    }

    protected void finishTaskInLabelContext(String message) {
        if (!locked) {
            labelContext.setText(message);
            labelContext.setBackground((new JTextField()).getBackground());
            labelContext.setForeground(Color.black);
            this.setVisible(true);
        }
    }

    protected static void lock() {
        locked = true;
    }

    protected static void unlock() {
        locked = false;
    }

    private class OpenURL implements ActionListener{
        private final String[] browsers = {
        "firefox", "opera", "konqueror", "epiphany",
        "seamonkey", "galeon", "kazehakase", "mozilla", "netscape",
        "safari"};

        public void actionPerformed(ActionEvent ae){
            openURL(((JButton) ae.getSource()).getText());

        }
        public void openURL(String url) {
            String osName = System.getProperty("os.name");

            try {
                if (BiggestsInstall.isMacOS(osName.toLowerCase())) {
                    Class<?> fileMgr = Class.forName("com.apple.eio.FileManager");
                    Method openURL = fileMgr.getDeclaredMethod("openURL", new Class[] {String.class});
                    openURL.invoke(null, new Object[] {url});
                }
                else if (BiggestsInstall.isWindows(osName.toLowerCase()))
                    Runtime.getRuntime().exec("rundll32 url.dll,FileProtocolHandler " + url);
                else {
                    //assume Unix or Linux
                    boolean found = false;
                    for (String browser : browsers)
                        if (!found) {
                            found = Runtime.getRuntime().exec(
                                    new String[] {
                                    "which", browser}).waitFor() == 0;
                            if (found)
                                Runtime.getRuntime().exec(
                                        new String[] {
                                        browser, url});
                        }
                    if (!found)
                        throw new Exception(Arrays.toString(browsers));
                }
            }
            catch (Exception e) {
                JOptionPane.showMessageDialog(null, "Error attempting to launch web browser\n" + e.toString());
            }
        }
    }
}


/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 */
class BiggestsFrame_this_windowAdapter extends WindowAdapter {
    private BiggestsFrame adaptee;
    BiggestsFrame_this_windowAdapter(BiggestsFrame adaptee) {
        this.adaptee = adaptee;
    }

    public void windowClosing(WindowEvent e) {
        adaptee.this_windowClosing(e);
    }
}


class BiggestsFrame_this_windowStateAdapter implements WindowStateListener {
    private BiggestsFrame adaptee;
    BiggestsFrame_this_windowStateAdapter(BiggestsFrame adaptee) {
        this.adaptee = adaptee;
    }

    public void windowStateChanged(WindowEvent e) {
        adaptee.this_windowStateChanged(e);
    }
}


class BiggestsFrame_this_componentAdapter extends ComponentAdapter {
    private BiggestsFrame adaptee;
    BiggestsFrame_this_componentAdapter(BiggestsFrame adaptee) {
        this.adaptee = adaptee;
    }

    public void componentResized(ComponentEvent e) {
        adaptee.this_componentResized(e);
    }
}
