package biggests.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;

import biggests.tree.*;
import biggests.utils.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;

/**
 * <p>Title: Biclustering Panel</p>
 *
 * <p>Description: Displays available biclustering techniques
 *                 and allows for setting their corresponding
 *                 parameters and apply them to the gene
 *                 expression data.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */

public class PanelBiclustering extends PanelBasicIOValidation {
    /**
     * Name of the 'number of errors' parameter (e-CCC-Biclustering).
     */
    public static final String NUMBER_OF_ERRORS = "number of errors";

    /**
     * Name of the 'shifts' parameter for CCC-Biclustering
     * and e-CCC-Biclustering algorithms.
     */
    public static final String SHIFTS = "shifts";

    /**
     * Name of the 'neighbors' parameter (e-CCC-Biclustering).
     */
    public static final String NEIGHBORS = "neighbors";

    /**
     * Name of the 'delta' parameter (CC-TSB-biclustering).
     */
    public static final String DELTA = "delta";

    /**
     * Name of the 'alpha' parameter (CC-TSB-biclustering).
     */
    public static final String ALPHA = "alpha";

    /**
     * Name of the 'number of biclusters' parameter (CC-TSB-biclustering).
     */
    public static final String NUMBER_OF_BICLUSTERS = "number of biclusters";

    /**
     * Name of the 'maximum number of iterations' parameter
     * (CC-TSB-biclustering).
     */
    public static final String MAX_ITERATIONS = "maximum number of iterations";

    /**
     * Maximum time lag.
     */
    public static final String MAX_TIME_LAG = "max time lag";

    /**
     * CCC-Biclustering algorithm name.
     */
    public static final String CCC_BICLUSTERING = "CCC-Biclustering";

    /**
     * e-CCC-Biclustering algorithm name.
     */
    public static final String E_CCC_BICLUSTERING = "e-CCC-Biclustering";

    /**
     * CC-TSB-Biclustering algorithm name.
     */
    public static final String CC_TSB_BICLUSTERING = "CC-TSB-Biclustering";

    /**
     * Biclustering algorithms for discrete data.
     */
    private static final String[] discreteAlgorithms = {
            CCC_BICLUSTERING,
            E_CCC_BICLUSTERING};

    /**
     * Biclustering algorithms for continuous data.
     */
    private static final String[] floatAlgorithms = {CC_TSB_BICLUSTERING};

    /**
     * The application's main frame.
     */
    private BiggestsFrame mainFrame;

    /**
     * The node object which contains the original, preprocessed or
     * discretized matrix on which biclustering algorithm will be performed.
     */
    private NodeObjectInterface expressionMatrix;

    /**
     * The dataset tree selected path which corresponds to the path of
     * the dataset which leads to the node with the matrix on which
     * biclustering algorithm will be performed.
     */
    private TreePath selectionPath;

    /**
     * The layout of this panel.
     */
    private SpringLayout springLayout;

    /**
     * The biclustering algorithms label.
     */
    private JLabel biclusteringAlgorithmsLabel;

    /**
     * Gene shifts for CCC-Biclustering and e-CCC-Biclustering algorithms'
     * check box. Allows the user to select/deselect gene shifts for
     * these algorithms.
     */
    private JCheckBox geneShiftsCheckBox;

    /**
     * Label for number of shifts for CCC-Biclustering
     * and e-CCC-Biclustering algorithms.
     */
    private JLabel shiftsLabel;

    /**
     * Text field for the input of CCC-Biclustering and e-CCC-Biclustering
     * algorithm's number of shifts.
     */
    private JTextField shiftsTextField;

    /**
     * Restricted errors for e-CCC-Biclustering algorithm check box.
     * Allows the user to select/deselect restricted errors for
     * e-CCC-Biclustering algorithm.
     */
    private JCheckBox restrictedErrorsECCCCheckBox;

    /**
     * Sign changes check box.
     */
    private JCheckBox signChangesCheckBox;

    /**
     * Time lags check box.
     */
    private JCheckBox timeLagsCheckBox;

    /**
     * Maximum time lags text field.
     */
    private JTextField maxTimeLagTextField;

    /**
     * Maximum time lags label.
     */
    private JLabel maxTimeLagLabel;

    /**
     * Label for the number of neighbors on e-CCC-Biclustering algorithm.
     */
    private JLabel neighborsECCCLabel;

    /**
     * Text field for the input of number of neighbors for
     * e-CCC-Biclustering algorithm.
     */
    private JTextField neighborsECCCTextField;

    /**
     * Label for the number of errors allowed in e-CCC-Biclustering algorithm.
     */
    private JLabel errorsNumberLabel;

    /**
     * Text field for the input of the number of errors allowed in
     * e-CCC-Biclustering algorithm.
     */
    private JTextField errorsNumberTextField;

    /**
     * Delta input label;
     */
    private JLabel deltaLabel;

    /**
     * Delta input text field.
     */
    private JTextField deltaTextField;

    /**
     * Alpha input label.
     */
    private JLabel alphaLabel;

    /**
     * Alpha input text field.
     */
    private JTextField alphaTextField;

    /**
     * Number of biclusters label.
     */
    private JLabel biclustersNumberLabel;

    /**
     * Number of biclusters input text field.
     */
    private JTextField biclustersNumberTextField;

    /**
     * Maximum number of iterations label.
     */
    private JLabel maxIterationsLabel;

    /**
     * Maximum number of iterations input text field.
     */
    private JTextField maxIterationsTextField;

    private JLabel missingValuesLabel;
    private JRadioButton jumpMissingsButton;
    private JRadioButton allowMissingsButton;
    private ButtonGroup missingsButtonGroup;

    /**
     * Panel with biclustering algorithm name and parameters.
     */
    private JPanel biclusteringAlgorithmPanel;

    /**
     * List with biclustering algorithms.
     */
    private JList algorithmsList;

    /**
     * The apply button. When the user clicks it, biclustering
     * algorithms are applied to the matrix.
     */
    private JButton applyButton;

    /**
     * Information strings on the new bicluster group object.
     */
    private ArrayList<String> infoStrings;

    /**
     * Styles for the information strings on the new bicluster group object.
     */
    private ArrayList<String> infoStyles;

    /**
     * Creates a new biclustering panel, with biclustering algorithms
     * options.
     *
     * @param frame the main <code>BiggestsFrame</code>, in which this
     *              panel is being displayed
     */
    public PanelBiclustering(BiggestsFrame frame) {
        mainFrame = frame;

        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Component initialization.
     *
     * @throws Exception when components cannot be initialized
     */
    private void jbInit() throws Exception {
        this.setPreferredSize(new Dimension(750, 700));
        springLayout = new SpringLayout();
        this.setLayout(springLayout);
        this.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));

        this.biclusteringAlgorithmsLabel = new JLabel("Biclustering Algorithms");
        Font font = biclusteringAlgorithmsLabel.getFont();
        this.biclusteringAlgorithmsLabel.setFont(new Font(font.getName(),
                Font.BOLD, font.getSize() + 2));

        this.algorithmsList = new JList();
        algorithmsList.addListSelectionListener(new
                                                PanelBiclustering_algorithmsList_listSelectionAdapter(this));
        this.biclusteringAlgorithmPanel = new JPanel();
        this.biclusteringAlgorithmPanel.setPreferredSize(new Dimension(300, 270));

        this.applyButton = new JButton("Apply");
        applyButton.setPreferredSize(new Dimension(400, 50));

        applyButton.addActionListener(
            new
            PanelBiclustering_applyButton_actionAdapter(this));

        this.add(biclusteringAlgorithmsLabel);
        this.add(this.algorithmsList);
        springLayout.putConstraint(SpringLayout.WEST,
                                   algorithmsList,
                                   100,
                                   SpringLayout.EAST,
                                   biclusteringAlgorithmsLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   algorithmsList,
                                   0,
                                   SpringLayout.NORTH,
                                   biclusteringAlgorithmsLabel);

        this.add(this.biclusteringAlgorithmPanel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   biclusteringAlgorithmPanel,
                                   0,
                                   SpringLayout.WEST,
                                   biclusteringAlgorithmsLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   biclusteringAlgorithmPanel,
                                   30,
                                   SpringLayout.SOUTH,
                                   biclusteringAlgorithmsLabel);

        this.add(this.applyButton);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.applyButton,
                                   0,
                                   SpringLayout.WEST,
                                   this.biclusteringAlgorithmPanel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.applyButton,
                                   0,
                                   SpringLayout.SOUTH,
                                   this.biclusteringAlgorithmPanel);
    }

    /**
     * Paints biclustering algorithm panel with CCC-Biclustering algorithm.
     */
    public void paintBiclusteringAlgorithmPanelWithCCC() {
        SpringLayout sp = new SpringLayout();
        this.biclusteringAlgorithmPanel.setLayout(sp);

        JLabel name = new JLabel("CCC-Biclustering");

        this.shiftsLabel = new JLabel("(nr of shift(s))");
        this.shiftsTextField = new JTextField(3);
        this.shiftsTextField.setText("1");
        this.geneShiftsCheckBox = new JCheckBox("gene shifts");
        this.signChangesCheckBox = new JCheckBox("sign changes");
        signChangesCheckBox.addItemListener(new
                                            PanelBiclustering_signChangesCheckBox_itemAdapter(this));
        this.timeLagsCheckBox = new JCheckBox("time lags");
        timeLagsCheckBox.addItemListener(new
                                         PanelBiclustering_timeLagsCheckBox_itemAdapter(this));
        this.maxTimeLagLabel = new JLabel("(maximum time lag)");
        this.maxTimeLagTextField = new JTextField(3);
        this.maxTimeLagTextField.setText("1");
        this.setGeneShiftsCheckBoxSelected(false);
        geneShiftsCheckBox.addItemListener(new
                                           PanelBiclustering_geneShiftsCheckBox_itemAdapter(this));

        this.biclusteringAlgorithmPanel.add(name);

        this.biclusteringAlgorithmPanel.add(this.signChangesCheckBox);
        sp.putConstraint(SpringLayout.WEST,
                         this.signChangesCheckBox,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.signChangesCheckBox,
                         15,
                         SpringLayout.SOUTH,
                         name);

        this.biclusteringAlgorithmPanel.add(this.timeLagsCheckBox);
        sp.putConstraint(SpringLayout.WEST,
                         this.timeLagsCheckBox,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.timeLagsCheckBox,
                         5,
                         SpringLayout.SOUTH,
                         signChangesCheckBox);

        this.biclusteringAlgorithmPanel.add(this.maxTimeLagTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.maxTimeLagTextField,
                         10,
                         SpringLayout.EAST,
                         name);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.maxTimeLagTextField,
                         0,
                         SpringLayout.SOUTH,
                         timeLagsCheckBox);

        this.biclusteringAlgorithmPanel.add(this.maxTimeLagLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.maxTimeLagLabel,
                         5,
                         SpringLayout.EAST,
                         maxTimeLagTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.maxTimeLagLabel,
                         0,
                         SpringLayout.SOUTH,
                         timeLagsCheckBox);

        this.biclusteringAlgorithmPanel.add(geneShiftsCheckBox);
        sp.putConstraint(SpringLayout.WEST,
                         this.geneShiftsCheckBox,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.geneShiftsCheckBox,
                         5,
                         SpringLayout.SOUTH,
                         timeLagsCheckBox);

        this.biclusteringAlgorithmPanel.add(this.shiftsTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.shiftsTextField,
                         10,
                         SpringLayout.EAST,
                         this.geneShiftsCheckBox);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.shiftsTextField,
                         0,
                         SpringLayout.SOUTH,
                         this.geneShiftsCheckBox);

        this.biclusteringAlgorithmPanel.add(this.shiftsLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.shiftsLabel,
                         5,
                         SpringLayout.EAST,
                         this.shiftsTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.shiftsLabel,
                         0,
                         SpringLayout.SOUTH,
                         this.shiftsTextField);
    }

    /**
     * Paints biclustering algorithm panel with E-CCC-Biclustering algorithm.
     */
    public void paintBiclusteringAlgorithmPanelWithECCC() {
        SpringLayout sp = new SpringLayout();
        this.biclusteringAlgorithmPanel.setLayout(sp);

        JLabel name = new JLabel("e-CCC-Biclustering");

        if (((DiscretizedExpressionMatrix)expressionMatrix).hasMissingValues()) {
            this.missingValuesLabel = new JLabel("Missing values");
            this.jumpMissingsButton = new JRadioButton("Jump");
            this.jumpMissingsButton.setSelected(true);
            this.allowMissingsButton = new JRadioButton("Consider as errors");
            allowMissingsButton.addMouseListener(new
                    PanelBiclustering_allowMissingsButton_mouseAdapter(this));
            this.missingsButtonGroup = new ButtonGroup();
            missingsButtonGroup.add(jumpMissingsButton);
            missingsButtonGroup.add(allowMissingsButton);
        }

        this.errorsNumberLabel = new JLabel("error(s)");
        this.errorsNumberTextField = new JTextField(3);
        this.errorsNumberTextField.setText("1");
        this.shiftsLabel = new JLabel("(nr of shift(s))");
        this.shiftsTextField = new JTextField(3);
        this.shiftsTextField.setText("1");
        this.geneShiftsCheckBox = new JCheckBox("gene shifts");
        this.setGeneShiftsCheckBoxSelected(false);
        geneShiftsCheckBox.addItemListener(new
                                           PanelBiclustering_geneShiftsCheckBox_itemAdapter(this));

        this.restrictedErrorsECCCCheckBox = new JCheckBox("restricted errors");
        this.restrictedErrorsECCCCheckBox.setSelected(true);
        restrictedErrorsECCCCheckBox.addItemListener(new
                PanelBiclustering_restrictedErrorsECCCCheckBox_itemAdapter(this));
        this.neighborsECCCLabel = new JLabel("(nr of neighbor(s))");
        this.neighborsECCCTextField = new JTextField(3);
        this.neighborsECCCTextField.setText("1");

        this.signChangesCheckBox = new JCheckBox("sign changes");
        signChangesCheckBox.addItemListener(new
                                            PanelBiclustering_signChangesCheckBox_itemAdapter(this));
        this.timeLagsCheckBox = new JCheckBox("time lags");
        timeLagsCheckBox.addItemListener(new
                                         PanelBiclustering_timeLagsCheckBox_itemAdapter(this));
        this.maxTimeLagLabel = new JLabel("(maximum time lag)");
        this.maxTimeLagTextField = new JTextField(3);
        this.maxTimeLagTextField.setText("1");


        this.biclusteringAlgorithmPanel.add(name);

        if (((DiscretizedExpressionMatrix)expressionMatrix).hasMissingValues()) {
            this.biclusteringAlgorithmPanel.add(missingValuesLabel);
            sp.putConstraint(SpringLayout.WEST,
                             this.missingValuesLabel,
                             20,
                             SpringLayout.WEST,
                             name);
            sp.putConstraint(SpringLayout.NORTH,
                             this.missingValuesLabel,
                             15,
                             SpringLayout.SOUTH,
                             name);

            this.biclusteringAlgorithmPanel.add(jumpMissingsButton);
            sp.putConstraint(SpringLayout.WEST,
                             this.jumpMissingsButton,
                             40,
                             SpringLayout.WEST,
                             name);
            sp.putConstraint(SpringLayout.NORTH,
                             this.jumpMissingsButton,
                             5,
                             SpringLayout.SOUTH,
                             this.missingValuesLabel);

            this.biclusteringAlgorithmPanel.add(allowMissingsButton);
            sp.putConstraint(SpringLayout.WEST,
                             this.allowMissingsButton,
                             40,
                             SpringLayout.WEST,
                             name);
            sp.putConstraint(SpringLayout.NORTH,
                             this.allowMissingsButton,
                             5,
                             SpringLayout.SOUTH,
                             this.jumpMissingsButton);

            this.biclusteringAlgorithmPanel.add(this.errorsNumberTextField);
            sp.putConstraint(SpringLayout.WEST,
                             this.errorsNumberTextField,
                             20,
                             SpringLayout.WEST,
                             name);
            sp.putConstraint(SpringLayout.NORTH,
                             this.errorsNumberTextField,
                             15,
                             SpringLayout.SOUTH,
                             allowMissingsButton);
        }
        else {
            this.biclusteringAlgorithmPanel.add(this.errorsNumberTextField);
            sp.putConstraint(SpringLayout.WEST,
                             this.errorsNumberTextField,
                             20,
                             SpringLayout.WEST,
                             name);
            sp.putConstraint(SpringLayout.NORTH,
                             this.errorsNumberTextField,
                             15,
                             SpringLayout.SOUTH,
                             name);
        }

        this.biclusteringAlgorithmPanel.add(this.errorsNumberLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.errorsNumberLabel,
                         5,
                         SpringLayout.EAST,
                         this.errorsNumberTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.errorsNumberLabel,
                         0,
                         SpringLayout.SOUTH,
                         this.errorsNumberTextField);

        this.biclusteringAlgorithmPanel.add(this.signChangesCheckBox);
        sp.putConstraint(SpringLayout.WEST,
                         this.signChangesCheckBox,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.signChangesCheckBox,
                         5,
                         SpringLayout.SOUTH,
                         errorsNumberTextField);

        this.biclusteringAlgorithmPanel.add(this.timeLagsCheckBox);
        sp.putConstraint(SpringLayout.WEST,
                         this.timeLagsCheckBox,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.timeLagsCheckBox,
                         5,
                         SpringLayout.SOUTH,
                         signChangesCheckBox);

        this.biclusteringAlgorithmPanel.add(this.maxTimeLagTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.maxTimeLagTextField,
                         10,
                         SpringLayout.EAST,
                         name);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.maxTimeLagTextField,
                         0,
                         SpringLayout.SOUTH,
                         timeLagsCheckBox);

        this.biclusteringAlgorithmPanel.add(this.maxTimeLagLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.maxTimeLagLabel,
                         5,
                         SpringLayout.EAST,
                         maxTimeLagTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.maxTimeLagLabel,
                         0,
                         SpringLayout.SOUTH,
                         timeLagsCheckBox);

        this.biclusteringAlgorithmPanel.add(this.geneShiftsCheckBox);
        sp.putConstraint(SpringLayout.WEST,
                         this.geneShiftsCheckBox,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.geneShiftsCheckBox,
                         5,
                         SpringLayout.SOUTH,
                         this.timeLagsCheckBox);

        this.biclusteringAlgorithmPanel.add(this.shiftsTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.shiftsTextField,
                         10,
                         SpringLayout.EAST,
                         this.geneShiftsCheckBox);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.shiftsTextField,
                         0,
                         SpringLayout.SOUTH,
                         this.geneShiftsCheckBox);

        this.biclusteringAlgorithmPanel.add(this.shiftsLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.shiftsLabel,
                         5,
                         SpringLayout.EAST,
                         this.shiftsTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                 this.shiftsLabel,
                 0,
                 SpringLayout.SOUTH,
                 this.shiftsTextField);

        this.biclusteringAlgorithmPanel.add(this.restrictedErrorsECCCCheckBox);
        sp.putConstraint(SpringLayout.WEST,
                         this.restrictedErrorsECCCCheckBox,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.restrictedErrorsECCCCheckBox,
                         5,
                 SpringLayout.SOUTH,
                 this.shiftsTextField);

        this.biclusteringAlgorithmPanel.add(this.neighborsECCCTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.neighborsECCCTextField,
                 10,
                 SpringLayout.EAST,
                 this.restrictedErrorsECCCCheckBox);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.neighborsECCCTextField,
                         0,
                         SpringLayout.SOUTH,
                         this.restrictedErrorsECCCCheckBox);

        this.biclusteringAlgorithmPanel.add(this.neighborsECCCLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.neighborsECCCLabel,
                         5,
                         SpringLayout.EAST,
                         this.neighborsECCCTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.neighborsECCCLabel,
                         0,
                         SpringLayout.SOUTH,
                         this.neighborsECCCTextField);
    }

    /**
     * Paints biclustering algorithm panel with CC-TSB-Biclustering algorithm.
     */
    public void paintBiclusteringAlgorithmPanelWithCCTSB() {
        SpringLayout sp = new SpringLayout();
        this.biclusteringAlgorithmPanel.setLayout(sp);

        JLabel name = new JLabel("CC-TSB-Biclustering");

        this.deltaLabel = new JLabel("delta");
        this.deltaTextField = new JTextField(3);
        this.deltaTextField.setText("0");
        this.alphaLabel = new JLabel("alpha");
        this.alphaTextField = new JTextField(3);
        this.alphaTextField.setText("1");
        this.biclustersNumberLabel = new JLabel("nr of biclusters to extract");
        this.biclustersNumberTextField = new JTextField(3);
        this.biclustersNumberTextField.setText("10");
        this.maxIterationsLabel = new JLabel("maximum nr of iterations");
        this.maxIterationsTextField = new JTextField(3);
        this.maxIterationsTextField.setText("50");

        this.biclusteringAlgorithmPanel.add(name);

        this.biclusteringAlgorithmPanel.add(deltaTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.deltaTextField,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.deltaTextField,
                         5,
                         SpringLayout.SOUTH,
                         name);

        this.biclusteringAlgorithmPanel.add(this.deltaLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.deltaLabel,
                         5,
                         SpringLayout.EAST,
                         this.deltaTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.deltaLabel,
                         0,
                         SpringLayout.SOUTH,
                         this.deltaTextField);

        this.biclusteringAlgorithmPanel.add(this.alphaTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.alphaTextField,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.alphaTextField,
                         5,
                         SpringLayout.SOUTH,
                         this.deltaTextField);

        this.biclusteringAlgorithmPanel.add(this.alphaLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.alphaLabel,
                         5,
                         SpringLayout.EAST,
                         this.alphaTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.alphaLabel,
                         0,
                         SpringLayout.SOUTH,
                         this.alphaTextField);

        this.biclusteringAlgorithmPanel.add(this.biclustersNumberTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.biclustersNumberTextField,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.biclustersNumberTextField,
                         5,
                         SpringLayout.SOUTH,
                         this.alphaTextField);

        this.biclusteringAlgorithmPanel.add(this.biclustersNumberLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.biclustersNumberLabel,
                         5,
                         SpringLayout.EAST,
                         this.biclustersNumberTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.biclustersNumberLabel,
                         0,
                         SpringLayout.SOUTH,
                         this.biclustersNumberTextField);

        this.biclusteringAlgorithmPanel.add(this.maxIterationsTextField);
        sp.putConstraint(SpringLayout.WEST,
                         this.maxIterationsTextField,
                         20,
                         SpringLayout.WEST,
                         name);
        sp.putConstraint(SpringLayout.NORTH,
                         this.maxIterationsTextField,
                         5,
                         SpringLayout.SOUTH,
                         this.biclustersNumberTextField);

        this.biclusteringAlgorithmPanel.add(this.maxIterationsLabel);
        sp.putConstraint(SpringLayout.WEST,
                         this.maxIterationsLabel,
                         5,
                         SpringLayout.EAST,
                         this.maxIterationsTextField);
        sp.putConstraint(SpringLayout.SOUTH,
                         this.maxIterationsLabel,
                         0,
                         SpringLayout.SOUTH,
                         this.maxIterationsTextField);
    }

    /**
     * Paints this biclustering panel with options to perform biclustering
     * on the <code>discretizedExpressionMatrix</code> and options to
     * perform function analysis.
     *
     * @param discretizedExpressionMatrix the <code>DiscretizedExpressionMatrix</code>
     *        to which biclustering algorithm and function analysis will
     *        be applied.
     * @param selectionPath the <code>TreePath</code> selected on the dataset
     *        tree; corresponds to the node that contains the
     *        <code>discretizedExpressionMatrix</code>
     */
    public void paint(DiscretizedExpressionMatrix discretizedExpressionMatrix,
                      TreePath selectionPath) {
        this.expressionMatrix = discretizedExpressionMatrix;
        this.selectionPath = selectionPath;
//        if (!discretizedExpressionMatrix.getOriginalExpressionMatrix().hasMissingValues()) {
            this.algorithmsList.setListData(this.discreteAlgorithms);
/*        }
        else {
            this.algorithmsList.setListData(new String[] {CCC_BICLUSTERING});
        }
 */
        this.algorithmsList.setSelectedIndex(0);
        this.maxTimeLagLabel.setEnabled(false);
        this.maxTimeLagTextField.setEnabled(false);
    }

    /**
     * Paints this biclustering panel with options to perform biclustering
     * and function analysis on <code>AbstractExpressionMatrix</code> objects
     * (<code>ExpressionMatrix</code> and
     * <code>PreProcessedExpressionMatrix</code>).
     *
     * @param expressionMatrix the <code>AbstractExpressionMatrix</code>
     *        to which biclustering algorithm and function analysis will
     *        be applied
     * @param selectionPath the <code>TreePath</code> selected on the dataset
     *        tree; corresponds to the node that contains the
     *        <code>expressionMatrix</code>
     */
    public void paint(AbstractExpressionMatrix expressionMatrix,
                      TreePath selectionPath) {
        this.expressionMatrix = expressionMatrix;
        this.selectionPath = selectionPath;
        this.algorithmsList.setListData(this.floatAlgorithms);

        this.algorithmsList.setSelectedIndex(0);
    }

    /**
     * Clears this panel's components and other attributes that are
     * no longer needed.
     */
    public void clear() {
        this.biclusteringAlgorithmPanel.removeAll();
        this.expressionMatrix = null;
        this.selectionPath = null;
        this.infoStrings = null;
        this.infoStyles = null;
    }

    /**
     * Sets gene shifts check box selected/deselected, along with
     * the corresponding enabling/disabling of shifts label and
     * input text field.
     *
     * @param selected <code>boolean</code> <code>true</code> if
     * gene shifts check box should be selected and <code>false</code>
     * otherwise
     */
    public void setGeneShiftsCheckBoxSelected(boolean selected) {
        this.geneShiftsCheckBox.setEnabled(true);
        this.geneShiftsCheckBox.setSelected(selected);
        this.shiftsLabel.setEnabled(selected);
        this.shiftsTextField.setEnabled(selected);
    }

    /**
     * Sets restricted errors for e-CCC-Biclustering algorithm check
     * box selected/deselected, along with the corresponding
     * enabling/disabling of neighbors label and text field.
     *
     * @param selected <code>boolean</code> value <code>true</code> if
     *        the check box should get selected; <code>false</code> otherwise
     */
    public void setRestrictedErrorsECCCCheckBoxSelected(boolean selected) {
        this.restrictedErrorsECCCCheckBox.setEnabled(true);
        this.restrictedErrorsECCCCheckBox.setSelected(selected);
        this.neighborsECCCLabel.setEnabled(selected);
        this.neighborsECCCTextField.setEnabled(selected);
    }

    /**
     * Handles the item state changed event fired when the user
     * selects/deselects restricted errors for e-CCC-Biclustering
     * algorithm check box.
     *
     * @param e the <code>ItemEvent</code> fired when the user
     *          selects/deselects restricted errors for e-CCC-Biclustering
     *          algorithm check box
     */
    public void restrictedErrorsECCCCheckBox_itemStateChanged(ItemEvent e) {
        boolean selected = this.restrictedErrorsECCCCheckBox.isSelected();
        this.setRestrictedErrorsECCCCheckBoxSelected(selected);
        if (selected && allowMissingsButton.isSelected()) {
            jumpMissingsButton.setSelected(true);
        }
    }

    /**
     * Handles the mouse event fired when the user clicks the apply button.
     *
     * @param e the <code>ActionEvent</code> fired when the user clicks
     *          the apply button
     */
    public void applyButton_actionPerformed(ActionEvent e) {
        this.applyBiclusteringAlgorithm();
    }

    /**
     * Applies biclustering algorithm.
     */
    public void applyBiclusteringAlgorithm() {
        String algorithm = (String)this.algorithmsList.getSelectedValue();
        if (expressionMatrix instanceof DiscretizedExpressionMatrix) {
            if (algorithm.equals(CCC_BICLUSTERING)) {
                try {
                    applyCCCBiclusteringAlgorithm();
                } catch (Exception ex) {
                    ex.printStackTrace();
                    return;
                }
            } else
            if (algorithm.equals(E_CCC_BICLUSTERING)) {
                try {
                    applyECCCBiclusteringAlgorithm();
                } catch (Exception ex) {
                    ex.printStackTrace();
                    return;
                }
            } else {
                JOptionPane.showMessageDialog(this, "Select a biclustering " +
                                              "algorithm to apply",
                                              "No algorithm selected",
                                              JOptionPane.INFORMATION_MESSAGE);
            }
        } else {
            if (algorithm.equals(CC_TSB_BICLUSTERING)) {
                try {
                    applyCCTSBBiclusteringAlgorithm();
                } catch (Exception ex) {
                    ex.printStackTrace();
                    return;
                }
            } else {
                JOptionPane.showMessageDialog(this, "Select a biclustering " +
                                              "algorithm to apply",
                                              "No algorithm selected",
                                              JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }

    /**
     * Applies CC-TSB-biclustering algorithm.
     *
     * @throws Exception
     */
    public void applyCCTSBBiclusteringAlgorithm() throws
            Exception {
        Biclustering biclustering = this.preCCTSB();

        new ThreadBiclustering(mainFrame, biclustering, selectionPath);

        if (biclustering == null) {
            throw new Exception();
        }
    }

    /**
     * Applies CCC-Biclustering algorithm.
     *
     * @throws Exception
     */
    public void applyCCCBiclusteringAlgorithm() throws
            Exception {
        Biclustering biclustering = this.preCCC();

        new ThreadBiclustering(mainFrame, biclustering, selectionPath);

        if (biclustering == null) {
            throw new Exception();
        }
    }

    /**
     * Applies e-CCC-Biclustering algorithm.
     *
     * @throws Exception
     */
    public void applyECCCBiclusteringAlgorithm() throws
            Exception {
        Biclustering biclustering = this.preECCC();

        new ThreadBiclustering(mainFrame, biclustering, selectionPath);

        if (biclustering == null) {
            throw new Exception();
        }
    }

    /**
     * Constructs a CC_TSB_Biclustering object with parameters
     * entered by the user.
     *
     * @return CC_TSB_Biclustering the group of biclusters over
     *         which CC-TSB-Biclustering algorithm will be performed.
     * @throws Exception
     */
    private CC_TSB_Biclustering preCCTSB() throws Exception {
        float delta = 0;
        float alpha = 1;
        int biclustersNumber = 10;
        int maxIterations = 50;

        delta = this.readParameterFloat(0, PanelBiclustering.DELTA);
        alpha = this.readParameterFloat(1, PanelBiclustering.ALPHA);
        biclustersNumber = this.readParameterInteger(1,
                PanelBiclustering.NUMBER_OF_BICLUSTERS);
        maxIterations = this.readParameterInteger(
            1, PanelBiclustering.MAX_ITERATIONS);

        CC_TSB_Biclustering biclustering = new CC_TSB_Biclustering(
                (AbstractExpressionMatrix) expressionMatrix,
                delta,
                alpha,
                biclustersNumber,
                maxIterations);

        return biclustering;
    }

    /**
     *
     * Constructs a CCC_Biclustering object with parameters
     * entered by the user.
     *
     * @return CCC_Biclustering the group of biclusters over
     *         which CCC-Biclustering algorithm will be performed.
     * @throws Exception
     */
    private CCC_Biclustering preCCC() throws Exception {
        int shifts = 1;
        boolean geneShifts;
        boolean signChanges;
        boolean timeLags;
        int maxTimeLag = 1;

        DiscretizedExpressionMatrix discreteMatrix = (
                DiscretizedExpressionMatrix)
                expressionMatrix;
        geneShifts = this.geneShiftsCheckBox.isSelected();
        signChanges = this.signChangesCheckBox.isSelected();
        timeLags = this.timeLagsCheckBox.isSelected();

        CCC_Biclustering biclustering = null;

        if (geneShifts) {
            shifts = this.readParameterInteger(
            1, discreteMatrix.getAlphabet().length, this.SHIFTS);
            if (discreteMatrix.getOriginalExpressionMatrix().hasMissingValues()) {
                biclustering =
                        new CCC_Biclustering_GeneShifts_Jumping_MissingValues(
                                discreteMatrix, shifts);
            }
            else {
                biclustering =
                        new CCC_Biclustering_GeneShifts(discreteMatrix, shifts);
            }
        } else
        if (timeLags) {
            maxTimeLag = this.readParameterInteger(1,
                    discreteMatrix.getNumberOfConditions() - 1,
                    this.MAX_TIME_LAG);
            if (signChanges) {
                if (discreteMatrix.hasMissingValues()) {
                    biclustering =
                            new
                            CCC_Biclustering_SignChanges_TimeLags_Jumping_MissingValues(
                                    discreteMatrix, maxTimeLag);
                }
                else {
                    biclustering = new CCC_Biclustering_SignChanges_TimeLags(
                            discreteMatrix, maxTimeLag);
                }
            } else {
                if (discreteMatrix.hasMissingValues()) {
                    biclustering =
                            new CCC_Biclustering_TimeLags_Jumping_MissingValues(
                                    discreteMatrix, maxTimeLag);
                }
                else {
                    biclustering =
                            new CCC_Biclustering_TimeLags(
                                    discreteMatrix,
                                    maxTimeLag);
                }
            }
        } else
        if (signChanges) {
            if (discreteMatrix.hasMissingValues()) {
                biclustering =
                        new CCC_Biclustering_SignChanges_Jumping_MissingValues(
                                discreteMatrix);
            }
            else {
                biclustering =
                        new CCC_Biclustering_SignChanges(discreteMatrix);
            }
        } else {
            if (discreteMatrix.hasMissingValues()) {
                biclustering =
                        new CCC_Biclustering_Jumping_MissingValues(
                                discreteMatrix);
            }
            else {
                biclustering = new CCC_Biclustering(discreteMatrix);
            }
        }

        return biclustering;
    }

    /**
     *
     * Constructs an E_CCC_Biclustering object with parameters
     * entered by the user.
     *
     * @return E_CCC_Biclustering the group of biclusters over
     *         which e-CCC-Biclustering algorithm will be performed.
     * @throws Exception
     */
    private E_CCC_Biclustering preECCC() throws Exception {
        int numberOfErrors;
        int shifts = 1;
        int neighbors = 0;
        boolean geneShifts;
        boolean restrictedErrors;
        boolean signChanges;
        boolean timeLags;
        int maxTimeLag = 1;

        DiscretizedExpressionMatrix discretizedMatrix = (
                DiscretizedExpressionMatrix) expressionMatrix;
        int upperLimit = discretizedMatrix.getNumberOfConditions();
        numberOfErrors = this.readParameterInteger(1,
                upperLimit,
                this.NUMBER_OF_ERRORS);

        geneShifts = this.geneShiftsCheckBox.isSelected();
        restrictedErrors = this.restrictedErrorsECCCCheckBox.isSelected();
        signChanges = this.signChangesCheckBox.isSelected();
        timeLags = this.timeLagsCheckBox.isSelected();

        if (geneShifts) {
            shifts = this.readParameterInteger(1,
                                               discretizedMatrix.getAlphabet().
                                               length,
                                               this.SHIFTS);
        }
        if (restrictedErrors) {
            neighbors = this.readParameterInteger(0, upperLimit,
                                                  "neighbors");
        }
        if (timeLags) {
            maxTimeLag = this.readParameterInteger(1,
                    discretizedMatrix.getNumberOfConditions() - 1,
                    this.MAX_TIME_LAG);
        }

        E_CCC_Biclustering biclustering = null;
        if (!discretizedMatrix.hasMissingValues()) {
            if (geneShifts) {
                if (restrictedErrors) {
                    biclustering = new
                            E_CCC_Biclustering_GeneShifts_RestrictedErrors(
                                    discretizedMatrix,
                                    numberOfErrors, shifts, neighbors);
                }
                else {
                    biclustering = new E_CCC_Biclustering_GeneShifts(
                            discretizedMatrix, numberOfErrors, shifts);
                }
            }
            else {
                if (restrictedErrors) {
                    if (signChanges && timeLags) {
                        biclustering =
                                new
                                E_CCC_Biclustering_SignChanges_TimeLags_RestrictedErrors(
                                        discretizedMatrix, numberOfErrors, maxTimeLag, neighbors);
                    }
                    else
                    if (signChanges) {
                        biclustering =
                                new
                                E_CCC_Biclustering_SignChanges_RestrictedErrors(
                                discretizedMatrix, numberOfErrors, neighbors);
                    }
                    else
                    if (timeLags) {
                        biclustering =
                                new E_CCC_Biclustering_TimeLags_RestrictedErrors(
                                discretizedMatrix, numberOfErrors,
                                maxTimeLag, neighbors);
                    }
                    else {
                        biclustering = new E_CCC_Biclustering_RestrictedErrors(
                                discretizedMatrix, numberOfErrors, neighbors);
                    }
                }
                else {
                    if (signChanges && timeLags) {
                        biclustering = new
                                       E_CCC_Biclustering_SignChanges_TimeLags(
                                               discretizedMatrix,
                                               numberOfErrors, maxTimeLag);
                    } else
                    if (signChanges) {
                        biclustering = new E_CCC_Biclustering_SignChanges(
                                discretizedMatrix, numberOfErrors);
                    }
                    else
                    if (timeLags) {
                        biclustering = new E_CCC_Biclustering_TimeLags(
                                discretizedMatrix, numberOfErrors, maxTimeLag);
                    }
                    else {
                        biclustering = new E_CCC_Biclustering(
                                discretizedMatrix, numberOfErrors);
                    }
                } // else (if (restrictedErrors))
            } // else (if (geneShifts))
        }
        else {
            boolean jumpMissings = jumpMissingsButton.isSelected();
            if (restrictedErrors) {
                if (signChanges && timeLags && jumpMissings) {
                    biclustering =
                            new
                            E_CCC_Biclustering_SignChanges_TimeLags_RestrictedErrors_Jumping_MissingValues(
                                    discretizedMatrix, numberOfErrors,
                                    maxTimeLag, neighbors);
                }
                else
                if (signChanges && jumpMissings) {
                    biclustering =
                            new
                            E_CCC_Biclustering_SignChanges_RestrictedErrors_Jumping_MissingValues(
                                    discretizedMatrix, numberOfErrors,
                                    neighbors);
                }
                else
                if (timeLags && jumpMissings) {
                        biclustering =
                                new
                                E_CCC_Biclustering_TimeLags_RestrictedErrors_Jumping_MissingValues(
                                        discretizedMatrix, numberOfErrors,
                                        maxTimeLag, neighbors);
                }
                else {
                    if (jumpMissings) {
                        biclustering =
                                new
                                E_CCC_Biclustering_RestrictedErrors_Jumping_MissingValues(
                                discretizedMatrix, numberOfErrors, neighbors);
                    }
                }
            }
            else {
                if (signChanges && timeLags) {
                    if (jumpMissings) {
                        biclustering =
                                new
                                E_CCC_Biclustering_SignChanges_TimeLags_Jumping_MissingValues(
                                               discretizedMatrix,
                                               numberOfErrors, maxTimeLag);
                    }
                    else {
                        biclustering =
                                new
                                E_CCC_Biclustering_SignChanges_TimeLags_Allowing_MissingValues(
                                               discretizedMatrix,
                                               numberOfErrors, maxTimeLag);
                    }
                } else
                if (signChanges) {
                    if (jumpMissings) {
                        biclustering = new E_CCC_Biclustering_SignChanges_Jumping_MissingValues(
                                discretizedMatrix, numberOfErrors);
                    }
                    else {
                        biclustering = new E_CCC_Biclustering_SignChanges_Allowing_MissingValues(
                                discretizedMatrix, numberOfErrors);
                    }
                }
                else
                if (timeLags) {
                    if (jumpMissings) {
                        biclustering = new E_CCC_Biclustering_TimeLags_Jumping_MissingValues(
                                discretizedMatrix, numberOfErrors, maxTimeLag);
                    }
                    else {
                        biclustering = new E_CCC_Biclustering_TimeLags_Allowing_MissingValues(
                                discretizedMatrix, numberOfErrors, maxTimeLag);
                    }
                }
                else {
                    if (jumpMissings) {
                        biclustering = new
                                       E_CCC_Biclustering_Jumping_MissingValues(
                                               discretizedMatrix,
                                               numberOfErrors);
                    } else {
                        biclustering = new
                                E_CCC_Biclustering_Allowing_MissingValues(
                                        discretizedMatrix, numberOfErrors);
                    }
                }
            } // else (if (restrictedErrors))
        }

/*        if (!geneShifts && !restrictedErrors) {
            biclustering = new E_CCC_Biclustering(
                    (DiscretizedExpressionMatrix) expressionMatrix,
                    numberOfErrors);
        } else
        if (geneShifts && !restrictedErrors) {
            biclustering = new E_CCC_Biclustering_GeneShifts(
                    (DiscretizedExpressionMatrix) expressionMatrix,
                    numberOfErrors, shifts);
        } else
        if (!geneShifts && restrictedErrors) {
            biclustering = new E_CCC_Biclustering_RestrictedErrors(
                    (DiscretizedExpressionMatrix) expressionMatrix,
                    numberOfErrors, neighbors);
        } else {
            biclustering = new E_CCC_Biclustering_GeneShifts_RestrictedErrors(
                    (DiscretizedExpressionMatrix) expressionMatrix,
                    numberOfErrors, shifts, neighbors);
        }
 */
        return biclustering;
    }

    /**
     * Sets strings and styles attributes according to
     * the type and parameters of <code>biclustering</code> object.
     *
     * @param biclustering <code>Biclustering</code>
     */
    public void setStringsAndStyles(Biclustering biclustering) {
        if (biclustering instanceof CC_TSB_Biclustering) {
            this.setStringsAndStylesCCTSB((CC_TSB_Biclustering) biclustering);
            return;
        }
        if (biclustering instanceof CCC_Biclustering) {
            this.setStringsAndStylesCCC((CCC_Biclustering) biclustering);
            return;
        }
        this.setStringsAndStylesECCC((E_CCC_Biclustering) biclustering);
    }

    /**
     * Sets strings and styles attributes according to the
     * parameters of the <code>CC_TSB_Biclustering</code> object.
     *
     * @param biclustering <code>CC_TSB_Biclustering</code>
     */
    public void setStringsAndStylesCCTSB(CC_TSB_Biclustering biclustering) {
        infoStrings = new ArrayList<String>();
        infoStyles = new ArrayList<String>();

        infoStrings.add("Biclustering algorithm:");
        infoStyles.add("bold");
        infoStrings.add(" CC-TSB-biclustering\n");
        infoStyles.add("regular");
        infoStrings.add("   delta: " + biclustering.getDelta() + "\n");
        infoStyles.add("regular");
        infoStrings.add("   alpha: " + biclustering.getAlpha() + "\n");
        infoStyles.add("regular");
        infoStrings.add("   n. of biclusters: " +
                               biclustering.getnumberOfBiclustersToExtract() +
                               "\n");
        infoStyles.add("regular");
        infoStrings.add("   maximum n. of iterations: " +
                               biclustering.getMaxNumberOfIterations() + "\n");
        infoStyles.add("regular");
    }

    /**
     * Sets strings and styles attributes according to the
     * parameters of the <code>CCC_Biclustering</code> object.
     *
     * @param biclustering <code>CCC_Biclustering</code>
     */
    public void setStringsAndStylesCCC(CCC_Biclustering biclustering) {
        infoStrings = new ArrayList<String>();
        infoStyles = new ArrayList<String>();
        infoStrings.add("Biclustering algorithm:");
        infoStyles.add("bold");
        infoStrings.add(" CCC-Biclustering\n");
        infoStyles.add("regular");

        if (biclustering instanceof IMissingValuesJumped) {
            infoStrings.add("   missing values: jumped over\n");
            infoStyles.add("regular");
        }
        if (biclustering instanceof CCC_Biclustering_GeneShifts) {
            infoStrings.add("   gene shifts: " +
                                   ((CCC_Biclustering_GeneShifts) biclustering).
                                   getNumberOfShifts() + "\n");
            infoStyles.add("regular");
        }
        else {
            if (biclustering instanceof CCC_Biclustering_SignChanges ||
            biclustering instanceof CCC_Biclustering_SignChanges_TimeLags) {
                infoStrings.add("   sign changes: yes\n");
                infoStyles.add("regular");
            }
            if (biclustering instanceof CCC_Biclustering_TimeLags) {
                infoStrings.add("   time lags: max. time lag " +
                                ((CCC_Biclustering_TimeLags) biclustering).
                                getMaxTimeLag() + "\n");
                infoStyles.add("regular");
            }
        }
    }

    /**
     * Sets strings and styles attributes according to the
     * parameters of the <code>E_CCC_Biclustering</code> object.
     *
     * @param biclustering <code>E_CCC_Biclustering</code>
     */
    public void setStringsAndStylesECCC(E_CCC_Biclustering biclustering) {
        infoStrings = new ArrayList<String>();
        infoStyles = new ArrayList<String>();

        infoStrings.add("Biclustering algorithm:");
        infoStyles.add("bold");
        infoStrings.add(" e-CCC-Biclustering\n");
        infoStyles.add("regular");
        infoStrings.add("   max. n. of errors: " +
                               biclustering.getMaxNumberOfErrors() + "\n");
        infoStyles.add("regular");


        if (biclustering instanceof IRestrictedErrors) {
             infoStrings.add("   restricted errors: " + ((IRestrictedErrors) biclustering).
                             getNumberOfNeighborsInRestrictedErrors() +
                             " neighbor(s)\n");
             infoStyles.add("regular");
        }
        if (biclustering instanceof IMissingValuesAllowedAsErrors) {
            infoStrings.add("   missing values: allowed as errors\n");
            infoStyles.add("regular");
        }
        else if (biclustering instanceof IMissingValuesJumped) {
            infoStrings.add("   missing values: jumped over\n");
            infoStyles.add("regular");
        }

        if (biclustering instanceof E_CCC_Biclustering_GeneShifts) {
            infoStrings.add("   gene shifts: " +
                            ((E_CCC_Biclustering_GeneShifts) biclustering).
                                   getNumberOfShifts() + "\n");
            infoStyles.add("regular");
        }
        else {
            if (biclustering instanceof E_CCC_Biclustering_SignChanges ||
                biclustering instanceof E_CCC_Biclustering_SignChanges_TimeLags) {
                infoStrings.add("   sign changes: yes\n");
                infoStyles.add("regular");
            }
            if (biclustering instanceof E_CCC_Biclustering_TimeLags) {
                infoStrings.add("   time lags: max. time lag " +
                                ((E_CCC_Biclustering_TimeLags)biclustering).getMaxTimeLag() + "\n");
                infoStyles.add("regular");
            }
        }
    }

    /**
     * Adds a bicluster group to the dataset tree.
     *
     * @param biclustering <code>Biclustering</code> - biclusters' group
     * @param selectionPath TreePath
     * @param go the <code>GOFrontEnd</code> for the biclusters' group
     */
    public void addBiclusterGroupToDatasetTree(Biclustering biclustering,
                                               TreePath selectionPath,
                                               GOFrontEnd go) {
        ArrayList<String> newStrings = new ArrayList<String>();
        ArrayList<String> newStyles = new ArrayList<String>();
        newStrings.add("Number of biclusters: ");
        newStyles.add("bold");
        newStrings.add(biclustering.getNumberOfBiclusters() +
                                    "\n\n");
        newStyles.add("regular");
        newStrings.add("Datatype: ");
        newStyles.add("bold");
        if (biclustering instanceof CCC_Biclustering ||
            biclustering instanceof E_CCC_Biclustering) {
            newStrings.add("discrete symbols\n");
            newStyles.add("regular");
        } else {
            newStrings.add("floating point values\n");
            newStyles.add("regular");
        }
        newStrings.add("Organism: ");
        newStyles.add("bold");

        IMatrix matrix = (IMatrix) biclustering.getMatrix();
        String organism;
        if (matrix instanceof AbstractExpressionMatrix) {
            organism = ((AbstractExpressionMatrix) matrix).getOrganism();
        } else {
            organism = ((DiscretizedExpressionMatrix) matrix).
                       getOriginalExpressionMatrix().getOrganism();
        }
        newStrings.add(organism + "\n\n");
        newStyles.add("regular");
        newStrings.add("\n");
        newStyles.add("regular");
        if (biclustering instanceof IMissingValues) {
            infoStrings.add("   missing values: yes\n");
            infoStyles.add("regular");
        }

        newStrings.addAll(this.infoStrings);
        newStyles.addAll(this.infoStyles);

        try {
            NodeInfo biclusteringNodeInfo = new NodeInfo(
                    biclustering,
                    NodeInfo.BICLUSTER_GROUP,
                    selectionPath,
                    mainFrame.getTree(),
                    (NodeInfo) ((DefaultMutableTreeNode) selectionPath.
                                getLastPathComponent()).getUserObject(),
                    newStrings,
                    newStyles);
            TreePath newPath = mainFrame.getTree().addNewNode(
                    biclusteringNodeInfo, selectionPath, true);

            mainFrame.getPanelAnalyzingFunctionTable().putGOFrontEnd(newPath,
                    go);
            mainFrame.getTree().setSelectionPath(newPath);
        } catch (InvalidNodeObjectTypeException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "Unable to add the " +
                                          "new bicluster group node to the tree.",
                                          "Unable to add bicluster group node to the tree",
                                          JOptionPane.ERROR_MESSAGE);
        }
    }

    public int readParameterInteger(int min, int max, String parameterToRead) throws
            Exception {
        int parameter;
        parameter = readParameterInteger(parameterToRead);

        if (parameterToRead.equals(this.NUMBER_OF_ERRORS) && (parameter == 0)) {
            JOptionPane.showMessageDialog(
            this,
            "The e-CCC-Biclustering algorithm is\n" +
            "designed to identify biclusters with\n" +
            "approximate expression patterns.\n\n" +
            "For discovering biclusters with exact\n" +
            "expression patterns, we recommend using\n" +
            "the CCC-Biclustering algorithm instead.",
            "Use CCC-Biclustering",
            JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        else if ((parameter < min) || (parameter > max)) {
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be between " + min + " and " +
                                          max,
                                          parameterToRead +
                                          " must be between [" + min + ", " +
                                          max + "]",
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads a parameter from its input text field box.
     *
     * @param parameterToRead <code>String</code> the name of the
     *        parameter to read
     * @return String the parameter read from the input text field
     * @throws Exception if parameter could not be read
     * @see PanelBasicIOValidation#readParameter(String parameterToRead)
     */
    public String readParameter(String parameterToRead) throws Exception {
        String parameter = null;
        if (parameterToRead.equals(this.NUMBER_OF_ERRORS)) {
            parameter = this.errorsNumberTextField.getText();
        } else
        if (parameterToRead.equals(this.SHIFTS)) {
            parameter = this.shiftsTextField.getText();
        } else
        if (parameterToRead.equals(this.NEIGHBORS)) {
            parameter = this.neighborsECCCTextField.getText();
        } else
        if (parameterToRead.equals(this.DELTA)) {
            parameter = this.deltaTextField.getText();
        } else
        if (parameterToRead.equals(this.ALPHA)) {
            parameter = this.alphaTextField.getText();
        } else
        if (parameterToRead.equals(this.NUMBER_OF_BICLUSTERS)) {
            parameter = this.biclustersNumberTextField.getText();
        } else
        if (parameterToRead.equals(this.MAX_ITERATIONS)) {
            parameter = this.maxIterationsTextField.getText();
        } else
        if (parameterToRead.equals(this.MAX_TIME_LAG)) {
            parameter = this.maxTimeLagTextField.getText();
        }

        if ((parameter == null) ||
            (parameter.equals(""))) {
            JOptionPane.showMessageDialog(this, "Please type "
                                          + parameterToRead + ".",
                                          "Type " + parameterToRead,
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Returns the application's main frame.
     *
     * @return the main <code>BiggestsFrame</code>
     */
    public BiggestsFrame getMainFrame() {
        return mainFrame;
    }

    /**
     * Returns the node object which contains the original, preprocessed or
     * discretized matrix on which biclustering algorithm will be performed.
     *
     * @return the <code>NodeObjectInterface</code> which contains a gene
     *         expression matrix (real valued or discretized)
     */
    public NodeObjectInterface getExpressionMatrix() {
        return expressionMatrix;
    }

    /**
     * Returns the dataset tree selected path which corresponds to the path of
     * the dataset which in turn leads to the node with the matrix on which
     * biclustering algorithm will be performed.
     *
     * @return the <code>TreePath</code> selected in the dataset tree
     */
    public TreePath getSelectionPath() {
        return selectionPath;
    }

    /**
     * Returns the layout of this panel.
     *
     * @return <code>SpringLayout</code>
     */
    public SpringLayout getSpringLayout() {
        return springLayout;
    }

    /**
     * Returns biclustering algorithms label.
     *
     * @return the biclustering algorithms <code>JLabel</code>
     */
    public JLabel getBiclusteringAlgorithmsLabel() {
        return biclusteringAlgorithmsLabel;
    }

    /**
     * Returns gene shifts for CCC-Biclustering and e-CCC-Biclustering
     * algorithms check box.
     *
     * @return gene shifts for CCC-Biclustering and e-CCC-Biclustering
     *         algorithm <code>JCheckBox</code>
     */
    public JCheckBox getGeneShiftsCheckBox() {
        return geneShiftsCheckBox;
    }


    /**
     * Returns the label for number of shifts for CCC-Biclustering
     * and e-CCC-Biclustering algorithms.
     *
     * @return the <code>JLabel</code> for number of shifts for
     * CCC-Biclustering and e-CCC-Biclustering algorithm
     */
    public JLabel getShiftsLabel() {
        return shiftsLabel;
    }

    /**
     * Returns the text field for the input of CCC-Biclustering
     * and e-CCC-Biclustering algorithms number of shifts.
     *
     * @return the <code>JTextField</code> for the input of
     *         CCC-Biclustering and e-CCC-Biclustering algorithms
     *         number of shifts
     */
    public JTextField getShiftsTextField() {
        return shiftsTextField;
    }

    /**
     * Returns restricted errors for e-CCC-Biclustering algorithm check box.
     *
     * @return the restricted errors for e-CCC-Biclustering algorithm
     *         <code>JCheckBox</code>
     */
    public JCheckBox getRestrictedErrorsECCCCheckBox() {
        return restrictedErrorsECCCCheckBox;
    }

    /**
     * Label for the number of neighbors on e-CCC-Biclustering algorithm.
     *
     * @return the <code>JLabel</code> for the number of neighbors on
     *         e-CCC-Biclustering algorithm
     */
    public JLabel getNeighborsECCCLabel() {
        return neighborsECCCLabel;
    }

    /**
     * Text field for the input of number of neighbors for
     * e-CCC-Biclustering algorithm.
     *
     * @return the <code>JTextField</code> for the input of number
     *         of neighbors for e-CCC-Biclustering algorithm
     */
    public JTextField getNeighborsECCCTextField() {
        return neighborsECCCTextField;
    }

    /**
     * Label for the number of errors allowed in e-CCC-Biclustering algorithm.
     *
     * @return the <code>JLabel</code> for the number of errors allowed
     *         in e-CCC-Biclustering algorithm
     */
    public JLabel getErrorsNumberLabel() {
        return errorsNumberLabel;
    }

    /**
     * Text field for the input of the number of errors allowed in
     * e-CCC-Biclustering algorithm.
     *
     * @return the <code>JTextField</code> for the input of the number
     *         of errors allowed in e-CCC-Biclustering algorithm
     */
    public JTextField getErrorsNumberTextField() {
        return errorsNumberTextField;
    }

    /**
     * Returns biclustering panel.
     *
     * @return JPanel biclustering panel
     */
    public JPanel getBiclusteringPanel() {
        return this.biclusteringAlgorithmPanel;
    }

    /**
     * Returns apply button.
     *
     * @return apply <code>JButton</code>
     */
    public JButton getApplyButton() {
        return applyButton;
    }

    /**
     * Handles a selection change in algorithms' list.
     *
     * @param e <code>ListSelectionEvent</code>
     */
    public void algorithmsList_valueChanged(ListSelectionEvent e) {
        if (!e.getValueIsAdjusting()) {
            this.biclusteringAlgorithmPanel.setLayout(new SpringLayout());
            this.biclusteringAlgorithmPanel.removeAll();
            this.biclusteringAlgorithmPanel.repaint();
            String algorithm = (String) algorithmsList.getSelectedValue();
            if (algorithm == null) {
                return;
            }
            if (algorithm.equals(CCC_BICLUSTERING)) {
                this.paintBiclusteringAlgorithmPanelWithCCC();
                mainFrame.setVisible(true);
            } else
            if (algorithm.equals(E_CCC_BICLUSTERING)) {
                this.paintBiclusteringAlgorithmPanelWithECCC();
                mainFrame.setVisible(true);
            } else
            if (algorithm.equals(CC_TSB_BICLUSTERING)) {
                this.paintBiclusteringAlgorithmPanelWithCCTSB();
                mainFrame.setVisible(true);
            } else {
                JOptionPane.showMessageDialog(this, "Select a biclustering " +
                                              "algorithm to apply",
                                              "No algorithm selected",
                                              JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }

    /**
     * Handles a change in gene shifts check box state.
     *
     * @param e <code>ItemEvent</code>
     */
    public void geneShiftsCheckBox_itemStateChanged(ItemEvent e) {
        if (this.geneShiftsCheckBox.isSelected()) {
            this.signChangesCheckBox.setSelected(false);
            this.timeLagsCheckBox.setSelected(false);
        }
        this.setGeneShiftsCheckBoxSelected(this.geneShiftsCheckBox.isSelected());
    }

    public void signChangesCheckBox_itemStateChanged(ItemEvent e) {
        if (this.signChangesCheckBox.isSelected()) {
            this.geneShiftsCheckBox.setSelected(false);
        }
    }

    public void timeLagsCheckBox_itemStateChanged(ItemEvent e) {
        if (this.timeLagsCheckBox.isSelected()) {
            this.geneShiftsCheckBox.setSelected(false);
        }
        this.maxTimeLagLabel.setEnabled(this.timeLagsCheckBox.isSelected());
        this.maxTimeLagTextField.setEnabled(this.timeLagsCheckBox.isSelected());
    }

    public void allowMissingsButton_mouseClicked(MouseEvent e) {
        if (this.allowMissingsButton.isSelected() &&
            this.restrictedErrorsECCCCheckBox.isSelected()) {
            this.restrictedErrorsECCCCheckBox.setSelected(false);
        }
    }
}


class PanelBiclustering_allowMissingsButton_mouseAdapter extends MouseAdapter {
    private PanelBiclustering adaptee;
    PanelBiclustering_allowMissingsButton_mouseAdapter(PanelBiclustering
            adaptee) {
        this.adaptee = adaptee;
    }

    public void mouseClicked(MouseEvent e) {
        adaptee.allowMissingsButton_mouseClicked(e);
    }
}


class PanelBiclustering_timeLagsCheckBox_itemAdapter implements ItemListener {
    private PanelBiclustering adaptee;
    PanelBiclustering_timeLagsCheckBox_itemAdapter(PanelBiclustering adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.timeLagsCheckBox_itemStateChanged(e);
    }
}


class PanelBiclustering_signChangesCheckBox_itemAdapter implements ItemListener {
    private PanelBiclustering adaptee;
    PanelBiclustering_signChangesCheckBox_itemAdapter(PanelBiclustering adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.signChangesCheckBox_itemStateChanged(e);
    }
}


/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 */
class PanelBiclustering_algorithmsList_listSelectionAdapter implements
        ListSelectionListener {
    private PanelBiclustering adaptee;
    PanelBiclustering_algorithmsList_listSelectionAdapter(PanelBiclustering
            adaptee) {
        this.adaptee = adaptee;
    }

    public void valueChanged(ListSelectionEvent e) {
        adaptee.algorithmsList_valueChanged(e);
    }
}


/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 */
class PanelBiclustering_geneShiftsCheckBox_itemAdapter implements ItemListener {
    private PanelBiclustering adaptee;
    PanelBiclustering_geneShiftsCheckBox_itemAdapter(PanelBiclustering adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.geneShiftsCheckBox_itemStateChanged(e);
    }
}


/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 */
class PanelBiclustering_restrictedErrorsECCCCheckBox_itemAdapter implements
        ItemListener {
    private PanelBiclustering adaptee;
    PanelBiclustering_restrictedErrorsECCCCheckBox_itemAdapter(
            PanelBiclustering adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e <code>ItemEvent</code>
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.restrictedErrorsECCCCheckBox_itemStateChanged(e);
    }
}


class PanelBiclustering_applyButton_actionAdapter implements ActionListener {
    private PanelBiclustering adaptee;
    PanelBiclustering_applyButton_actionAdapter(PanelBiclustering adaptee) {
        this.adaptee = adaptee;
    }

    public void actionPerformed(ActionEvent e) {
        adaptee.applyButton_actionPerformed(e);
    }
}
