package biggests.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import javax.swing.*;
import javax.swing.tree.*;

import biggests.tree.*;
import biggests.utils.*;
import de.schlichtherle.io.File;
import smadeira.biclustering.*;
import smadeira.ontologizer.GOFrontEnd;

/**
 * <p>Title: Function Analysis Panel</p>
 *
 * <p>Description: This panel displays function analysis options to
 *                 apply to a given bicluster or group of biclusters.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
public class PanelFunctionAnalysis extends JPanel {
    public static final String OBO_FILE_NAME = "gene_ontology.1_0.obo";

    /**
     * Set of organisms to display in organism combo box.
     */
    public static final String[] organisms =
            {
            "Anaplasma phagocytophilum HZ",
            "Agrobacterium tumefaciensstr. C58",
            "Arabidopsis thaliana",
            "Aspergillus nidulans",
            "Bacillus anthracis Ames",
            "Bos taurus",
            "Carboxydothermus hydrogenoformans Z-2901",
            "Caenorhabditis elegans",
            "Campylobacter jejuni",
            "Candida albicans",
            "Clostridium perfringens",
            "Colwellia psychrerythraea 34H",
            "Coxiella burnetii RSA 493",
            "Danio rerio",
            "Dehalococcoides ethenogenes 195",
            "Dictyostelium discoideum",
            "Drosophila melanogaster",
            "Ehrlichia chaffeensis Arkansas",
            "Escherichia coli",
            "Gallus gallus",
            "Geobacter sulfurreducens PCA",
            "Homo sapiens",
            "Hyphomonas neptunium ATCC 15444",
            "Leishmania major",
            "Listeria monocytogenes 4b F2365",
            "Magnaporthe grisea",
            "Methylococcus capsulatus Bath",
            "Mus musculus",
            "Neorickettsia sennetsu Miyayama",
            "Oomycetes",
            "Oryza sativa",
            "Plasmodium falciparum",
            "Pseudomonas aeruginosa PAO1",
            "Pseudomonas fluorescens Pf-5",
            "Pseudomonas syringae DC3000",
            "Pseudomonas syringae pv. phaseolicola 1448A",
            "Rattus norvegicus",
            "Saccharomyces cerevisiae",
            "Schizosaccharomyces pombe",
            "Shewanella oneidensis MR-1",
            "Silicibacter pomeroyi DSS-3",
            "Solanaceae",
            "Trypanosoma brucei",
            "Vibrio cholerae"
    };

    /**
     * Annotation file extension.
     * File extension at annotationFileExtensions[i] corresponds to
     * organism at organisms[i].
     */
    private static final String[] annotationFileExtensions =
            {
            ".tigr_Aphagocytophilum",
            ".PAMGO_Atumefaciens",
            ".tair",
            ".aspgd",
            ".tigr_Banthracis",
            ".goa_cow",
            ".tigr_Chydrogenoformans",
            ".wb",
            ".tigr_Cjejuni",
            ".cgd",
            ".tigr_Cperfringens",
            ".tigr_Cpsychrerythraea",
            ".tigr_Cburnetii",
            ".zfin",
            ".tigr_Dethenogenes",
            ".dictyBase",
            ".fb",
            ".tigr_Echaffeensis",
            ".ecocyc",
            ".goa_chicken",
            ".tigr_Gsulfurreducens",
            ".goa_human",
            ".tigr_Hneptunium",
            ".GeneDB_Lmajor",
            ".tigr_Lmonocytogenes",
            ".PAMGO_Mgrisea",
            ".tigr_Mcapsulatus",
            ".mgi",
            ".tigr_Nsennetsu",
            ".PAMGO_Oomycetes",
            ".gramene_oryza",
            ".GeneDB_Pfalciparum",
            ".pseudocap",
            ".tigr_Pfluorescens",
            ".tigr_Psyringae",
            ".tigr_Psyringae_phaseolicola",
            ".rgd",
            ".sgd",
            ".GeneDB_Spombe",
            ".tigr_Soneidensis",
            ".tigr_Spomeroyi",
            ".sgn",
            ".GeneDB_Tbrucei",
            ".tigr_Vcholerae"
    };

    /**
     * Application's main frame.
     */
    private BiggestsFrame mainFrame;

    /**
     * The selection path.
     */
    private TreePath selectionPath;

    /**
     * The organism related with this function analysis.
     */
    private String organism;

    /**
     * Function analysis label.
     */
    private JLabel functionAnalysisLabel;

    /**
     * P-value label.
     */
    private JLabel pvalueLabel;

    /**
     * P-value text field.
     */
    private JTextField pvalueTextField;

    /**
     * Organism label.
     */
    private JLabel organismLabel;

    /**
     * Files to use on function analysis label.
     */
    private JLabel filesFunctionAnalysisLabel;

    /**
     * Gene ontology file path label.
     */
    private JLabel geneOntologyFilePathLabel;

    /**
     * Gene ontology file path text field.
     */
    private JTextField geneOntologyFilePathTextField;

    /**
     * Browse gene ontology file path button.
     */
    private JButton browseGeneOntologyFilePathButton;

    /**
     * Gene association file path label.
     */
    private JLabel geneAssociationFilePathLabel;

    /**
     * Gene association file path text field.
     */
    private JTextField geneAssociationFilePathTextField;

    /**
     * Browse gene association file path button.
     */
    private JButton browseGeneAssociationFilePathButton;

    /**
     * Download gene ontology file button.
     */
    private JButton downloadGeneOntologyFileButton;

    /**
     * Download gene association file button.
     */
    private JButton downloadGeneAssociationFileButton;

    /**
     * The layout for this panel.
     */
    private SpringLayout springLayout;

    /**
     * GO frontend check box.
     */
    private JCheckBox goFrontEndCheckBox;

    /**
     * Creates a new <code>PanelFunctionAnalysis</code>.
     *
     * @param frame the main <code>BiggestsFrame</code>
     */
    public PanelFunctionAnalysis(BiggestsFrame frame) {
        this.mainFrame = frame;
        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Component initialization.
     *
     * @throws Exception if components could not be initialized
     */
    private void jbInit() throws Exception {
        springLayout = new SpringLayout();
        this.setPreferredSize(new Dimension(450, 350));
        this.setLayout(springLayout);

        this.functionAnalysisLabel = new JLabel("Function Analysis");
        Font font = this.functionAnalysisLabel.getFont();
        this.functionAnalysisLabel.setFont(new Font(font.getName(), Font.BOLD,
                font.getSize() + 2));
        this.pvalueLabel = new JLabel("P-Value");
        this.pvalueTextField = new JTextField(5);
        this.pvalueTextField.setText("0.01");
        this.organismLabel = new JLabel("");
    }

    /**
     * Clears this panel. Removes components that are not needed and sets
     * others to default values.
     */
    public void clear() {
        this.selectionPath = null;
        this.removeAll();
        this.pvalueTextField.setText("0.01");
    }

    /**
     * Sets this function analysis panel enabled/disabled according to the
     * gene identifier of the bicluster group matrices.
     */
    private void setFunctionAnalysisPanelEnabled() {
        boolean enabled = true;
        if (!((NodeInfo) ((DefaultMutableTreeNode)this.selectionPath.
                          getLastPathComponent()).getUserObject()).isGeneID()) {
            enabled = false;
        }
        Component[] components = this.getComponents();
        for (Component comp : components) {
            comp.setEnabled(enabled);
        }
    }

    /**
     * Paints the panel with information from the given
     * <code>selectionPath</code>.
     *
     * @param selectionPath the <code>TreePath</code> selected in a dataset tree
     */
    public void paint(TreePath selectionPath) {
        boolean includeGOFrontEnd =
            mainFrame.
            getPanelAnalyzingFunctionTable().
            getGOFrontEndForSelectionPath(selectionPath) != null;
        if (includeGOFrontEnd) {
            this.goFrontEndCheckBox = new JCheckBox("Use available GO frontend", true);
            goFrontEndCheckBox.addItemListener(
                    new
                    PanelFunctionAnalysis_goFrontEndCheckBox_itemAdapter(this));
        }
        this.filesFunctionAnalysisLabel = new JLabel(
                "GO files to use in Function Analysis");

        this.geneOntologyFilePathLabel = new JLabel("Gene Ontology File");
        this.geneOntologyFilePathTextField = new JTextField("");
        this.geneOntologyFilePathTextField.setPreferredSize(new Dimension(350,
                22));
        this.browseGeneOntologyFilePathButton = new JButton("Browse...");
        this.browseGeneOntologyFilePathButton.setPreferredSize(new Dimension(
                170, 22));
        browseGeneOntologyFilePathButton.addActionListener(new
                PanelFunctionAnalysis_browseGOFilesPathButton_actionAdapter(this));
        this.downloadGeneOntologyFileButton = new JButton("Download");
        this.downloadGeneOntologyFileButton.setPreferredSize(new Dimension(170,
                22));
        downloadGeneOntologyFileButton.addActionListener(new
                PanelFunctionAnalysis_downloadGeneOntologyFileButton_actionAdapter(this));

        this.geneAssociationFilePathLabel = new JLabel("Gene Association File");
        this.geneAssociationFilePathTextField = new JTextField("");
        this.geneAssociationFilePathTextField.setPreferredSize(new Dimension(
                350, 22));
        this.browseGeneAssociationFilePathButton = new JButton("Browse...");
        this.browseGeneAssociationFilePathButton.setPreferredSize(new Dimension(
                170, 22));
        browseGeneAssociationFilePathButton.addActionListener(new
                PanelFunctionAnalysis_browseGeneAssociationFilesPathButton_actionAdapter(this));
        this.downloadGeneAssociationFileButton = new JButton("Download");
        this.downloadGeneAssociationFileButton.setPreferredSize(new Dimension(
                170, 22));
        downloadGeneAssociationFileButton.addActionListener(new
                PanelFunctionAnalysis_downloadGeneAssociationFileButton_actionAdapter(this));

        this.add(this.functionAnalysisLabel);

        this.add(this.pvalueLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.pvalueLabel,
                                   0,
                                   SpringLayout.WEST,
                                   this.functionAnalysisLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.pvalueLabel,
                                   30,
                                   SpringLayout.SOUTH,
                                   this.functionAnalysisLabel);

        this.add(this.pvalueTextField);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.pvalueTextField,
                                   10,
                                   SpringLayout.EAST,
                                   this.pvalueLabel);
        springLayout.putConstraint(SpringLayout.SOUTH,
                                   this.pvalueTextField,
                                   0,
                                   SpringLayout.SOUTH,
                                   this.pvalueLabel);

        this.add(this.organismLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.organismLabel,
                                   0,
                                   SpringLayout.WEST,
                                   this.functionAnalysisLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.organismLabel,
                                   10,
                                   SpringLayout.SOUTH,
                                   this.pvalueLabel);

        int x = 20;
        if (includeGOFrontEnd) {
            this.add(this.goFrontEndCheckBox);
            springLayout.putConstraint(SpringLayout.WEST,
                                       this.goFrontEndCheckBox,
                                       -2,
                                       SpringLayout.WEST,
                                       this.organismLabel);
            springLayout.putConstraint(SpringLayout.NORTH,
                                       this.goFrontEndCheckBox,
                                       20,
                                       SpringLayout.SOUTH,
                                       this.organismLabel);
            x += 40;
        }

        this.add(this.filesFunctionAnalysisLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.filesFunctionAnalysisLabel,
                                   0,
                                   SpringLayout.WEST,
                                   this.organismLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.filesFunctionAnalysisLabel,
                                   x,
                                   SpringLayout.SOUTH,
                                   this.organismLabel);

        this.add(this.geneOntologyFilePathLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.geneOntologyFilePathLabel,
                                   20,
                                   SpringLayout.WEST,
                                   this.filesFunctionAnalysisLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.geneOntologyFilePathLabel,
                                   5,
                                   SpringLayout.SOUTH,
                                   this.filesFunctionAnalysisLabel);

        this.add(this.geneOntologyFilePathTextField);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.geneOntologyFilePathTextField,
                                   0,
                                   SpringLayout.WEST,
                                   this.geneOntologyFilePathLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.geneOntologyFilePathTextField,
                                   5,
                                   SpringLayout.SOUTH,
                                   this.geneOntologyFilePathLabel);

        this.add(this.browseGeneOntologyFilePathButton);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.browseGeneOntologyFilePathButton,
                                   0,
                                   SpringLayout.WEST,
                                   this.geneOntologyFilePathTextField);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.browseGeneOntologyFilePathButton,
                                   5,
                                   SpringLayout.SOUTH,
                                   this.geneOntologyFilePathTextField);

        this.add(this.downloadGeneOntologyFileButton);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.downloadGeneOntologyFileButton,
                                   10,
                                   SpringLayout.EAST,
                                   this.browseGeneOntologyFilePathButton);
        springLayout.putConstraint(SpringLayout.SOUTH,
                                   this.downloadGeneOntologyFileButton,
                                   0,
                                   SpringLayout.SOUTH,
                                   this.browseGeneOntologyFilePathButton);

        this.add(this.geneAssociationFilePathLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.geneAssociationFilePathLabel,
                                   0,
                                   SpringLayout.WEST,
                                   this.browseGeneOntologyFilePathButton);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.geneAssociationFilePathLabel,
                                   10,
                                   SpringLayout.SOUTH,
                                   this.browseGeneOntologyFilePathButton);

        this.add(this.geneAssociationFilePathTextField);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.geneAssociationFilePathTextField,
                                   0,
                                   SpringLayout.WEST,
                                   this.geneAssociationFilePathLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.geneAssociationFilePathTextField,
                                   5,
                                   SpringLayout.SOUTH,
                                   this.geneAssociationFilePathLabel);

        this.add(this.browseGeneAssociationFilePathButton);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.browseGeneAssociationFilePathButton,
                                   0,
                                   SpringLayout.WEST,
                                   this.geneAssociationFilePathTextField);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   this.browseGeneAssociationFilePathButton,
                                   5,
                                   SpringLayout.SOUTH,
                                   this.geneAssociationFilePathTextField);

        this.add(this.downloadGeneAssociationFileButton);
        springLayout.putConstraint(SpringLayout.WEST,
                                   this.downloadGeneAssociationFileButton,
                                   10,
                                   SpringLayout.EAST,
                                   this.browseGeneAssociationFilePathButton);
        springLayout.putConstraint(SpringLayout.SOUTH,
                                   this.downloadGeneAssociationFileButton,
                                   0,
                                   SpringLayout.SOUTH,
                                   this.browseGeneAssociationFilePathButton);

        this.selectionPath = selectionPath;
        NodeInfo nodeInfo = ((NodeInfo)
                             ((DefaultMutableTreeNode)
                              ((DefaultMutableTreeNode)this.selectionPath.
                               getPath()[1])
                              .getChildAt(0))
                             .getUserObject());

        this.organism = ((ExpressionMatrix) nodeInfo.getNodeObject()).
                        getOrganism();
        this.organismLabel.setText("Organism: " + organism);
        try {
            String pathToGeneOntologyFile = BiggestsIO.findFileInDirectoryPath(
                    PanelFunctionAnalysis.OBO_FILE_NAME,
                    mainFrame.getPathToGO_Files());
            String pathToGeneAssociationFile = BiggestsIO.
                                               findFileInDirectoryPath(
                    findGeneAssociationFileName(organism, true),
                    mainFrame.getPathToGO_Files());

            if (pathToGeneOntologyFile == null) {
                this.geneOntologyFilePathTextField.setText("");
            } else {
                this.geneOntologyFilePathTextField.setText(
                        pathToGeneOntologyFile);
            }
            if (pathToGeneAssociationFile == null) {
                this.geneAssociationFilePathTextField.setText("");
            } else {
                this.geneAssociationFilePathTextField.setText(
                        pathToGeneAssociationFile);
            }
        } catch (IOException e) {
            e.printStackTrace();
        } catch (Exception e) {
            e.printStackTrace();
            this.geneOntologyFilePathTextField.setText("");
        }
        this.setFunctionAnalysisPanelEnabled();
        if (includeGOFrontEnd) {
            for (int i = 5; i < this.getComponentCount(); i++) {
                this.getComponent(i).setEnabled(false);
            }
        }
    }

    /**
     * Finds a specific filename or extension for a given organism.
     *
     * @param organism <code>String</code> with the organism to find a
     *                 gene association file for
     * @param filename_extension <code>boolean</code> <code>true</code> if
     *                 the String to return is the filename of the
     *                 corresponding organism gene association file;
     *                 <code>false</code> if the String to return is the
     *                 file extension of the corresponding organism gene
     *                 association file
     * @return <code>String</code> the filename/extension of the corresponding
     *                 organism gene association file
     */
    public static String findGeneAssociationFileName(String organism,
            boolean filename_extension) {
        String fileExtension = findAnnotationFileExtensionForOrganism(organism);
        if (filename_extension) {
            return "gene_association" + fileExtension;
        }
        return fileExtension;
    }

    /**
     * Finds and returns the corresponding annotation file extension for
     * a given organism.
     *
     * @param organism <code>String</code> the name of the organism to find
     *                 the annotation file extension for
     * @return <code>String</code> the annotation file extension for the
     *         given organism
     */
    public static String findAnnotationFileExtensionForOrganism(String organism) {
        int i = 0;
        for (String org : organisms) {
            if (organism.equals(org)) {
                return annotationFileExtensions[i];
            }
            i++;
        }
        return null;
    }

    /**
     * Checks the input of the p-value on <code>pvalueTextField</code>
     *
     * @return a valid p-value <code>float</code> entered by the user
     */
    public float checkPValue() {
        float pvalue = -1000;
        String parameter = this.pvalueTextField.getText();
        if ((parameter == null) ||
            (parameter.equals(""))) {
            JOptionPane.showMessageDialog(this, "Please type a p-value.",
                                          "Type p-value",
                                          JOptionPane.
                                          INFORMATION_MESSAGE);
            return -1000;
        }
        try {
            pvalue = Float.parseFloat(parameter);

        } catch (NumberFormatException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this,
                                          "The p-value must be a floating " +
                                          "point value. ",
                                          "P-value must be floating point.",
                                          JOptionPane.
                                          INFORMATION_MESSAGE);
            return -1000;
        }
        if (pvalue < (float)0 || pvalue > (float)1.0) {
            JOptionPane.showMessageDialog(this,
                                          "The p-value must be a floating " +
                                          "point >=0 and <=1. ",
                                          "P-value must be in [0,1].",
                                          JOptionPane.
                                          INFORMATION_MESSAGE);

            return -1000;
        }
        return pvalue;
    }

    /**
     * Checks the existence of the required gene ontology files (the
     * gene ontology file and the specific gene association file for
     * the specified organism) in the GO Files directory. In case there
     * is some error displays a message to user and asks if he/she wants
     * to correct it.<br/>
     * <p>
     * In case the gene ontology files are correct, returns:<br/>
     * File[0] - containing the gene ontology file.<br/>
     * File[1] - containing the gene association file.
     * </p>
     * <p>
     * Returns <code>null</code> if there was some error on the input
     * (p-value or file paths are not valid) and the user wants to ignore it.
     * </p>
     * <p>
     * Returns <code>File[0]</code> if there was some error on the input
     * (p-value or file paths are not valid) and the user wants to correct it.
     * </p>
     *
     * @param biclustering_function <code>boolean</code> that indicates if
     *        this panel is being used in biclustering (true) or in
     *        function panel (false)
     * @return the <code>File[]</code> list of files
     */
    public File[] checkGOFiles(boolean biclustering_function) {
        String pathToGOFile = this.geneOntologyFilePathTextField.getText();
        String pathToGeneAssociationFile = this.
                                           geneAssociationFilePathTextField.
                                           getText();

        int invalid = 0;
        if (pathToGOFile == null || pathToGOFile.equals("")) {
            invalid = 1;
        }

        if (pathToGeneAssociationFile == null ||
            pathToGeneAssociationFile.equals("")) {
            invalid += 2;
        }

        if (invalid != 0) {
            return this.showInvalidGOFilesPathsMessage(invalid,
                    biclustering_function);
        }

        invalid = 0;
        if (!pathToGOFile.toLowerCase().endsWith(".obo")) {
            invalid = 1;
        }

        if (!pathToGeneAssociationFile.toLowerCase().endsWith(
                findGeneAssociationFileName(organism, false))) {
            invalid += 2;
        }

        if (invalid != 0) {
            return this.showInvalidGOFilesPathsMessage(invalid,
                    biclustering_function);
        }

        invalid = 0;
        File goFile = new File(pathToGOFile);
        if (!goFile.exists() || !goFile.isFile()) {
            invalid = 1;
        }
        File geneAssociationFile = new File(pathToGeneAssociationFile);
        if (!geneAssociationFile.exists() || !geneAssociationFile.isFile()) {
            invalid += 2;
        }

        if (invalid != 0) {
            return this.showInvalidGOFilesPathsMessage(invalid,
                    biclustering_function);
        }

        File[] files = {goFile, geneAssociationFile};
        return files;
    }

    /**
     * Check if GO frontend can be used for function analysis.
     * Returns null if GO frontend does not exist for this dataset or if
     * GO frontend is not selected by the user. Returns the correct GOFrontEnd
     * object if GO frontend should be used.
     *
     * @return GOFrontEnd
     */
    public GOFrontEnd useGOFrontEnd() {
        if (goFrontEndCheckBox == null) {
            return null;
        }
        if (goFrontEndCheckBox.isSelected()) {
            return mainFrame.getPanelAnalyzingFunctionTable().
                    getGOFrontEndForSelectionPath(selectionPath);
        }
        return null;
    }

    /**
     * Displays a message to user indicating which filenames are invalid
     * and asks him/her if he/she wants to correct it/them.<br/>
     * Returns <code>null</code> if user wants to ignore the error.<br/>
     * Returns <code>File[0]</code> if user wants to correct the error.<br/>
     *
     * @param invalid <code>int</code> signaling which file(s) name(s)
     *        is/are incorrect<br/>
     *        1 - the first one<br/>
     *        2 - the second one<br/>
     *        3 - both<br/>
     * @param biclustering_function <code>boolean</code> that indicates if
     *        this panel is being used in biclustering (true) or in
     *        function panel (false)
     * @return the <code>File[]</code> list of files (see description
     *         of return values)
     */
    private File[] showInvalidGOFilesPathsMessage(int invalid,
                                                  boolean biclustering_function) {
        String message = "";
        switch (invalid) {
        case 1:
            message =
                    "The path to gene ontology file is not valid. Correct it?";
            break;
        case 2:
            message =
                    "The path to gene association file is not valid. Correct it?";
            break;
        case 3:
            message = "The path to gene ontology and gene association files are not valid. Correct them?";
            break;
        }

        if (biclustering_function) {
            message += "\n\nIf you ignore this, only biclustering algorithm " +
                    "will be performed.\nYou may perform function analysis " +
                    "later on, selecting the group\nof biclusters and " +
                    "choosing [Analyzing -> Function -> Table] option.";
        }

        String[] options = {"Correct", "Ignore"};
        int option = JOptionPane.showOptionDialog(this,
                                                  message,
                                                  "Invalid file path",
                                                  JOptionPane.OK_CANCEL_OPTION,
                                                  JOptionPane.QUESTION_MESSAGE, null,
                                                  options,
                                                  options[1]);
        if (option == 1) {
            return null;
        }

        return new File[0];
    }

    /**
     * Handles event fired when the user clicks browse gene
     * ontology file path button.
     *
     * @param e <code>ActionEvent</code> fired when the user clicks
     *          browse gene ontology file path button
     */
    public void browseGOFilesPathButton_actionPerformed(ActionEvent e) {
        if (!this.browseGeneOntologyFilePathButton.isEnabled()) {
            return;
        }

        JFileChooser datasetFileChooser = new JFileChooser(mainFrame.
                getPathToGO_Files());
        try {
            datasetFileChooser.setFileFilter(new FileFilterExtension(".obo"));
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        datasetFileChooser.setAcceptAllFileFilterUsed(false);
        int option = datasetFileChooser.showOpenDialog(this);

        if (option == JFileChooser.APPROVE_OPTION) {
            this.geneOntologyFilePathTextField.setText(datasetFileChooser.
                    getSelectedFile().getPath());
        }
    }

    /**
     * Handles event fired when the user clicks on browse gene
     * association file path button.
     *
     * @param e <code>ActionEvent</code> fired when the user clicks
     *          on browse gene association file path button
     */
    public void browseGeneAssociationFilesPathButton_actionPerformed(ActionEvent e) {
        if (!this.browseGeneAssociationFilePathButton.isEnabled()) {
            return;
        }

        JFileChooser datasetFileChooser = new JFileChooser(mainFrame.
                getPathToGO_Files());
        try {
            datasetFileChooser.setFileFilter(new FileFilterExtension(this.
                    findGeneAssociationFileName(this.organism, false)));
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        datasetFileChooser.setAcceptAllFileFilterUsed(false);
        int option = datasetFileChooser.showOpenDialog(this);

        if (option == JFileChooser.APPROVE_OPTION) {
            this.geneAssociationFilePathTextField.setText(datasetFileChooser.
                    getSelectedFile().getPath());
        }
    }

    /**
     * Returns the application's main frame.
     *
     * @return the main <code>BiggestsFrame</code>
     */
    public BiggestsFrame getMainFrame() {
        return mainFrame;
    }

    /**
     * Returns the dataset tree selected path which corresponds to the path of
     * the dataset which in turn leads to the node with the matrix on which
     * biclustering algorithm will be performed.
     *
     * @return the <code>TreePath</code> selected in the dataset tree
     */
    public TreePath getSelectionPath() {
        return selectionPath;
    }

    /**
     * Returns the organism related with this function analysis.
     *
     * @return <code>String</code> organism
     */
    public String getOrganism() {
        return organism;
    }

    /**
     * Returns function analysis label.
     *
     * @return function analysis <code>JLabel</code>
     */
    public JLabel getFunctionAnalysisLabel() {
        return functionAnalysisLabel;
    }

    /**
     * Returns p-value label.
     *
     * @return p-value <code>JLabel</code>
     */
    public JLabel getPvalueLabel() {
        return pvalueLabel;
    }

    /**
     * Returns text field for the input of a p-value.
     *
     * @return <code>JTextField</code> for the input of a p-value
     */
    public JTextField getPvalueTextField() {
        return pvalueTextField;
    }

    /**
     * Returns the organism label.
     *
     * @return organism <code>JLabel</code>
     */
    public JLabel getOrganismLabel() {
        return organismLabel;
    }

    /**
     * Returns label of files to use on function analysis.
     *
     * @return code>JLabel</code> of files to use on function analysis
     */
    public JLabel getFilesFunctionAnalysisLabel() {
        return filesFunctionAnalysisLabel;
    }

    /**
     * Returns gene ontology file path label.
     *
     * @return gene ontology file path <code>JLabel</code>
     */
    public JLabel getGeneOntologyFilePathLabel() {
        return geneOntologyFilePathLabel;
    }

    /**
     * Returns gene ontology file path text field.
     *
     * @return gene ontology file path <code>JTextField</code>
     */
    public JTextField getGeneOntologyFilePathTextField() {
        return geneOntologyFilePathTextField;
    }

    /**
     * Returns browse gene ontology file path button.
     *
     * @return browse gene ontology file path <code>JButton</code>
     */
    public JButton getBrowseGeneOntologyFilePathButton() {
        return browseGeneOntologyFilePathButton;
    }

    /**
     * Returns gene association file path label.
     *
     * @return gene association file path <code>JLabel</code>
     */
    public JLabel getGeneAssociationFilePathLabel() {
        return geneAssociationFilePathLabel;
    }

    /**
     * Returns gene association file path text field.
     *
     * @return gene association file path <code>JTextField</code>
     */
    public JTextField getGeneAssociationFilePathTextField() {
        return geneAssociationFilePathTextField;
    }

    /**
     * Returns browse gene association file path button.
     *
     * @return browse gene association file path <code>JButton</code>
     */
    public JButton getBrowseGeneAssociationFilePathButton() {
        return browseGeneAssociationFilePathButton;
    }

    /**
     * Returns the layout for this panel.
     *
     * @return <code>SpringLayout</code>
     */
    public SpringLayout springLayout() {
        return springLayout;
    }

    /**
     * Handles a mouse click event on download gene ontology file button.
     * Downloads the gene ontology file to GO_Files folder.
     *
     * @param e ActionEvent
     */
    public void downloadGeneOntologyFileButton_actionPerformed(ActionEvent e) {
        if (!downloadGeneOntologyFileButton.isEnabled()) {
            return;
        }
        ThreadDownloadGOFiles downloadThread = new ThreadDownloadGOFiles(
                this.mainFrame,
                this,
                PanelFunctionAnalysis.OBO_FILE_NAME);
    }

    /**
     * Handles a mouse click event on download gene association file button.
     * Downloads the gene association file to GO_Files folder.
     *
     * @param e ActionEvent
     */
    public void downloadGeneAssociationFileButton_actionPerformed(ActionEvent e) {
        if (!downloadGeneAssociationFileButton.isEnabled()) {
            return;
        }
        ThreadDownloadGOFiles downloadThread = new ThreadDownloadGOFiles(
                this.mainFrame, this,
                this.findGeneAssociationFileName(this.organism, true));
    }

    public void goFrontEndCheckBox_itemStateChanged(ItemEvent e) {
        boolean enabled = !this.goFrontEndCheckBox.isSelected();
        for (int i = 5; i < this.getComponentCount(); i++) {
            this.getComponent(i).setEnabled(enabled);
        }
    }
}


class PanelFunctionAnalysis_goFrontEndCheckBox_itemAdapter implements
        ItemListener {
    private PanelFunctionAnalysis adaptee;
    PanelFunctionAnalysis_goFrontEndCheckBox_itemAdapter(PanelFunctionAnalysis
            adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.goFrontEndCheckBox_itemStateChanged(e);
    }
 }

class PanelFunctionAnalysis_downloadGeneAssociationFileButton_actionAdapter implements
        ActionListener {
    private PanelFunctionAnalysis adaptee;
    PanelFunctionAnalysis_downloadGeneAssociationFileButton_actionAdapter(
            PanelFunctionAnalysis adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ActionEvent
     */
    public void actionPerformed(ActionEvent e) {
        adaptee.downloadGeneAssociationFileButton_actionPerformed(e);
    }
}


class PanelFunctionAnalysis_downloadGeneOntologyFileButton_actionAdapter implements
        ActionListener {
    private PanelFunctionAnalysis adaptee;
    PanelFunctionAnalysis_downloadGeneOntologyFileButton_actionAdapter(
            PanelFunctionAnalysis adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ActionEvent
     */
    public void actionPerformed(ActionEvent e) {
        adaptee.downloadGeneOntologyFileButton_actionPerformed(e);
    }
}


class PanelFunctionAnalysis_browseGeneAssociationFilesPathButton_actionAdapter implements
        ActionListener {
    private PanelFunctionAnalysis adaptee;
    PanelFunctionAnalysis_browseGeneAssociationFilesPathButton_actionAdapter(
            PanelFunctionAnalysis adaptee) {
        this.adaptee = adaptee;
    }

    /**
     * Associates the specific handler of mouse clicks over browse
     * gene association file path button with the overriden method
     * <code>mouseClicked</code>.
     *
     * @param e <code>ActionEvent</code> fired when the user clicks
     *          some graphic component
     */
    public void actionPerformed(ActionEvent e) {
        adaptee.browseGeneAssociationFilesPathButton_actionPerformed(e);
    }
}


/**
 *
 * <p>Title: </p>
 *
 * <p>Description: </p>
 */
class PanelFunctionAnalysis_browseGOFilesPathButton_actionAdapter implements
        ActionListener {
    private PanelFunctionAnalysis adaptee;
    PanelFunctionAnalysis_browseGOFilesPathButton_actionAdapter(
            PanelFunctionAnalysis adaptee) {
        this.adaptee = adaptee;
    }

    /**
     * Associates the specific handler of mouse clicks over browse
     * gene ontology file path button with the overriden method
     * <code>mouseClicked</code>.
     *
     * @param e <code>ActionEvent</code> fired when the user clicks
     *          some graphic component
     */
    public void actionPerformed(ActionEvent e) {
        adaptee.browseGOFilesPathButton_actionPerformed(e);
    }
}
