package biggests.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.tree.*;

import biggests.tree.*;
import biggests.utils.*;
import smadeira.biclustering.*;

/**
 * <p>Title: Preprocessing Panel</p>
 *
 * <p>Description: Displays options to select and apply different
 *                 preprocessing techniques on gene expression
 *                 data matrices.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
public class PanelPreProcessing extends PanelBasicIOValidation {
    /**
     * 'Percentage of genes to filter' parameter name.
     */
    public static final String FILTER_PERCENTAGE =
            "percentage of genes to filter";

    /**
     * 'Value to fill missings' parameter name;
     */
    public static final String FILL_VALUE = "value to fill missings";

    /**
     * 'Number of neighbors' parameter name.
     */
    public static final String FILL_NEIGHBORS = "number of neighbors";

    /**
     * 'Mean value to normalize expression data' parameter name.
     */
    public static final String NORMALIZE_MEAN =
            "mean value to normalize expression data";

    /**
     * 'Std value to normalize expression data' parameter name.
     */
    public static final String NORMALIZE_STD =
            "std value to normalize expression data";

    /**
     * 'Smoothing window' parameter name.
     */
    public static final String SMOOTH_WINDOW = "smoothing window";

    /**
     * 'Percentage for discretization' parameter name.
     */
    public static final String DISCRETIZE_PERCENT =
            "percentage for discretization";

    /**
     * 'Number of symbols' parameter name.
     */
    public static final String DISCRETIZE_SYMBOLS = "number of symbols";

    /**
     * 'Alpha' discretization parameter name.
     */
    public static final String DISCRETIZE_ALPHA = "alpha value";

    /**
     * 'Threshold' discretization parameter name.
     */
    public static final String DISCRETIZE_THRESHOLD = "threshold";

    /**
     * Discretization techniques available;
     */
    private static final String[] discretizeAlgorithms = {"expression average",
            "gene expression average",
            "mid-range",
            "gene mid-range",
            "max-minus percent-max",
            "gene max-minus percent-max",
            "equal frequency",
            "equal frequency by gene",
            "equal width",
            "equal width by gene",
            "expression mean and std",
            "gene expression mean and std",
            "transitional state discrimination (2 symbols)",
            "variation between time points (2 symbols)",
            "variation between time points (3 symbols)"};

    /**
     * BiGGEsTS main window.
     */
    private BiggestsFrame mainFrame;

    /**
     * The original or preprocessed expression matrix in which
     * preprocessing techniques will be performed.
     */
    private AbstractExpressionMatrix expressionMatrix;

    /**
     * The path selected on the dataset tree. Corresponds to the path in
     * the tree from the root node to the node that contains the original
     * or preprocessed matrix to apply preprocessing techniques.
     */
    private TreePath selectionPath;

    /**
     * The layout of this panel.
     */
    private SpringLayout springLayout;

    /**
     * Filter check box.
     */
    private JCheckBox filterCheckBox;

    /**
     * Filter options button group.
     */
    private ButtonGroup filterButtonGroup;

    /**
     * Filter all genes with missing values button.
     */
    private JRadioButton filterAllGenesButton;

    /**
     * Filter genes with more than a % of missing values button.
     */
    private JRadioButton filterPercentGenesButton;

    /**
     * Filter percent input text field.
     */
    private JTextField filterPercentTextField;

    /**
     * Filter percent label.
     */
    private JLabel filterPercentLabel;

    /**
     * Fill check box.
     */
    private JCheckBox fillCheckBox;

    /**
     * Fill options button group.
     */
    private ButtonGroup fillButtonGroup;

    /**
     * Fill with a certain value button.
     */
    private JRadioButton fillValueButton;

    /**
     * Fill with average over the k-neighbors button.
     */
    private JRadioButton fillKNeighborsButton;

    /**
     * Fill with average over the same row button.
     */
    private JRadioButton fillAverageRowButton;

    /**
     * Fill with a certain value input text field.
     */
    private JTextField fillValueTextField;

    /**
     * Fill with average over the k-neighbors input text field.
     */
    private JTextField fillKNeighborsTextField;

    /**
     * Fill with average over the k-neighbors label.
     */
    private JLabel fillKNeighborsLabel;

    /**
     * Normalize check box.
     */
    private JCheckBox normalizeCheckBox;

    /**
     * Normalize options button group.
     */
    private ButtonGroup normalizeButtonGroup;

    /**
     * Normalize overall values button.
     */
    private JRadioButton normalizeOverallButton;

    /**
     * Normalize by gene button.
     */
    private JRadioButton normalizeByGeneButton;

    /**
     * Normalize to mean 0 and std 1 button.
     */
    private JRadioButton normalizeMean0Std1Button;

    /**
     * Normalize by mean and std values button.
     */
    private JRadioButton normalizeMeanStdValuesButton;

    /**
     * Normalize mean input text field.
     */
    private JTextField normalizeMeanValueTextField;

    /**
     * Normalize std input text field.
     */
    private JTextField normalizeStdValueTextField;

    /**
     * Normalize to mean and std values label.
     */
    private JLabel normalizeMeanStdValuesLabel;

    /**
     * Normalize to mean and std options button group.
     */
    private ButtonGroup normalizeMeanStdButtonGroup;

    /**
     * Smooth check box.
     */
    private JCheckBox smoothCheckBox;

    /**
     * Smooth window label.
     */
    private JLabel smoothWindowLabel;

    /**
     * Smooth window input text field.
     */
    private JTextField smoothWindowTextField;

    /**
     * Discretize check box.
     */
    private JCheckBox discretizeCheckBox;

    /**
     * Discretize label.
     */
    private JLabel discretizeLabel;

    /**
     * Discretize combo box.
     */
    private JComboBox discretizeComboBox;

    /**
     * Discretize parameter input text field.
     */
    private JTextField discretizeParameterTextField;

    /**
     * Discretize parameter label.
     */
    private JLabel discretizeParameterLabel;

    /**
     * Apply preprocessing options button.
     */
    private JButton applyButton;

    /**
     * Creates a new preprocessing panel.
     *
     * @param frame the main window
     */
    public PanelPreProcessing(BiggestsFrame frame) {
        mainFrame = frame;
        expressionMatrix = null;
        selectionPath = null;
        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Component initialization.
     *
     * @throws Exception if components cannot be initialized
     */
    private void jbInit() throws Exception {
        this.setPreferredSize(new Dimension(550, 700));
        this.springLayout = new SpringLayout();
        this.setLayout(springLayout);
        this.setBorder(BorderFactory.createEmptyBorder(50, 50, 50, 50));

        this.filterCheckBox = new JCheckBox("Filter");
        this.filterAllGenesButton = new JRadioButton(
                "all genes with missing values");
        this.filterPercentGenesButton = new JRadioButton("genes with more than");
        this.filterPercentTextField = new JTextField(3);
        this.filterPercentLabel = new JLabel("% of missing values");
        this.filterButtonGroup = new ButtonGroup();
        filterPercentGenesButton.addItemListener(
            new
            PanelPreProcessing_filterPercentGenesButton_itemAdapter(this));
        filterCheckBox.addItemListener(new
                PanelPreProcessing_filterCheckBox_itemAdapter(this));
        this.filterButtonGroup.add(this.filterAllGenesButton);
        this.filterButtonGroup.add(this.filterPercentGenesButton);

        this.add(filterCheckBox);
        this.add(filterAllGenesButton);
        springLayout.putConstraint(SpringLayout.NORTH, filterAllGenesButton, 5,
                                   SpringLayout.SOUTH, filterCheckBox);
        springLayout.putConstraint(SpringLayout.WEST, filterAllGenesButton, 20,
                                   SpringLayout.WEST, filterCheckBox);

        this.add(filterPercentGenesButton);
        springLayout.putConstraint(SpringLayout.NORTH, filterPercentGenesButton,
                                   5,
                                   SpringLayout.SOUTH, filterAllGenesButton);
        springLayout.putConstraint(SpringLayout.WEST, filterPercentGenesButton,
                                   20,
                                   SpringLayout.WEST, filterCheckBox);

        this.add(filterPercentTextField);
        springLayout.putConstraint(SpringLayout.WEST, filterPercentTextField, 5,
                                   SpringLayout.EAST, filterPercentGenesButton);
        springLayout.putConstraint(SpringLayout.SOUTH, filterPercentTextField,
                                   -4,
                                   SpringLayout.SOUTH, filterPercentGenesButton);

        this.add(filterPercentLabel);
        springLayout.putConstraint(SpringLayout.WEST, filterPercentLabel, 5,
                                   SpringLayout.EAST, filterPercentTextField);
        springLayout.putConstraint(SpringLayout.SOUTH, filterPercentLabel, -5,
                                   SpringLayout.SOUTH, filterPercentGenesButton);

        this.fillCheckBox = new JCheckBox("Fill missings", true);
        fillCheckBox.addItemListener(new
                PanelPreProcessing_fillCheckBox_itemAdapter(this));
        this.fillValueButton = new JRadioButton("with value");
        this.fillValueButton.addItemListener(
            new
            PanelPreProcessing_fillValueButton_itemAdapter(this));
        this.fillValueTextField = new JTextField(3);
        this.fillKNeighborsButton = new JRadioButton(
                "with the average over the K-neighbors");
        this.fillKNeighborsButton.addItemListener(
            new
            PanelPreProcessing_fillKNeighborsButton_itemAdapter(this));
        this.fillKNeighborsTextField = new JTextField(3);
        this.fillKNeighborsLabel = new JLabel("(k value)");
        this.fillAverageRowButton = new JRadioButton(
                "with the average over the same row");
        this.fillAverageRowButton.addItemListener(
            new
            PanelPreProcessing_fillAverageRowButton_itemAdapter(this));
        this.fillButtonGroup = new ButtonGroup();
        this.fillButtonGroup.add(this.fillValueButton);
        this.fillButtonGroup.add(this.fillKNeighborsButton);
        this.fillButtonGroup.add(this.fillAverageRowButton);

        this.add(fillCheckBox);
        springLayout.putConstraint(SpringLayout.WEST, fillCheckBox, 0,
                                   SpringLayout.WEST, filterCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH, fillCheckBox, 30,
                                   SpringLayout.SOUTH, filterPercentGenesButton);

        this.add(fillValueButton);
        springLayout.putConstraint(SpringLayout.WEST, fillValueButton, 20,
                                   SpringLayout.WEST, fillCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH, fillValueButton, 5,
                                   SpringLayout.SOUTH, fillCheckBox);

        this.add(fillValueTextField);
        springLayout.putConstraint(SpringLayout.WEST, fillValueTextField, 5,
                                   SpringLayout.EAST, fillValueButton);
        springLayout.putConstraint(SpringLayout.SOUTH, fillValueTextField, -4,
                                   SpringLayout.SOUTH, fillValueButton);

        this.add(fillKNeighborsButton);
        springLayout.putConstraint(SpringLayout.WEST, fillKNeighborsButton, 20,
                                   SpringLayout.WEST, fillCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH, fillKNeighborsButton, 5,
                                   SpringLayout.SOUTH, fillValueButton);

        this.add(fillKNeighborsTextField);
        springLayout.putConstraint(SpringLayout.WEST, fillKNeighborsTextField,
                                   5,
                                   SpringLayout.EAST, fillKNeighborsButton);
        springLayout.putConstraint(SpringLayout.SOUTH, fillKNeighborsTextField,
                                   -4,
                                   SpringLayout.SOUTH, fillKNeighborsButton);

        this.add(fillKNeighborsLabel);
        springLayout.putConstraint(SpringLayout.WEST, fillKNeighborsLabel, 5,
                                   SpringLayout.EAST, fillKNeighborsTextField);
        springLayout.putConstraint(SpringLayout.SOUTH, fillKNeighborsLabel, -5,
                                   SpringLayout.SOUTH, fillKNeighborsButton);

        this.add(fillAverageRowButton);
        this.springLayout.putConstraint(SpringLayout.WEST, fillAverageRowButton,
                                        20,
                                        SpringLayout.WEST, fillCheckBox);
        this.springLayout.putConstraint(SpringLayout.NORTH,
                                        fillAverageRowButton, 5,
                                        SpringLayout.SOUTH,
                                        fillKNeighborsButton);

        this.normalizeCheckBox = new JCheckBox("Normalize");
        this.normalizeCheckBox.addItemListener(new
                                               PanelPreProcessing_normalizeCheckBox_itemAdapter(this));
        this.normalizeOverallButton = new JRadioButton("by overall values");
        this.normalizeOverallButton.addItemListener(new
                PanelPreProcessing_normalizeOverallButton_itemAdapter(this));
        this.normalizeByGeneButton = new JRadioButton("by gene");
        this.normalizeByGeneButton.addItemListener(new
                PanelPreProcessing_normalizeByGeneButton_itemAdapter(this));
        this.normalizeMean0Std1Button = new JRadioButton("to mean 0 and std 1");
        this.normalizeMean0Std1Button.addItemListener(new
                PanelPreProcessing_normalizeMean0Std1Button_itemAdapter(this));
        this.normalizeMeanStdValuesButton = new JRadioButton("to mean");
        this.normalizeMeanStdValuesButton.addItemListener(new
                PanelPreProcessing_normalizeMeanStdValuesButton_itemAdapter(this));
        this.normalizeMeanValueTextField = new JTextField(3);
        this.normalizeMeanStdValuesLabel = new JLabel("and std");
        this.normalizeStdValueTextField = new JTextField(3);
        this.normalizeButtonGroup = new ButtonGroup();
        this.normalizeButtonGroup.add(this.normalizeOverallButton);
        this.normalizeButtonGroup.add(this.normalizeByGeneButton);
        this.normalizeMeanStdButtonGroup = new ButtonGroup();
        this.normalizeMeanStdButtonGroup.add(this.normalizeMean0Std1Button);
        this.normalizeMeanStdButtonGroup.add(this.normalizeMeanStdValuesButton);

        this.add(normalizeCheckBox);
        this.springLayout.putConstraint(SpringLayout.NORTH, normalizeCheckBox,
                                        30,
                                        SpringLayout.SOUTH,
                                        fillAverageRowButton);
        this.springLayout.putConstraint(SpringLayout.WEST, normalizeCheckBox, 0,
                                        SpringLayout.WEST, fillCheckBox);

        this.add(normalizeOverallButton);
        this.springLayout.putConstraint(SpringLayout.WEST,
                                        normalizeOverallButton, 20,
                                        SpringLayout.WEST, normalizeCheckBox);
        this.springLayout.putConstraint(SpringLayout.NORTH,
                                        normalizeOverallButton, 5,
                                        SpringLayout.SOUTH, normalizeCheckBox);

        this.add(normalizeMean0Std1Button);
        springLayout.putConstraint(SpringLayout.WEST, normalizeMean0Std1Button,
                                   20,
                                   SpringLayout.EAST, normalizeOverallButton);
        springLayout.putConstraint(SpringLayout.SOUTH, normalizeMean0Std1Button,
                                   0,
                                   SpringLayout.SOUTH, normalizeOverallButton);

        this.add(normalizeByGeneButton);
        springLayout.putConstraint(SpringLayout.WEST, normalizeByGeneButton, 20,
                                   SpringLayout.WEST, normalizeCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH, normalizeByGeneButton, 5,
                                   SpringLayout.SOUTH, normalizeOverallButton);

        this.add(normalizeMeanStdValuesButton);
        springLayout.putConstraint(SpringLayout.WEST,
                                   normalizeMeanStdValuesButton, 0,
                                   SpringLayout.WEST, normalizeMean0Std1Button);
        springLayout.putConstraint(SpringLayout.SOUTH,
                                   normalizeMeanStdValuesButton, 0,
                                   SpringLayout.SOUTH, normalizeByGeneButton);

        this.add(normalizeMeanValueTextField);
        springLayout.putConstraint(SpringLayout.WEST,
                                   normalizeMeanValueTextField, 5,
                                   SpringLayout.EAST,
                                   normalizeMeanStdValuesButton);
        springLayout.putConstraint(SpringLayout.SOUTH,
                                   normalizeMeanValueTextField, -4,
                                   SpringLayout.SOUTH,
                                   normalizeMeanStdValuesButton);

        this.add(normalizeMeanStdValuesLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   normalizeMeanStdValuesLabel, 5,
                                   SpringLayout.EAST,
                                   normalizeMeanValueTextField);
        springLayout.putConstraint(SpringLayout.SOUTH,
                                   normalizeMeanStdValuesLabel, -5,
                                   SpringLayout.SOUTH,
                                   normalizeMeanStdValuesButton);

        this.add(normalizeStdValueTextField);
        springLayout.putConstraint(SpringLayout.WEST,
                                   normalizeStdValueTextField, 5,
                                   SpringLayout.EAST,
                                   normalizeMeanStdValuesLabel);
        springLayout.putConstraint(SpringLayout.SOUTH,
                                   normalizeStdValueTextField, -4,
                                   SpringLayout.SOUTH,
                                   normalizeMeanStdValuesButton);

        this.smoothCheckBox = new JCheckBox("Smooth");
        smoothCheckBox.addItemListener(new
                                       PanelPreProcessing_smoothCheckBox_itemAdapter(this));
        this.smoothWindowLabel = new JLabel("with window");
        this.smoothWindowTextField = new JTextField(3);

        this.add(smoothCheckBox);
        springLayout.putConstraint(SpringLayout.WEST, smoothCheckBox, 0,
                                   SpringLayout.WEST, normalizeCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH, smoothCheckBox, 30,
                                   SpringLayout.SOUTH,
                                   normalizeMeanStdValuesButton);

        this.add(smoothWindowLabel);
        springLayout.putConstraint(SpringLayout.WEST, smoothWindowLabel, 20,
                                   SpringLayout.EAST, smoothCheckBox);
        springLayout.putConstraint(SpringLayout.SOUTH, smoothWindowLabel, -4,
                                   SpringLayout.SOUTH, smoothCheckBox);

        this.add(smoothWindowTextField);
        springLayout.putConstraint(SpringLayout.WEST, smoothWindowTextField, 5,
                                   SpringLayout.EAST, smoothWindowLabel);
        springLayout.putConstraint(SpringLayout.SOUTH, smoothWindowTextField, 0,
                                   SpringLayout.SOUTH, smoothWindowLabel);

        this.discretizeCheckBox = new JCheckBox("Discretize");
        discretizeCheckBox.addItemListener(new
                                           PanelPreProcessing_discretizeCheckBox_itemAdapter(this));
        this.discretizeLabel = new JLabel("Discretization Technique");
        this.discretizeComboBox = new JComboBox(discretizeAlgorithms);
        this.discretizeComboBox.setMaximumRowCount(20);
        discretizeComboBox.addItemListener(new
                                           PanelPreProcessing_discretizeComboBox_itemAdapter(this));
        this.discretizeParameterTextField = new JTextField(3);
        this.discretizeParameterLabel = new JLabel("(discretization parameter)");

        this.add(discretizeCheckBox);
        springLayout.putConstraint(SpringLayout.WEST, discretizeCheckBox, 0,
                                   SpringLayout.WEST, smoothCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH, discretizeCheckBox, 30,
                                   SpringLayout.SOUTH, smoothCheckBox);

        this.add(discretizeLabel);
        springLayout.putConstraint(SpringLayout.WEST, discretizeLabel, 20,
                                   SpringLayout.WEST, discretizeCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH, discretizeLabel, 5,
                                   SpringLayout.SOUTH, discretizeCheckBox);

        this.add(discretizeComboBox);
        springLayout.putConstraint(SpringLayout.WEST, discretizeComboBox, 5,
                                   SpringLayout.EAST, discretizeLabel);
        springLayout.putConstraint(SpringLayout.SOUTH, discretizeComboBox, 0,
                                   SpringLayout.SOUTH, discretizeLabel);

        this.add(discretizeParameterTextField);
        springLayout.putConstraint(SpringLayout.WEST,
                                   discretizeParameterTextField, 0,
                                   SpringLayout.WEST, discretizeComboBox);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   discretizeParameterTextField, 5,
                                   SpringLayout.SOUTH, discretizeComboBox);

        this.add(discretizeParameterLabel);
        springLayout.putConstraint(SpringLayout.WEST, discretizeParameterLabel,
                                   5,
                                   SpringLayout.EAST,
                                   discretizeParameterTextField);
        springLayout.putConstraint(SpringLayout.SOUTH, discretizeParameterLabel,
                                   0,
                                   SpringLayout.SOUTH,
                                   discretizeParameterTextField);

        this.applyButton = new JButton("Apply");
        applyButton.setPreferredSize(new Dimension(450, 50));

        applyButton.addActionListener(new
                PanelPreProcessing_applyButton_actionAdapter(this));
        this.add(applyButton);
        springLayout.putConstraint(SpringLayout.WEST, applyButton, 0,
                                   SpringLayout.WEST, discretizeCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH, applyButton, 60,
                                   SpringLayout.SOUTH, discretizeComboBox);
    }

    /**
     * Paints this preprocessing panel with the preprocessing options
     * related with a given <code>expressionMatrix</code>.
     *
     * @param expressionMatrix the <code>ExpressionMatrix</code> in which
     *        preprocessing options will be performed
     * @param selectionPath the <code>TreePath</code> selected in the
     *        dataset tree; corresponds to the path from the root node to
     *        the node that contains the <code>expressionMatrix</code>
     */
    public void paint(ExpressionMatrix expressionMatrix, TreePath selectionPath) {
        this.expressionMatrix = expressionMatrix;
        this.selectionPath = selectionPath;

        this.fillValueTextField.setText("-100");
        this.fillKNeighborsTextField.setText("4");
        this.setNormalizeCheckBoxSelected(true);
        this.setSmoothCheckBoxSelected(false);
        this.setDiscretizeCheckBoxSelected(true);

        if (this.expressionMatrix.hasMissingValues()) {
            this.setFilterCheckBoxSelected(true);
            this.setFilterPercentGenesButtonSelected(false);
            this.fillCheckBox.setEnabled(true);
            this.setFillCheckBoxSelected(false);
        }
        else {
            this.setFilterCheckBoxSelected(false);
            this.setFilterPercentGenesButtonSelected(false);
            this.setFillCheckBoxSelected(false);
            this.fillCheckBox.setEnabled(false);
        }
    }

    /**
     * Paints this preprocessing panel with the preprocessing options
     * related with a given <code>expressionMatrix</code>.
     *
     * @param preProcessedExpMat the <code>PreProcessedExpressionMatrix</code>
     *        in which preprocessing options will be performed
     * @param selectionPath the <code>TreePath</code> selected in the
     *        dataset tree; corresponds to the path from the root node to
     *        the node that contains the <code>preProcessedExpMat</code>
     */
    public void paint(PreProcessedExpressionMatrix preProcessedExpMat,
                      TreePath selectionPath) {
        this.expressionMatrix = preProcessedExpMat;
        this.selectionPath = selectionPath;

//        this.setFilterCheckBoxSelected(false);
//        this.setFilterPercentGenesButtonSelected(false);
        this.fillValueTextField.setText("-100");
        this.fillKNeighborsTextField.setText("4");
        this.setNormalizeCheckBoxSelected(true);
        this.setSmoothCheckBoxSelected(false);
        this.setDiscretizeCheckBoxSelected(true);

        if (this.expressionMatrix.hasMissingValues()) {
            this.setFilterCheckBoxSelected(true);
            this.setFilterPercentGenesButtonSelected(false);
            this.fillCheckBox.setEnabled(true);
            this.setFillCheckBoxSelected(false);
        } else {
            this.setFilterCheckBoxSelected(false);
            this.setFilterPercentGenesButtonSelected(false);
            this.setFillCheckBoxSelected(false);
            this.fillCheckBox.setEnabled(false);
        }

    }

    /**
     * Sets the filter check box selected/deselected, enabling/disabling
     * the corresponding filter options.
     *
     * @param selected <code>boolean</code> <code>true</code> if the
     *        filter check box is to select; <code>false</code> otherwise
     */
    public void setFilterCheckBoxSelected(boolean selected) {
        this.filterCheckBox.setSelected(selected);
//        this.filterCheckBox.setEnabled(false);
        this.filterAllGenesButton.setEnabled(selected);
        this.filterPercentGenesButton.setEnabled(selected);
        this.filterPercentLabel.setEnabled(selected);
        this.filterPercentTextField.setEnabled(selected);

        if (selected) {
            this.filterAllGenesButton.setSelected(true);
            this.filterPercentTextField.setText("20");
        }
    }

    /**
     * Sets the filter genes with more than a percent of missing values
     * check box selected/deselected, enabling/disabling
     * the corresponding filter options.
     *
     * @param selected <code>boolean</code> <code>true</code> if the
     *        filter check box is to select; <code>false</code> otherwise
     */
    private void setFilterPercentGenesButtonSelected(boolean selected) {
        this.filterPercentTextField.setEnabled(selected);
        this.setFillCheckBoxSelected(selected);
    }

    private void setFillCheckBoxSelected(boolean selected) {
        this.fillCheckBox.setSelected(selected);
        this.fillValueButton.setEnabled(selected);
        this.fillValueTextField.setEnabled(false);
        this.fillKNeighborsButton.setEnabled(selected);
        this.fillKNeighborsLabel.setEnabled(selected);
        this.fillKNeighborsTextField.setEnabled(false);
        this.fillAverageRowButton.setEnabled(selected);

        if (selected) {
            this.fillAverageRowButton.setSelected(true);
            this.fillValueTextField.setText("-100");
            this.fillKNeighborsTextField.setText("4");
        }
    }

    /**
     * Sets the normalize check box selected/deselected, enabling/disabling
     * the corresponding normalize options.
     *
     * @param selected <code>boolean</code> <code>true</code> if the
     *        normalize check box is to select; <code>false</code> otherwise
     */
    public void setNormalizeCheckBoxSelected(boolean selected) {
        normalizeCheckBox.setSelected(selected);
        normalizeOverallButton.setEnabled(selected);
        normalizeByGeneButton.setEnabled(selected);
        normalizeMean0Std1Button.setEnabled(selected);
        normalizeMeanStdValuesButton.setEnabled(selected);
        normalizeMeanValueTextField.setEnabled(false);
        normalizeMeanStdValuesLabel.setEnabled(selected);
        normalizeStdValueTextField.setEnabled(false);

        if (selected) {
            normalizeByGeneButton.setSelected(true);
            normalizeMean0Std1Button.setSelected(true);
            normalizeMeanValueTextField.setText("0");
            normalizeStdValueTextField.setText("1");
        }
    }

    /**
     * Sets the smooth check box selected/deselected, enabling/disabling
     * the corresponding smooth options.
     *
     * @param selected <code>boolean</code> <code>true</code> if the
     *        smooth check box is to select; <code>false</code> otherwise
     */
    public void setSmoothCheckBoxSelected(boolean selected) {
        this.smoothCheckBox.setSelected(selected);
        this.smoothWindowLabel.setEnabled(selected);
        this.smoothWindowTextField.setEnabled(selected);

        if (selected) {
            smoothWindowTextField.setText("3");
        }
    }

    /**
     * Sets the discretize check box selected/deselected, enabling/disabling
     * the corresponding discretize options.
     *
     * @param selected <code>boolean</code> <code>true</code> if the
     *        discretize check box is to select; <code>false</code> otherwise
     */
    public void setDiscretizeCheckBoxSelected(boolean selected) {
        this.discretizeCheckBox.setSelected(selected);
        this.discretizeComboBox.setEnabled(selected);
        this.discretizeLabel.setEnabled(selected);
        this.discretizeParameterLabel.setEnabled(selected);
        this.discretizeParameterTextField.setEnabled(selected);
        this.discretizeParameterLabel.setText("(discretization parameter)");
        this.discretizeParameterTextField.setText("");

        if (selected) {
            this.discretizeComboBox.setSelectedIndex(14);
            this.discretizeParameterTextField.setText("1");
            this.discretizeParameterLabel.setText("(threshold)");
        }
    }

    /**
     * Reads the parameter string for its input parameter text field.
     *
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return the parameter <code>String</code> read from the
     *         discretize parameter text field
     * @throws Exception if there is no parameter written in the text field
     */
    public String readParameter(String parameterToRead) throws Exception {
        String parameter;
        if (parameterToRead.equals(PanelPreProcessing.FILTER_PERCENTAGE)) {
            parameter = this.filterPercentTextField.getText();
        } else
        if (parameterToRead.equals(PanelPreProcessing.FILL_VALUE)) {
            parameter = this.fillValueTextField.getText();
        } else
        if (parameterToRead.equals(PanelPreProcessing.FILL_NEIGHBORS)) {
            parameter = this.fillKNeighborsTextField.getText();
        } else
        if (parameterToRead.equals(PanelPreProcessing.NORMALIZE_MEAN)) {
            parameter = this.normalizeMeanValueTextField.getText();
        } else
        if (parameterToRead.equals(PanelPreProcessing.NORMALIZE_STD)) {
            parameter = this.normalizeStdValueTextField.getText();
        } else
        if (parameterToRead.equals(PanelPreProcessing.SMOOTH_WINDOW)) {
            parameter = this.smoothWindowTextField.getText();
        } else {
            parameter = this.discretizeParameterTextField.getText();
        }
        if ((parameter == null) ||
            (parameter.equals(""))) {
            JOptionPane.showMessageDialog(this, "Please type "
                                          + parameterToRead + ".",
                                          "Type " + parameterToRead,
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Shows the discretization error message.
     *
     * @param algorithmName <code>String</code> the name of the
     *        discretization algorithm that was applied
     */
    public void showDiscretizeErrorMessage(String algorithmName) {
        JOptionPane.showMessageDialog(this, "Unable to compute " +
                                      algorithmName + "discretization" +
                                      " algorithm", "Unable to discretize",
                                      JOptionPane.ERROR_MESSAGE);
    }

    /**
     * Handles the item state changed event fired when the user selects
     * the filter genes with more than a % of missing values radio button.
     *
     * @param e <code>ItemEvent</code> fired when the user selects
     * the filter genes with more than a % of missing values radio button
     */
    public void filterPercentGenesButton_itemStateChanged(ItemEvent e) {
        setFilterPercentGenesButtonSelected(filterPercentGenesButton.isSelected());
    }

    /**
     * Handles the item state changed event fired when the user selects
     * the normalize overall values radio button.
     *
     * @param e <code>ItemEvent</code> fired when the user selects
     * the normalize overall values radio button
     */
    public void normalizeOverallButton_itemStateChanged(ItemEvent e) {
        if (this.normalizeOverallButton.isSelected()) {
            this.normalizeMean0Std1Button.setEnabled(true);
            this.normalizeMean0Std1Button.setSelected(true);
            this.normalizeMeanStdValuesButton.setEnabled(true);
            this.normalizeMeanValueTextField.setEnabled(false);
            this.normalizeStdValueTextField.setEnabled(false);
            this.normalizeMeanStdValuesLabel.setEnabled(true);
        }
    }

    /**
     * Handles the item state changed event fired when the user selects
     * the normalize by gene radio button.
     *
     * @param e <code>ItemEvent</code> fired when the user selects
     * the normalize by gene radio button
     */
    public void normalizeByGeneButton_itemStateChanged(ItemEvent e) {
        if (this.normalizeByGeneButton.isSelected()) {
            this.normalizeMean0Std1Button.setEnabled(true);
            this.normalizeMean0Std1Button.setSelected(true);
            this.normalizeMeanStdValuesButton.setEnabled(false);
            this.normalizeMeanValueTextField.setEnabled(false);
            this.normalizeStdValueTextField.setEnabled(false);
            this.normalizeMeanStdValuesLabel.setEnabled(false);
        }
    }

    /**
     * Handles the item state changed event fired when the user
     * selects/deselects the normalize check box.
     *
     * @param e <code>ItemEvent</code> fired when the user selects/deselects
     *        the normalize option check box
     */
    public void normalizeCheckBox_itemStateChanged(ItemEvent e) {
        setNormalizeCheckBoxSelected(this.normalizeCheckBox.isSelected());
    }

    /**
     * Handles the item state changed event fired when the user
     * selects/deselects the smooth check box.
     *
     * @param e <code>ItemEvent</code> fired when the user selects/deselects
     *        the smooth option check box
     */
    public void smoothCheckBox_itemStateChanged(ItemEvent e) {
        setSmoothCheckBoxSelected(this.smoothCheckBox.isSelected());
    }

    /**
     * Handles the item state changed event fired when the user
     * selects/deselects the discretize check box.
     *
     * @param e <code>ItemEvent</code> fired when the user selects/deselects
     *        the discretize option check box
     */
    public void discretizeCheckBox_itemStateChanged(ItemEvent e) {
        setDiscretizeCheckBoxSelected(this.discretizeCheckBox.isSelected());
    }

    /**
     * Handles the item state changed event fired when the user
     * changes the selection in the discretize combo box.
     *
     * @param e <code>ItemEvent</code> fired when the user changes
     *        the selection in the discretize combo box
     */
    public void discretizeComboBox_itemStateChanged(ItemEvent e) {
        int selectedIndex = this.discretizeComboBox.getSelectedIndex();
        if ((selectedIndex == 0) || (selectedIndex == 1) ||
            (selectedIndex == 2) || (selectedIndex == 3) ||
            (selectedIndex == 12)) {
            this.discretizeParameterLabel.setEnabled(false);
            this.discretizeParameterTextField.setEnabled(false);
            this.discretizeParameterLabel.setText("(discretization parameter)");
            this.discretizeParameterTextField.setText("");
            return;
        }
        this.discretizeParameterLabel.setEnabled(true);
        this.discretizeParameterTextField.setEnabled(true);
        if ((selectedIndex == 4) || (selectedIndex == 5)) {
            this.discretizeParameterTextField.setText("20");
            this.discretizeParameterLabel.setText("(%)");
            return;
        }
        if ((selectedIndex == 6) || (selectedIndex == 7) ||
            (selectedIndex == 8) || (selectedIndex == 9)) {
            this.discretizeParameterTextField.setText("2");
            this.discretizeParameterLabel.setText("(number of symbols)");
            return;
        }
        if ((selectedIndex == 10) || (selectedIndex == 11) ||
            (selectedIndex == 13)) {
            this.discretizeParameterTextField.setText("0.5");
            this.discretizeParameterLabel.setText("(alpha)");
            return;
        }
        if (selectedIndex == 14) {
            this.discretizeParameterTextField.setText("1");
            this.discretizeParameterLabel.setText("(threshold)");
            return;
        }
    }

    /**
     * Handles the mouse event fired when the user clicks on the apply button.
     *
     * @param e <code>ActionEvent</code> fired when the user clicks on
     *        the apply button
     */
    public void applyButton_actionPerformed(ActionEvent e) {
        ArrayList<String> infoStrings = new ArrayList<String>();
        ArrayList<String> infoStyles = new ArrayList<String>();

        try {
            if ((this.expressionMatrix == null) || (this.selectionPath == null)) {
                throw (new Exception("An expression matrix and a selection " +
                                     "path are needed to perform preprocessing"));
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        boolean wasPreProcessed = false;

        PreProcessedExpressionMatrix preprocessed =
                new PreProcessedExpressionMatrix(this.expressionMatrix);
//        if (this.expressionMatrix instanceof ExpressionMatrix) {
            if (preprocessed.hasMissingValues()) {
                if (this.filterCheckBox.isSelected()) {
                    if (this.filterPercentGenesButton.isSelected()) {
                        float percentage = 100;
                        try {
                            percentage = this.readParameterFloat(1, 100,
                                    PanelPreProcessing.FILTER_PERCENTAGE);
                        } catch (Exception ex) {
                            ex.printStackTrace();
                            return;
                        }

                        preprocessed.
                                removeGenesWithMoreThanAPercentageOfMissingValues(
                                        percentage);
                        infoStrings.add("   Gene filtering: ");
                        infoStyles.add("bold");
                        infoStrings.add("genes with more than " + percentage +
                                        "% of missings");
                        infoStyles.add("regular");
                        wasPreProcessed = true;

                        if (this.fillValueButton.isSelected()) {
                            float value = -10000;
                            try {
                                value = this.readParameterFloat(
                                        PanelPreProcessing.
                                        FILL_VALUE);
                            } catch (Exception ex) {
                                ex.printStackTrace();
                                return;
                            }

                            preprocessed.fillMissingValuesWithValue(value);
                            infoStrings.add("   Missing values filling: ");
                            infoStrings.add("filled with value " + value +
                                            "\n");
                            infoStyles.add("bold");
                            infoStyles.add("regular");
                            wasPreProcessed = true;
                        } else
                        if (this.fillKNeighborsButton.isSelected()) {
                            int k = 4;

                            try {
                                k = this.readParameterInteger(
                                        PanelPreProcessing.
                                        FILL_NEIGHBORS);
                            } catch (Exception ex) {
                                ex.printStackTrace();
                                return;
                            }

                            preprocessed.
                                    fillMissingsNeighborsAverageGeneExpression(
                                            k);
                            infoStrings.add("   Missing values filling: ");
                            infoStrings.add("average over the " + k +
                                            " neighbors\n");
                            infoStyles.add("bold");
                            infoStyles.add("regular");
                            wasPreProcessed = true;
                        } else {
                            preprocessed.fillMissingsAverageGeneExpression();
                            infoStrings.add("   Missing values filling: ");
                            infoStrings.add("average over the same row\n");
                            infoStyles.add("bold");
                            infoStyles.add("regular");
                            wasPreProcessed = true;
                        }
                    } else {
                        /**
                         * filter all genes with missing values
                         */
                        preprocessed.removeGenesWithMissingValues();
                        infoStrings.add("   Gene filtering: ");
                        infoStrings.add("all genes with missing values\n");
                        infoStyles.add("bold");
                        infoStyles.add("regular");
                        wasPreProcessed = true;
                    }
                }
            }
            /*else {
                /**
              * expression matrix does not have missing values; don't have
              * to check filter and fill options
              */
             //        }
//        }
        /**
         * check and process normalize, smooth and discretize options
         */
        if (this.normalizeCheckBox.isSelected()) {
            if (this.normalizeOverallButton.isSelected()) {
                if (this.normalizeMean0Std1Button.isSelected()) {
                    try {
                        preprocessed.normalizeGeneExpressionMatrix();
                        infoStrings.add("   Normalization: ");
                        infoStrings.add("all data - mean 0, std 1\n");
                        infoStyles.add("bold");
                        infoStyles.add("regular");
                        wasPreProcessed = true;
                    } catch (Exception ex) {
                        ex.printStackTrace();
                        JOptionPane.showMessageDialog(this,
                                "Unable to compute " +
                                "normalization.",
                                "Unable to normalize",
                                JOptionPane.ERROR_MESSAGE);
                    }
                } else {
                    /**
                     * selected normalize option: by overall values - to
                     * specified mean and std values
                     */
                    float mean = (float) 0.0;
                    float std = (float) 1.0;

                    try {
                        mean = this.readParameterInteger(PanelPreProcessing.
                                NORMALIZE_MEAN);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                        return;
                    }

                    try {
                        std = this.readParameterInteger(PanelPreProcessing.
                                NORMALIZE_STD);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                        return;
                    }

                    try {
                        preprocessed.normalizeGeneExpressionMatrix(mean, std);
                        infoStrings.add("   Normalization: ");
                        infoStrings.add("all data - mean " + mean +
                                               ", std " + std + "\n");
                        infoStyles.add("bold");
                        infoStyles.add("regular");
                        wasPreProcessed = true;
                    } catch (Exception ex) {
                        ex.printStackTrace();
                        JOptionPane.showMessageDialog(this,
                                "Unable to compute " +
                                "normalization.",
                                "Unable to normalize",
                                JOptionPane.ERROR_MESSAGE);
                    }
                }
            } else {
                /**
                 * selected normalize option: normalize by gene - to mean 0
                 * and std 1
                 */
                try {
                    preprocessed.normalizeGeneExpressionMatrixByGene();
                    infoStrings.add("   Normalization: ");
                    infoStrings.add("by gene - mean 0, std 1\n");
                    infoStyles.add("bold");
                    infoStyles.add("regular");
                    wasPreProcessed = true;
                } catch (Exception ex) {
                    ex.printStackTrace();
                    JOptionPane.showMessageDialog(this, "Unable to compute " +
                                                  "normalization.",
                                                  "Unable to normalize",
                                                  JOptionPane.ERROR_MESSAGE);
                }
            }
        }
        if (this.smoothCheckBox.isSelected()) {
            int window = 3;

            try {
                window = this.readParameterInteger(3,
                        preprocessed.getNumberOfConditions(),
                        PanelPreProcessing.SMOOTH_WINDOW);
            } catch (Exception ex) {
                ex.printStackTrace();
                return;
            }

            if ((window % 2) == 0) {
                JOptionPane.showMessageDialog(this, "The smooth " +
                                              "window must be an odd integer between 3 and " +
                                              preprocessed.
                                              getNumberOfConditions() + ".",
                                              "Smooth window must be odd value in [3," +
                                              preprocessed.
                                              getNumberOfConditions() + "[",
                                              JOptionPane.INFORMATION_MESSAGE);
                return;
            }

            try {
                preprocessed.computeSmoothedExpressionMatrix(window);
                infoStrings.add("   Smoothing: ");
                infoStrings.add("with window " + window + "\n");
                infoStyles.add("bold");
                infoStyles.add("regular");
                wasPreProcessed = true;
            } catch (Exception ex) {
                ex.printStackTrace();
                JOptionPane.showMessageDialog(this, "Unable to compute " +
                                              "smoothed expression matrix.",
                                              "Unable to smooth matrix",
                                              JOptionPane.ERROR_MESSAGE);
            }
        }
        /**
         * load the new preprocessed expression matrix in the dataset tree
         */
        NodeInfo preprocessedNodeInfo = null;
        if (wasPreProcessed) {
            ArrayList<String> newStrings = new ArrayList<String>();
            newStrings.add("Number of genes: ");
            newStrings.add("" + preprocessed.getNumberOfGenes() + "\n");
            newStrings.add("Number of time points: ");
            newStrings.add("" + preprocessed.getNumberOfConditions() + "\n\n");
            newStrings.add("Missing values: ");
            newStrings.add("" + preprocessed.hasMissingValues() + "\n");
            newStrings.add("Missings filled with value: ");
            newStrings.add("" + preprocessed.getMissingValue() + "\n\n");
            newStrings.add("Organism: ");
            newStrings.add(preprocessed.getOrganism() + "\n\n");
            newStrings.add("Preprocessing\n");

            ArrayList<String> newStyles = new ArrayList<String>();
            for (int i = 0; i < newStrings.size(); i++) {
                if (i % 2 == 0) {
                    newStyles.add("bold");
                } else {
                    newStyles.add("regular");
                }
            }

            newStrings.addAll(infoStrings);
            newStyles.addAll(infoStyles);

            try {
                preprocessedNodeInfo = new NodeInfo(preprocessed,
                        NodeInfo.PREPROCESSED,
                        selectionPath,
                        mainFrame.getTree(),
                        ((NodeInfo) ((DefaultMutableTreeNode) selectionPath.
                                     getLastPathComponent()).getUserObject()),
                        newStrings,
                        newStyles);
                if (!this.discretizeCheckBox.isSelected()) {
                    mainFrame.getTree().addNewNode(preprocessedNodeInfo,
                            selectionPath, true, true);
                    return;
                }
            } catch (InvalidNodeObjectTypeException ex) {
                ex.printStackTrace();
                JOptionPane.showMessageDialog(this, "Unable to load " +
                                              "the new preprocessed matrix in the tree.",
                                              "Unable to load preprocessed in the tree",
                                              JOptionPane.ERROR_MESSAGE);
                return;
            }
        }

        DiscretizedExpressionMatrix discretized;
        if (wasPreProcessed) {
            discretized = new DiscretizedExpressionMatrix(preprocessed);
        } else {
            discretized = new DiscretizedExpressionMatrix(this.expressionMatrix);
        }

        ArrayList<String> iStrings = new ArrayList<String>();
        ArrayList<String> iStyles = new ArrayList<String>();

        iStrings.add("Number of genes: ");
        iStyles.add("bold");
        iStrings.add("" +
                            discretized.getOriginalExpressionMatrix().getNumberOfGenes() +
                            "\n");
        iStyles.add("regular");
        iStrings.add("Number of time points: ");
        iStyles.add("bold");
        iStrings.add("" +
                            discretized.getOriginalExpressionMatrix().getNumberOfConditions() +
                            "\n\n");
        iStyles.add("regular");
        iStrings.add("Organism: ");
        iStyles.add("bold");
        if (expressionMatrix instanceof ExpressionMatrix) {
            iStrings.add(((ExpressionMatrix) expressionMatrix).
                                getOrganism() + "\n\n");
            iStyles.add("regular");
        } else {
            iStrings.add(((PreProcessedExpressionMatrix)
                                 expressionMatrix).getOrganism() + "\n\n");
            iStyles.add("regular");
        }
        iStrings.add("Discretization technique:\n");
        iStyles.add("bold");

        int algorithm = this.discretizeComboBox.getSelectedIndex();
        switch (algorithm) {
        case 0:
            /**
             * expression average
             */
            try {
                discretized.computeDiscretizedMatrix_Mean_OverallMatrix('D',
                        'U');
                iStrings.add("   expression average\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("expression average");
                return;
            }
            break;
        case 1:

            /**
             * gene expression average
             */
            try {
                discretized.computeDiscretizedMatrix_Mean_ByGene('D', 'U');
                iStrings.add("   gene expression average\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("gene expression average");
                return;
            }
            break
                    ;
        case 2:

            /**
             * mid-range
             */
            try {
                discretized.
                        computeDiscretizedMatrix_MidRanged_OverallMatrix(
                                'D', 'U');
                iStrings.add("   mid-range\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("mid-range");
                return;
            }
            break
                    ;
        case 3:

            /**
             * gene mid-range
             */
            try {
                discretized.computeDiscretizedMatrix_MidRanged_ByGene(
                        'D', 'U');
                iStrings.add("   gene mid-range\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("gene mid-range");
                return;
            }
            break
                    ;
        case 4:

            /**
             * max-minus percent-max
             */
            float percent;
            try {
                percent = (float)this.readParameterDouble(1.0, 100.0,
                        PanelPreProcessing.DISCRETIZE_PERCENT);
            } catch (Exception ex) {
                /**
                 * Messages were already displayed to the user.
                 */
                ex.printStackTrace();
                return;
            }
            try {
                discretized.
                        computeDiscretizedMatrix_MaxMinusPercentMax_OverallMatrix(
                                percent, 'D', 'U');
                iStrings.add("   max-minus percent-max\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("percentage - " + percent + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("max-minus percent-max");
                return;
            }
            break
                    ;
        case 5:

            /**
             * gene max-minus percent-max
             */
            try {
                percent = (float)this.readParameterDouble(1.0, 100.0,
                        PanelPreProcessing.DISCRETIZE_PERCENT);
            } catch (Exception ex) {
                ex.printStackTrace();
                return;
            }
            try {
                discretized.
                        computeDiscretizedMatrix_MaxMinusPercentMax_ByGene(
                                percent, 'D', 'U');
                iStrings.add("   gene max-minus percent-max\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("percentage - " + percent + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("gene max-minus percent-max");
                return;
            }
            break
                    ;
        case 6:

            /**
             * equal frequency
             */
            int numSymbols;
            try {
                numSymbols = this.readParameterInteger(2, 20,
                        PanelPreProcessing.DISCRETIZE_SYMBOLS);
            } catch (Exception ex) {
                return;
            }
            try {
                discretized.computeDiscretizedMatrix_EqualFrequency_K_Symbols_OverallMatrix(numSymbols);
                iStrings.add("   equal frequency\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("nr of symbols - " + numSymbols + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("equal frequency");
                return;
            }
            break;
        case 7:
            /**
             * equal frequency by gene
             */
            try {
                numSymbols = this.readParameterInteger(2, 20,
                        PanelPreProcessing.DISCRETIZE_SYMBOLS);
            } catch (Exception ex) {
                return;
            }
            try {
                discretized.computeDiscretizedMatrix_EqualFrequency_K_Symbols_ByGene(numSymbols);
                iStrings.add("   equal frequency by gene\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("nr of symbols - " + numSymbols + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("equal frequency by gene");
                return;
            }
            break;
        case 8:
            /**
             * equal width
             */
            try {
                numSymbols = this.readParameterInteger(2, 20,
                        PanelPreProcessing.DISCRETIZE_SYMBOLS);
            } catch (Exception ex) {
                return;
            }
            try {
                discretized.computeDiscretizedMatrix_EqualWidth_K_Symbols_OverallMatrix(numSymbols);
                iStrings.add("   equal width\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("nr of symbols - " + numSymbols + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("equal width");
                return;
            }
            break;
        case 9:
            /**
             * equal width by gene
             */
            try {
                numSymbols = this.readParameterInteger(2, 20,
                        PanelPreProcessing.DISCRETIZE_SYMBOLS);
            } catch (Exception ex) {
                return;
            }
            try {
                discretized.computeDiscretizedMatrix_EqualWidth_K_Symbols_ByGene(numSymbols);
                iStrings.add("   equal width by gene\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("nr of symbols - " + numSymbols + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("equal width by gene");
                return;
            }
            break;
        case 10:
            /**
             * expression mean and std
             */
            double alpha;
            try {
                alpha = this.readParameterDouble(0.0, 1.0,
                                                 PanelPreProcessing.DISCRETIZE_ALPHA);
            } catch (Exception ex) {
                return;
            }
            try {
                discretized.
                        computeDiscretizedMatrix_MeanAndStd_OverallMatrix(
                                alpha, 'D', 'N', 'U');
                iStrings.add("   expression mean and std\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("alpha - " + alpha + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("expression mean and std");
                return;
            }
            break
                    ;
        case 11:

            /**
             * gene expression mean and std
             */
            try {
                alpha = this.readParameterDouble(0.0, 1.0,
                                                 PanelPreProcessing.DISCRETIZE_ALPHA);
            } catch (Exception ex) {
                ex.printStackTrace();
                return;
            }
            try {
                discretized.
                        computeDiscretizedMatrix_MeanAndStd_OverallMatrix(
                                alpha, 'D', 'N', 'U');
                iStrings.add("   gene expression mean and std\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("alpha - " + alpha + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("gene expression mean and std");
                return;
            }
            break
                    ;
        case 12:

            /**
             * Transitional State Discrimination (TSD)
             */
            try {
                discretized.
                        computeDiscretizedMatrix_VariationsBetweenTimePoints_TSD(
                                'D', 'U');
                iStrings.add("   transitional state discrimination\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("TSD");
            }
            break
                    ;
        case 13:

            /**
             * variation between time points (2 symbols)
             */
            try {
                alpha = this.readParameterDouble(0.0, 1.0,
                                                 PanelPreProcessing.DISCRETIZE_ALPHA);
            } catch (Exception ex) {
                ex.printStackTrace();
                return;
            }
            try {
                discretized.
                        computeDiscretizedMatrix_VariationsBetweenTimePoints((float)
                        alpha, 'D', 'U');
                iStrings.add("   variations between time points\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("alpha - " + alpha + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("variation between time points");
            }
            break
                    ;
        case 14:

            /**
             * variation between time points (3 symbols)
             */
            double threshold;
            try {
                threshold = this.readParameterDouble(PanelPreProcessing.
                        DISCRETIZE_THRESHOLD);
            } catch (Exception ex) {
                ex.printStackTrace();
                return;
            }
            if (threshold <= 0.0) {
                JOptionPane.showMessageDialog(this, "The discretization " +
                                              "threshold must be greater" +
                                              " than 0",
                                              "Threshold must be > 0",
                                              JOptionPane.INFORMATION_MESSAGE);
                return;
            }
            try {
                discretized.
                        computeDiscretizedMatrix_VariationsBetweenTimePoints(
                                (float) threshold, 'D', 'N', 'U');
                iStrings.add("   variations between time points\n");
                iStyles.add("regular");
                iStrings.add("   parameter: ");
                iStyles.add("regular");
                iStrings.add("threshold - " + threshold + "\n");
                iStyles.add("regular");
            } catch (Exception ex) {
                ex.printStackTrace();
                this.showDiscretizeErrorMessage("variation between time points");
            }
            break
                    ;
        }

        /**
         * load preprocessed and discretized matrix in the dataset tree
         */
        try {
            char[] alphabet = discretized.getAlphabet();
            String alph = "{" + alphabet[0];
            for (int i = 1; i < alphabet.length; i++) {
                alph += ", " + alphabet[i];
            }
            alph += "}";
            iStrings.add("Alphabet: ");
            iStyles.add("bold");
            iStrings.add(alph + "\n");
            iStyles.add("regular");

            NodeInfo discretizedNodeInfo = null;
            if (wasPreProcessed) {
                mainFrame.getTree().addNewNode(
                        preprocessedNodeInfo,
                        selectionPath, false);

                discretizedNodeInfo =
                        new NodeInfo(discretized,
                                     NodeInfo.DISCRETIZED,
                                     selectionPath,
                                     mainFrame.getTree(),
                                     preprocessedNodeInfo,
                                     iStrings,
                                     iStyles);
            } else {
                discretizedNodeInfo =
                        new NodeInfo(discretized,
                                     NodeInfo.DISCRETIZED,
                                     selectionPath,
                                     mainFrame.getTree(),
                                     (NodeInfo) ((DefaultMutableTreeNode)
                                                 selectionPath.
                                                 getLastPathComponent()).
                                     getUserObject(),
                                     iStrings,
                                     iStyles);
            }
            mainFrame.getTree().addNewNode(discretizedNodeInfo, selectionPath, true, true);
        } catch (InvalidNodeObjectTypeException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "Unable to load the " +
                                          "discretized matrix in the " +
                                          "dataset tree.",
                                          "Unable to load discretized matrix",
                                          JOptionPane.INFORMATION_MESSAGE);
        }
    }

    /**
     * Handles the item state changed event fired when the user selects
     * the normalize by mean and std values radio button.
     *
     * @param e <code>ItemEvent</code> fired when the user selects
     * the normalize by mean and std values radio button
     */
    public void normalizeMeanStdValuesButton_itemStateChanged(ItemEvent e) {
        if (this.normalizeMeanStdValuesButton.isSelected()) {
            this.normalizeMeanValueTextField.setEnabled(true);
            this.normalizeStdValueTextField.setEnabled(true);
        }
    }

    /**
     * Handles the item state changed event fired when the user selects
     * the normalize by mean 0 and std 1 radio button.
     *
     * @param e <code>ItemEvent</code> fired when the user selects
     * the normalize by mean 0 and std 1 radio button
     */
    public void normalizeMean0Std1Button_itemStateChanged(ItemEvent e) {
        if (this.normalizeMean0Std1Button.isSelected()) {
            this.normalizeMeanValueTextField.setEnabled(false);
            this.normalizeStdValueTextField.setEnabled(false);
        }
    }

    /**
     * Handles the item state changed event fired when the user selects
     * the fill missings with a given value radio button.
     *
     * @param e <code>ItemEvent</code> fired when the user selects
     * the fill missings with a given value radio button
     */
    public void fillValueButton_itemStateChanged(ItemEvent e) {
        if (this.fillValueButton.isSelected()) {
            this.fillKNeighborsTextField.setEnabled(false);
            this.fillValueTextField.setEnabled(this.fillCheckBox.isSelected());
            this.fillKNeighborsLabel.setEnabled(this.fillCheckBox.isSelected());
        }
    }

    /**
     * Handles the item state changed event fired when the user selects
     * the fill missings with average over the k-neighbors radio button.
     *
     * @param e <code>ItemEvent</code> fired when the user selects
     * the fill missings with average over the k-neighbors radio button
     */
    public void fillKNeighborsButton_itemStateChanged(ItemEvent e) {
        if (this.fillKNeighborsButton.isSelected()) {
            this.fillValueTextField.setEnabled(false);
            this.fillKNeighborsLabel.setEnabled(this.fillCheckBox.isSelected());
            this.fillKNeighborsTextField.setEnabled(this.fillCheckBox.isSelected());
        }
    }

    /**
     * Handles the item state changed event fired when the user selects
     * the fill missings with average over the same row radio button.
     *
     * @param e <code>ItemEvent</code> fired when the user selects
     * the fill missings with average over the same row radio button
     */
    public void fillAverageRowButton_itemStateChanged(ItemEvent e) {
        if (this.fillAverageRowButton.isSelected()) {
            this.fillValueTextField.setEnabled(false);
            this.fillKNeighborsTextField.setEnabled(false);
            this.fillKNeighborsLabel.setEnabled(this.fillCheckBox.isSelected());
        }
    }

    /**
     * Returns the application's main frame.
     *
     * @return the main window
     */
    public BiggestsFrame getMainFrame() {
        return mainFrame;
    }

    /**
     * Returns the original or preprocessed expression matrix in which
     * preprocessing techniques will be performed.
     *
     * @return original or preprocessed <code>AbstractExpressionMatrix</code>
     * in which preprocessing techniques will be performed
     */
    public AbstractExpressionMatrix getExpressionMatrix() {
        return expressionMatrix;
    }

    /**
     * Returns the path selected on the dataset tree. Corresponds to the path in
     * the tree from the root node to the node that contains the original
     * or preprocessed matrix to apply preprocessing techniques.
     *
     * @return <code>TreePath</code> selected on the dataset tree
     */
    public TreePath getSelectionPath() {
        return selectionPath;
    }

    /**
     * Returns the layout of this panel.
     *
     * @return <code>SpringLayout</code>
     */
    public SpringLayout getSpringLayout() {
        return springLayout;
    }

    /**
     * Returns the filter check box.
     *
     * @return the filter <code>JCheckBox</code>
     */
    public JCheckBox getFilterCheckBox() {
        return filterCheckBox;
    }

    /**
     * Returns the filter options button group.
     *
     * @return the filter options <code>ButtonGroup</code>
     */
    public ButtonGroup getFilterButtonGroup() {
        return filterButtonGroup;
    }

    /**
     * Returns filter all genes with missing values button.
     *
     * @return filter all genes with missing values <code>JRadioButton</code>
     */
    public JRadioButton getFilterAllGenesButton() {
        return filterAllGenesButton;
    }

    /**
     * Returns filter genes with more than a % of missing values button.
     *
     * @return filter genes with more than a % of missing values
     * <code>JRadioButton</code>
     */
    public JRadioButton getFilterPercentGenesButton() {
        return filterPercentGenesButton;
    }

    /**
     * Returns the filter percent input text field.
     *
     * @return the filter percent input <code>JTextField</code>
     */
    public JTextField getFilterPercentTextField() {
        return filterPercentTextField;
    }

    /**
     * Returns filter percent label.
     *
     * @return filter percent <code>JLabel</code>
     */
    public JLabel getFilterPercentLabel() {
        return filterPercentLabel;
    }

    /**
     * Returns fill check box.
     *
     * @return fill check <code>JCheckBox</code>
     */
    public JCheckBox getFillCheckBox() {
        return fillCheckBox;
    }

    /**
     * Returns fill options button group.
     *
     * @return fill options <code>ButtonGroup</code>
     */
    public ButtonGroup getFillButtonGroup() {
        return fillButtonGroup;
    }

    /**
     * Returns fill missings with a certain value button.
     *
     * @return fill missings with a certain value <code>JRadioButton</code>
     */
    public JRadioButton getFillValueButton() {
        return fillValueButton;
    }

    /**
     * Returns fill with average over the k-neighbors button.
     *
     * @return fill with average over the k-neighbors <code>JRadioButton</code>
     */
    public JRadioButton getFillKNeighborsButton() {
        return fillKNeighborsButton;
    }

    /**
     * Returns fill with average over the same row button.
     *
     * @return fill with average over the same row <code>JRadioButton</code>
     */
    public JRadioButton getFillAverageRowButton() {
        return fillAverageRowButton;
    }

    /**
     * Returns fill with a certain value input text field.
     *
     * @return fill with a certain value input <code>JTextField</code>
     */
    public JTextField getFillValueTextField() {
        return fillValueTextField;
    }

    /**
     * Returns fill with average over the k-neighbors input text field.
     *
     * @return fill with average over the k-neighbors input <code>JTextField</code>
     */
    public JTextField getFillKNeighborsTextField() {
        return fillKNeighborsTextField;
    }

    /**
     * Returns fill with average over the k-neighbors label.
     *
     * @return fill with average over the k-neighbors <code>JLabel</code>
     */
    public JLabel getFillKNeighborsLabel() {
        return fillKNeighborsLabel;
    }

    /**
     * Returns normalize check box.
     *
     * @return normalize <code>JCheckBox</code>
     */
    public JCheckBox getNormalizeCheckBox() {
        return normalizeCheckBox;
    }

    /**
     * Returns normalize options button group.
     *
     * @return normalize options <code>ButtonGroup</code>
     */
    public ButtonGroup getNormalizeButtonGroup() {
        return normalizeButtonGroup;
    }

    /**
     * Returns normalize overall values button.
     *
     * @return normalize overall values <code>JRadioButton</code>
     */
    public JRadioButton getNormalizeOverallButton() {
        return normalizeOverallButton;
    }

    /**
     * Returns normalize by gene button.
     *
     * @return normalize by gene <code>JRadioButton</code>
     */
    public JRadioButton getNormalizeByGeneButton() {
        return normalizeByGeneButton;
    }

    /**
     * Returns normalize to mean 0 and std 1 button.
     *
     * @return normalize to mean 0 and std 1 <code>JRadioButton</code>
     */
    public JRadioButton getNormalizeMean0Std1Button() {
        return normalizeMean0Std1Button;
    }

    /**
     * Returns normalize by mean and std values button.
     *
     * @return normalize by mean and std values <code>JRadioButton</code>
     */
    public JRadioButton getNormalizeMeanStdValuesButton() {
        return normalizeMeanStdValuesButton;
    }

    /**
     * Return normalize mean input text field.
     *
     * @return normalize mean input <code>JTextField</code>
     */
    public JTextField getNormalizeMeanValueTextField() {
        return normalizeMeanValueTextField;
    }

    /**
     * Returns normalize std input text field.
     *
     * @return normalize std input <code>JTextField</code>
     */
    public JTextField getNormalizeStdValueTextField() {
        return normalizeStdValueTextField;
    }

    /**
     * Returns normalize to mean and std values label.
     *
     * @return normalize to mean and std values <code>JLabel</code>
     */
    public JLabel getNormalizeMeanStdValuesLabel() {
        return normalizeMeanStdValuesLabel;
    }

    /**
     * Returns normalize to mean and std options button group.
     *
     * @return normalize to mean and std options <code>ButtonGroup</code>
     */
    public ButtonGroup getNormalizeMeanStdButtonGroup() {
        return normalizeMeanStdButtonGroup;
    }

    /**
     * Returns smooth check box.
     *
     * @return smooth <code>JCheckBox</code>
     */
    public JCheckBox getSmoothCheckBox() {
        return smoothCheckBox;
    }

    /**
     * Returns smooth window label.
     *
     * @return smooth window <code>JLabel</code>
     */
    public JLabel getSmoothWindowLabel() {
        return smoothWindowLabel;
    }

    /**
     * Returns smooth window input text field.
     *
     * @return smooth window input <code>JTextField</code>
     */
    public JTextField getSmoothWindowTextField() {
        return smoothWindowTextField;
    }

    /**
     * Returns discretize check box.
     *
     * @return discretize <code>JCheckBox</code>
     */
    public JCheckBox getDiscretizeCheckBox() {
        return discretizeCheckBox;
    }

    /**
     * Returns discretize label.
     *
     * @return discretize <code>JLabel</code>
     */
    public JLabel getDiscretizeLabel() {
        return discretizeLabel;
    }

    /**
     * Returns discretize combo box.
     *
     * @return discretize <code>JComboBox</code>
     */
    public JComboBox getDiscretizeComboBox() {
        return discretizeComboBox;
    }

    /**
     * Returns discretize parameter input text field.
     *
     * @return discretize parameter input <code>JTextField</code>
     */
    public JTextField getDiscretizeParameterTextField() {
        return discretizeParameterTextField;
    }

    /**
     * Returns discretize parameter label.
     *
     * @return discretize parameter <code>JLabel</code>
     */
    public JLabel getDiscretizeParameterLabel() {
        return discretizeParameterLabel;
    }

    /**
     * Returns apply preprocessing options button.
     *
     * @return apply preprocessing options <code>JButton</code>
     */
    public JButton applyButton() {
        return applyButton;
    }

    public void fillCheckBox_itemStateChanged(ItemEvent e) {
        boolean selected = fillCheckBox.isSelected();
        this.fillAverageRowButton.setEnabled(selected);
        this.fillKNeighborsButton.setEnabled(selected);
        this.fillKNeighborsLabel.setEnabled(selected);
        this.fillKNeighborsTextField.setEnabled(false);
        this.fillValueButton.setEnabled(selected);
        this.fillValueTextField.setEnabled(false);

        this.fillAverageRowButton.setSelected(selected);
        this.fillAverageRowButton_itemStateChanged(e);
        this.fillKNeighborsButton_itemStateChanged(e);
        this.fillValueButton_itemStateChanged(e);
    }

    public void filterCheckBox_itemStateChanged(ItemEvent e) {
        boolean selected = filterCheckBox.isSelected();
        this.filterAllGenesButton.setEnabled(selected);
        this.filterPercentGenesButton.setEnabled(selected);
        this.filterPercentLabel.setEnabled(selected);
        this.filterPercentTextField.setEnabled(selected);
    }
}

class PanelPreProcessing_fillCheckBox_itemAdapter implements ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_fillCheckBox_itemAdapter(PanelPreProcessing adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.fillCheckBox_itemStateChanged(e);
    }
}


class PanelPreProcessing_filterCheckBox_itemAdapter implements ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_filterCheckBox_itemAdapter(PanelPreProcessing adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.filterCheckBox_itemStateChanged(e);
    }
}


class PanelPreProcessing_fillAverageRowButton_itemAdapter implements
        ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_fillAverageRowButton_itemAdapter(PanelPreProcessing
            adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.fillAverageRowButton_itemStateChanged(e);
    }
}


class PanelPreProcessing_fillValueButton_itemAdapter implements ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_fillValueButton_itemAdapter(PanelPreProcessing adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.fillValueButton_itemStateChanged(e);
    }
}


class PanelPreProcessing_fillKNeighborsButton_itemAdapter implements
        ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_fillKNeighborsButton_itemAdapter(PanelPreProcessing
            adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.fillKNeighborsButton_itemStateChanged(e);
    }
}


class PanelPreProcessing_normalizeMean0Std1Button_itemAdapter implements
        ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_normalizeMean0Std1Button_itemAdapter(PanelPreProcessing
            adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.normalizeMean0Std1Button_itemStateChanged(e);
    }
}


class PanelPreProcessing_normalizeMeanStdValuesButton_itemAdapter implements
        ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_normalizeMeanStdValuesButton_itemAdapter(
            PanelPreProcessing adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.normalizeMeanStdValuesButton_itemStateChanged(e);
    }
}


class PanelPreProcessing_normalizeByGeneButton_itemAdapter implements
        ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_normalizeByGeneButton_itemAdapter(PanelPreProcessing
            adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.normalizeByGeneButton_itemStateChanged(e);
    }
}


class PanelPreProcessing_normalizeCheckBox_itemAdapter implements ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_normalizeCheckBox_itemAdapter(PanelPreProcessing adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.normalizeCheckBox_itemStateChanged(e);
    }
}


class PanelPreProcessing_smoothCheckBox_itemAdapter implements ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_smoothCheckBox_itemAdapter(PanelPreProcessing adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.smoothCheckBox_itemStateChanged(e);
    }
}


class PanelPreProcessing_discretizeCheckBox_itemAdapter implements ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_discretizeCheckBox_itemAdapter(PanelPreProcessing
            adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.discretizeCheckBox_itemStateChanged(e);
    }
}


class PanelPreProcessing_discretizeComboBox_itemAdapter implements ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_discretizeComboBox_itemAdapter(PanelPreProcessing
            adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.discretizeComboBox_itemStateChanged(e);
    }
}


class PanelPreProcessing_applyButton_actionAdapter implements ActionListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_applyButton_actionAdapter(PanelPreProcessing adaptee) {
        this.adaptee = adaptee;
    }

    public void actionPerformed(ActionEvent e) {
        adaptee.applyButton_actionPerformed(e);
    }
}


class PanelPreProcessing_filterPercentGenesButton_itemAdapter implements
        ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_filterPercentGenesButton_itemAdapter(PanelPreProcessing
            adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.filterPercentGenesButton_itemStateChanged(e);
    }
}


class PanelPreProcessing_normalizeOverallButton_itemAdapter implements
        ItemListener {
    private PanelPreProcessing adaptee;
    PanelPreProcessing_normalizeOverallButton_itemAdapter(PanelPreProcessing
            adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e ItemEvent
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.normalizeOverallButton_itemStateChanged(e);
    }
}
