package biggests.utils;

import java.io.*;

import javax.swing.filechooser.FileFilter;

/**
 * <p>Title: File Filter by extension.</p>
 *
 * <p>Description: Defines a file filter which filters the files which
 *                 are shown to and can be selected by the user in a
 *                 file chooser dialog.
 *                 Valid files are chosen according to a pecific file
 *                 extension.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */

public class FileFilterExtension extends FileFilter {
    /**
     * File extension of the files shown (no other file extension
     * files are allowed).
     */
    private String validFileExtension;

    /**
     * Creates a new extension file filter, to filter files based on their
     * extensions.
     *
     * @param extension the <code>String</code> file extension to be filtered
     * @throws Exception when the file extension is not valid
     */
    public FileFilterExtension(String extension) throws Exception {
        if (!extension.startsWith(".")) {
            throw new Exception(extension + " is not a valid file extension.");
        }

        validFileExtension = extension;
    }

    /**
     * Checks if a certain file is acceptable under the filter conditions.
     * Returns <code>true</code> if <code>file</code> is accepted, or
     * <code>false</code> otherwise.
     *
     * @param file the <code>File</code> to check
     * @return a <code>boolean</code> which is <code>true</code> if the file
     *         is accepted or <code>false</code> otherwise     */
    public boolean accept(File file) {
        if (file.isDirectory()) {
            return true;
        }

        String name = file.getName().toLowerCase();
        return (name.endsWith(validFileExtension));
    }

    /**
     * Returns the description of this file filter.
     *
     * @return the <code>String</code> description of this file filter
     */
    public String getDescription() {
        return (validFileExtension.substring(1).toUpperCase() +
                " Files (*" + validFileExtension + ")");
    }

    /**
     * Returns this file filter valid file extension.
     *
     * @return the <code>String</code> valid file extension for this file filter
     */
    public String getValidFileExtension() {
        return validFileExtension;
    }

    /**
     * Checks if <code>filePath</code> has the required file
     * extension on it. If not, adds the file extension to the
     * path and returns the new path.
     *
     * @param filePath <code>String</code> the file path to check
     * @return String the file path with the valid file extension
     */
    public String checkAndAddFileExtension(String filePath) {
        if (!filePath.toLowerCase().endsWith(validFileExtension)) {
            return filePath + validFileExtension;
        }
        return filePath;
    }
}
